using System;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using Object = UnityEngine.Object;
using System.Text.RegularExpressions;

namespace Cotyounoyume
{
    public class AddonTakeControlItem : MonoBehaviour 
    {
		public Atom atom;
		public EBATCore core;
		public bool autoStopOnGlobal = true;
		public bool autoStopMorphOnGlobal = true;
		public bool autoStopHeadOnGlobal = true;
		public bool autoStopBodyOnGlobal = false;
		
		protected List<string> potentialConflictsTable;
		
		protected Dictionary<string, bool> statusList;
		public Dictionary<string, string> triggersStore;
		public Dictionary<string, string> animationPatternsTriggersStore;
		public Dictionary<string, string> animationPatternAnimatedObjectStore;
		public Dictionary<string, string> pluginIDTable;
		public Dictionary<string, List<string>> pluginIDTableList;
		public Dictionary<string, JSONClass> motionAnimationControlsStore;
//		public Dictionary<string, Dictionary<string, string>> timelineControllerStore;
		
		public string headCurrentRotationState = "";
		public FreeControllerV3 headControl;
//		public string snapTimeline;
		
//		public TimelineLink tlLink;
		
		public string[] defaultKey = {
			"[AS]Stop Morph Control by non-plugins",
			"[AS]Stop Head Control by non-plugins",
			"[AS]Stop Body Control by non-plugins",
		};

		public string[] defaultKeyPlus1 = {
			"[AS]AutoStop Other Plugins",
			"[AS]Stop Morph Control by non-plugins",
			"[AS]Stop Head Control by non-plugins",
			"[AS]Stop Body Control by non-plugins",
		};
		
		
		public AddonTakeControlItem(EBATCore core){
			atom = core.atom;
			this.core = core;
			statusList = new Dictionary<string, bool>();
			potentialConflictsTable = new List<string>();
			
			triggersStore = new Dictionary<string, string>();
			animationPatternsTriggersStore = new Dictionary<string, string>();
			animationPatternAnimatedObjectStore = new Dictionary<string, string>();
			pluginIDTable = new Dictionary<string, string>();
			motionAnimationControlsStore = new Dictionary<string, JSONClass>();
//			timelineControllerStore = new Dictionary<string, Dictionary<string, string>>();
			pluginIDTableList = new Dictionary<string, List<string>>();

			headControl = Utility.GetFreeControllerV3(core.atom, "headControl");
//			tlLink = new TimelineLink(core);
		}
		
		public void OnDestroy(){
			ReturnControlOfPlugins();
			
			triggersStore.Clear();
			statusList.Clear();
			animationPatternsTriggersStore.Clear();
			animationPatternAnimatedObjectStore.Clear();
			motionAnimationControlsStore.Clear();
			pluginIDTable.Clear();			
			pluginIDTableList.Clear();			
			potentialConflictsTable.Clear();
//			timelineControllerStore.Clear();
//			defaultKey.Clear();
//			defaultKeyPlus1.Clear();
		}

		public void LoadSettings(){
			try{
				autoStopOnGlobal = core.settings["General"]["AutoStop OtherPlugins"].AsBool;
				ReconstructPTMCArray();
				potentialConflictsTable.Clear();
				MakePotentialConflictsTable();
				MakePluginIDTable();
				
				if(core.settingsRoot["Common"].AsObject.HasKey("PTMC On/Off")){
					foreach(var name in potentialConflictsTable){
						if(!core.settingsRoot["Common"]["PTMC On/Off"].AsObject.HasKey(name)){
							if(name.StartsWith("[AS]Stop Body")){
								core.settingsRoot["Common"]["PTMC On/Off"][name] = "false";
							}else{
								core.settingsRoot["Common"]["PTMC On/Off"][name] = "true";
							}
						}
					}
				}else{
					foreach(var name in potentialConflictsTable){
						string value = (name.StartsWith("[AS]Stop Body")) ? "false" : "true";
						core.settingsRoot["Common"]["PTMC On/Off"][name] = value;
					}
				}
				
				if(core.settingsRoot["Common"].AsObject.HasKey("PTMC Timing")){
					foreach(var name in potentialConflictsTable){
						if(!core.settingsRoot["Common"]["PTMC Timing"].AsObject.HasKey(name)){
							core.settingsRoot["Common"]["PTMC Timing"][name] = "5";
						}
					}
				}else{
					foreach(var name in potentialConflictsTable){
						core.settingsRoot["Common"]["PTMC Timing"][name] = "5";
					}
				}
				
				if(core.settingsRoot["Common"].AsObject.HasKey("AutoStop On/Off")){
					foreach(var name in defaultKeyPlus1){
						if(!core.settingsRoot["Common"]["AutoStop On/Off"].AsObject.HasKey(name)){
							if(name.StartsWith("[AS]Stop Body")){
								core.settingsRoot["Common"]["AutoStop On/Off"][name] = "false";
							}else{
								core.settingsRoot["Common"]["AutoStop On/Off"][name] = "true";
							}
						}
					}
				}else{
					foreach(var name in defaultKeyPlus1){
						string value = (name.StartsWith("[AS]Stop Body")) ? "false" : "true";
						core.settingsRoot["Common"]["AutoStop On/Off"][name] = value;
					}
				}
			}catch (Exception e){
				core.LoggerE(e);
			}
		}
		
		public void AddPotentialConflictsTableIndividual(string pluginName){
			if(!core.settingsRoot["Common"]["Plugins that might Confilict"].AsObject.HasKey(pluginName)){
				potentialConflictsTable.Add(pluginName);
			}
			if(!core.settingsRoot["Common"]["PTMC On/Off"].AsObject.HasKey(pluginName)){
				core.settingsRoot["Common"]["PTMC On/Off"][name] = "true";
			}
			if(!core.settingsRoot["Common"]["PTMC Timing"].AsObject.HasKey(pluginName)){
				core.settingsRoot["Common"]["PTMC Timing"][name] = "5";
			}
		}
		
		public void MakePluginIDTable(){
			pluginIDTable.Clear();
			pluginIDTableList.Clear();
			List<string> ids = atom.GetStorableIDs();
			foreach(var id in ids){
				if(id.StartsWith("plugin#")){
					string pluginName = NormalizePluginName(id);
					if(!pluginIDTable.ContainsKey(pluginName)){
						pluginIDTable.Add(pluginName, id);
					}
					if(!pluginIDTableList.ContainsKey(pluginName)){
						pluginIDTableList.Add(pluginName, new List<string>());
					}
					pluginIDTableList[pluginName].Add(id);
				}
			}
		}
		
		public void ReconstructPTMCArray()
		{
			if(!core.settingsRoot["Common"].AsObject.HasKey("Plugins that might Confilict")) {
				core.settingsRoot["Common"]["Plugins that might Confilict"] = JSONNode.Parse("[]");
				return;
			}

			var dic = new Dictionary<string, bool>();
			foreach(var key in core.settingsRoot["Common"]["Plugins that might Confilict"].AsArray)
			{
				string newKey = key.ToString().Trim('"');
				if (!dic.ContainsKey(newKey))
				{
					dic.Add(newKey, true);
				}
			}

			core.settingsRoot["Common"].Remove("Plugins that might Confilict");
			core.settingsRoot["Common"]["Plugins that might Confilict"] = JSONNode.Parse("[]");
			List<string> keys = dic.Keys.ToList<string>();
			keys.Sort();
			foreach(var key in keys)
			{
				core.settingsRoot["Common"]["Plugins that might Confilict"].Add(key);
			}
		}

		public void MakePotentialConflictsTable(){
			Dictionary<string, bool> hashMap = new Dictionary<string, bool>();
			foreach (var name in core.settingsRoot["Common"]["Plugins that might Confilict"].AsArray)
			{
				string key = name.ToString().Trim('"');
				if (!hashMap.ContainsKey(key))
				{
					hashMap.Add(key, true);
					potentialConflictsTable.Add(key);
				}
			}

			foreach (var key in defaultKeyPlus1){
				string newKey = key.ToString().Trim('"');
				if(!hashMap.ContainsKey(newKey)){
					hashMap.Add(newKey, true);
					potentialConflictsTable.Add(newKey);
					core.settingsRoot["Common"]["Plugins that might Confilict"].Add(newKey);
				}
			}
			
		}
		
		public void ApplySettings(){
			if(Utility.HasThisPlugin(core.atom, "timeline")){
				if(IsTimelineVerOk()){

				}else{
					if(HasTimeline6()){
						core.hint.ReplaceTimeline();
					}
				}
			}
		}

		public void Manage(float time){
		}	
		
		public void OnOrgasmStart()
		{
		}
		
		public void OnOrgasmStop()
		{
		}
		
		public List<string> GetPluginList(string mode = "unlisted", bool distinctEnabled = true){
			List<string> result = new List<string>();
			List<string> ids = atom.GetStorableIDs();
			Dictionary<string, bool> dic = new Dictionary<string, bool>();
			foreach(var id in ids){
				if(id.StartsWith("plugin#")){
					string pluginName = NormalizePluginName(id);
					if(mode == "unlisted" && !IsPluginListed(pluginName)){
						if(!dic.ContainsKey(pluginName)){
							dic.Add(pluginName, true);
						}
					}else if(mode == "all"){
						if(!dic.ContainsKey(pluginName)){
							dic.Add(pluginName, true);
						}
					}
				}
			}

			foreach(KeyValuePair<string, bool> kvp in dic){
				result.Add(kvp.Key);
			}
			result.Sort();
			
			return result;
		}
		
		public string NormalizePluginName(string orig){
			string[] pattern = {"_"};
			string[] words = orig.Split(pattern, StringSplitOptions.None);
			string pluginPrefix = words[0];
			string pluginName = orig.Remove(0, pluginPrefix.Length+1);
			return pluginName;
		}
		
		public void AutoStopPlugins(){
			if(pluginIDTable == null){
				MakePluginIDTable();
			}
			List<string> loadedPluginList = GetPluginList("all");
			int currentStage = core.GetStage();
			
			string ngWord1 = "Cotyounoyume.EBATCore";
			string ngWord2 = core.vammoanLinkOnGlobal ? "VAMMoanPlugin.VAMMoan" : "";
			string ngWord3 = "VamTimeline.AtomPlugin";
			foreach(var loadedPlugin in loadedPluginList){
				if(loadedPlugin == ngWord1){ continue; };
				if(loadedPlugin == ngWord2){ continue; };
				if(loadedPlugin == ngWord3){ continue; };
				
				if(IsPluginWithPotentialConflicts(loadedPlugin)){
					
					if(!pluginIDTableList.ContainsKey(loadedPlugin)){
						MakePluginIDTable();
					}
					if(IsStopTiming(loadedPlugin, currentStage)){
						if(!pluginIDTable.ContainsKey(loadedPlugin)){
							core.Info("ERROR: Unknown Bug1: " + loadedPlugin);
						}else{
							foreach(var pluginName in pluginIDTableList[loadedPlugin]){
								TemporarilyDeactivatePlugin(pluginName);
							}
						}
					}else{
						foreach(var pluginName in pluginIDTableList[loadedPlugin]){
							ReactivatePlugin(pluginName);
						}
					}
				}
			}

			foreach(var name in defaultKey){
				if(core.settingsRoot["Common"]["AutoStop On/Off"][name].AsBool){
					if(currentStage >= core.settingsRoot["Common"]["PTMC Timing"][name].AsInt){
						TakeControlOfTarget(name);
					}else{
						ReturnControlOfTarget(name);
					}
				}
			}			
		}
		
		public void TakeControlOfPlugins(){
			List<string> ids = atom.GetStorableIDs();
			foreach(var id in ids){
				if(id.StartsWith("plugin#")){
					string pluginName = NormalizePluginName(id);
					if(IsPluginWithPotentialConflicts(pluginName)){
						TemporarilyDeactivatePlugin(id);
					}
				}
			}
		}
		
		public bool IsStopTiming(string name, int currentStage){
			bool result = false;
			int stageTiming = core.settingsRoot["Common"]["PTMC Timing"][name].AsInt;
			if(currentStage >= stageTiming){
				result = true;
			}else{
				result = false;
			}
			return result;
		}
		
		public bool IsPluginListed(string id){
			if(core.settingsRoot["Common"]["PTMC On/Off"].AsObject.HasKey(id)){
				return true;
			}else{
				return false;
			}
		}
		
		public bool IsPluginWithPotentialConflicts(string id){
			if(IsPluginListed(id) && core.settingsRoot["Common"]["PTMC On/Off"][id].AsBool){
				return true;
			}else{
				return false;
			}
		}

		public void TemporarilyDeactivatePlugin(string id){
			JSONStorable jss = atom.GetStorableByID(id);
			bool currentStatus = jss.GetBoolParamValue("enabled");
			if(statusList.ContainsKey(id)){
//				statusList[id] = currentStatus;
			}else{
				statusList.Add(id, currentStatus);				
			}
			jss.SetBoolParamValue("enabled", false);
		}
		
		public void ReturnControlOfPlugins(){
			foreach(KeyValuePair<string, bool> kvp in statusList){
				ReactivatePlugin(kvp.Key);
			}
			statusList.Clear();
		}
		
		public void ReactivatePlugin(string id){
			JSONStorable jss = atom.GetStorableByID(id);
			if(statusList.ContainsKey(id)){
				jss.SetBoolParamValue("enabled", statusList[id]);
			}
		}
		
		#region TakeControl
		public void TakeControlOfTarget(string name, bool isMorph = true, bool isSpecificControl = true){
			switch(name){
				case "[AS]Stop Morph Control by non-plugins":
					StopVariableTrigger(bMorph:isMorph);
					StopAnimationPattern(bMorph:isMorph);
					core.expression2.StopAllExpressionInPose();
					break;
				case "[AS]Stop Head Control by non-plugins":
					StopAnimatedObject(bSpecificControl:true);
					StopMotionAnimation(bSpecificControl:true);
					StopTimeline(bSpecificControl:true);
					break;
				case "[AS]Stop Body Control by non-plugins":
					StopAnimatedObject(bSpecificControl:false);
					StopMotionAnimation(bSpecificControl:false);
					StopTimeline(bSpecificControl:false);
					break;
			}
		}		
		
		public void TakeControlOfEverythingButPlugins(bool isMorph = true, bool isSpecificControl = true){
			StopAnimationPattern(bMorph:isMorph);
			StopVariableTrigger(bMorph:isMorph);
			StopAnimatedObject(bSpecificControl:isSpecificControl);
			StopMotionAnimation(bSpecificControl:isSpecificControl);
			StopTimeline(bSpecificControl:isSpecificControl);
		}
		
		public void ReturnControlOfTarget(string name){
			switch(name){
				case "[AS]Stop Morph Control by non-plugins":
					RestartVariableTrigger();
					RestartAnimationPattern();
					break;
				case "[AS]Stop Head Control by non-plugins":
					RestartAnimatedObject();
					RestartMotionAnimation();
					RestartTimeline(bSpecificControl:true);
					break;
				case "[AS]Stop Body Control by non-plugins":
					RestartAnimatedObject();
					RestartMotionAnimation();
					RestartTimeline(bSpecificControl:false);
					break;				
			}
		}
		
		public void ReturnControlOfEverythingButPlugins(){
			RestartAnimationPattern();
			RestartVariableTrigger();
			RestartAnimatedObject();
			RestartMotionAnimation();
			RestartTimeline();
		}
		
		public void StopAnimationPattern(bool bMorph = true){
			foreach(KeyValuePair<string, AnimationPattern> kvp in Utility.CollectAnimationPatterns()){
				if(Utility.ContainsAtomInTriggers(core.atom, kvp.Value, isMorph:bMorph)){
					animationPatternsTriggersStore.Add(kvp.Key, kvp.Value.GetJSON().ToString());
					JSONClass erasedJsc = Utility.MakeJSCWithAtomRelatedTriggersExcluded(core.atom, kvp.Value, isMorph:bMorph);
					kvp.Value.LateRestoreFromJSON(erasedJsc);
				}
			}
		}
					
		public void RestartAnimationPattern(){
			foreach(KeyValuePair<string, AnimationPattern> kvp in Utility.CollectAnimationPatterns()){
				string atomUid = kvp.Key;
				if(animationPatternsTriggersStore.ContainsKey(atomUid)){					
					JSONClass jsc = JSONNode.Parse(animationPatternsTriggersStore[atomUid]).AsObject;
					kvp.Value.LateRestoreFromJSON(jsc);
					animationPatternsTriggersStore.Remove(atomUid);
				}
			}			
		}
					
		public void StopAnimatedObject(bool bSpecificControl = true){
			foreach(KeyValuePair<string, JSONStorable> kvp in Utility.CollectAnimatedObjects()){
				if(Utility.ContainsAtomInAnimationTarget(core.atom, kvp.Value, isSpecificControl:bSpecificControl)){
					animationPatternAnimatedObjectStore.Add(kvp.Key, kvp.Value.GetJSON().ToString());
					JSONClass erasedJsc = Utility.MakeJSCWithAtomRelatedReceiverExcluded(core.atom, kvp.Value, isSpecificControl:bSpecificControl);
					kvp.Value.RestoreFromJSON(erasedJsc);
				}
			}
		}
				
		public void RestartAnimatedObject(){
			foreach(KeyValuePair<string, JSONStorable> kvp in Utility.CollectAnimatedObjects()){
				string atomUid = kvp.Key;
				if(animationPatternAnimatedObjectStore.ContainsKey(atomUid)){
					JSONClass jsc = JSONNode.Parse(animationPatternAnimatedObjectStore[atomUid]).AsObject;
					kvp.Value.RestoreFromJSON(jsc);
					animationPatternAnimatedObjectStore.Remove(atomUid);
				}
			}
		}

		public void StopVariableTrigger(bool bMorph = true){
			foreach(KeyValuePair<string, VariableTrigger> kvp in Utility.CollectVariableTriggers()){
				if(Utility.ContainsAtomInTrigger(core.atom, kvp.Value, isMorph:bMorph)){
					triggersStore.Add(kvp.Key, kvp.Value.GetJSON().ToString());
					JSONClass erasedJsc = Utility.MakeJSCWithAtomRelatedTriggerExcluded(core.atom, kvp.Value, isMorph:bMorph);
					kvp.Value.LateRestoreFromJSON(erasedJsc);
				}
			}
		}
		
		public void RestartVariableTrigger(){
			foreach(KeyValuePair<string, VariableTrigger> kvp in Utility.CollectVariableTriggers()){
				string atomUid = kvp.Key;
				if(triggersStore.ContainsKey(atomUid)){
					JSONClass jsc = JSONNode.Parse(triggersStore[atomUid]).AsObject;
					kvp.Value.LateRestoreFromJSON(jsc);
					triggersStore.Remove(atomUid);
				}
			}
		}
		
		public void StopMotionAnimation(bool bSpecificControl = true){
			MotionAnimationControl[] MAC = core.atom.motionAnimationControls;
			foreach(var mac in MAC){
				JSONClass jsc = mac.GetJSON();
				string id = jsc["id"].Value.Trim('"');
				if(!bSpecificControl || Utility.CheckSpecificControl(id)){
					if(!motionAnimationControlsStore.ContainsKey(id)){
						motionAnimationControlsStore.Add(id, jsc);
					}else{
						motionAnimationControlsStore[id] = jsc;
					}
					mac.ClearAnimation();
				}
			}
		}
		
		public void RestartMotionAnimation(){
			MotionAnimationControl[] MAC = core.atom.motionAnimationControls;
			for(int i = 0; i < MAC.Length; i++){
				JSONClass jsc = MAC[i].GetJSON();
				string id = jsc["id"].Value.Trim('"');
				if(motionAnimationControlsStore.ContainsKey(id)){
					MAC[i].RestoreFromJSON(motionAnimationControlsStore[id]);
				}
			}
			motionAnimationControlsStore.Clear();
		}
		
		public bool IsTimelineVerOk(){
			int ver = Utility.GetPluginVer(core.atom, "VamTimeline.AtomPlugin");
			return (ver >= 253);
		}
		
		public bool HasTimeline6(){
			string varfile = Utility.GetLongestFilename("*Timeline*.var");
			int ver = 0;
			if(varfile != ""){
				Match m = Regex.Match(varfile, "^.+\\.(\\d+)\\.var$");
				if(m != null){
					ver = Convert.ToInt32(m.Groups[1].Value);			
				}
			}
			return (ver >= 253);
		}
		
		
		
		public void StopTimeline(bool bMorph = true, bool bSpecificControl = true){
			if(Utility.HasThisPlugin(core.atom, "timeline")){
				if(IsTimelineVerOk()){
					JSONStorable jsTimeline = Utility.GetReceiver(core.atom, "timeline");
					if(jsTimeline == null){ return; };
					
					foreach(var name in jsTimeline.GetFloatParamNames()){
						if(name.Contains("Controller Weight")){
							Match m = Regex.Match(name, "^Controller Weight (.*)$");
							string control = "";
							if(m != null){
								control = m.Groups[1].Value;
							}
							bool isTarget = !bSpecificControl || Utility.CheckSpecificControl(control);
							if(isTarget){
								core.StartCoroutine(SmoothUpDownWeightCoroutine(jsTimeline, name, false));
							}
						}
					}
				}
			}
		}
		
		public void RestartTimeline(bool bSpecificControl = true){
			if(Utility.HasThisPlugin(core.atom, "timeline")){
				if(IsTimelineVerOk()){
					JSONStorable jsTimeline = Utility.GetReceiver(core.atom, "timeline");
					if(jsTimeline == null){ return; };
					
					foreach(var name in jsTimeline.GetFloatParamNames()){
						if(name.Contains("Controller Weight")){
							Match m = Regex.Match(name, "^Controller Weight (.*)$");
							string control = "";
							if(m != null){
								control = m.Groups[1].Value;
							}
							bool isTarget = !bSpecificControl || Utility.CheckSpecificControl(control);
							if(isTarget){
								core.StartCoroutine(SmoothUpDownWeightCoroutine(jsTimeline, name, true));
							}
						}
					}
				}
			}
		}
		
		public IEnumerator SmoothUpDownWeightCoroutine(JSONStorable js, string control, bool bAscension){
			int step = 1;
			float interval = 0.1f;
			int from =  0;
			int min  =  0;
			int max  = 10;
			if(!bAscension){
				step = -1;
				interval = 0.02f;
				from = 10;
			}
			for(int i = from; (i >= min && i <= max); i += step){
				float f = (float)i/(float)max;
				js.SetFloatParamValue(control, f);
				yield return new WaitForSeconds(interval);
			}
			yield return null;
		}
		#endregion
	}
}