#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
#endregion

namespace Cotyounoyume{
	public class Belly
	{
		public Atom atom;
		public EBATCore core;
		public bool disableLog;

		
		protected Dictionary<string, string> backupRotationState;
		protected Dictionary<string, Quaternion> targetQuat;
		protected Dictionary<string, Quaternion> recoveryQuat;
		protected Dictionary<string, float> progress;
		protected Dictionary<string, CycleManager> timing;
		protected Dictionary<string, FreeControllerV3> ctrl;
		protected Dictionary<string, bool> isActive;

		protected string[] keyBreath = {
			"chestControl"
		};
		
		protected string[] keyOrgasm = {
			"abdomen2Control", 
			"pelvisControl", 
//			"abdomenControl", 
//			"hipControl", 
		};
				
		public Dictionary<string, float> targetStrength = new Dictionary<string, float>
		{
			["headControl"]       = -10
			, ["neckControl"]     = -10
			, ["chestControl"]    = -10
			, ["abdomenControl"]  = -30
			, ["abdomen2Control"] = -30
			, ["pelvisControl"]   = -30
			, ["hipControl"]      = -30
			, ["lThighControl"]   = 30
			, ["lKneeControl"]    = 30
			, ["lFootControl"]    = 30
			, ["rThighControl"]   = 30
			, ["rKneeControl"]    = 30
			, ["rFootControl"]    = 30
		};

		public Dictionary<string, float> closingTimeOffset1 = new Dictionary<string, float>
		{
			["breath"] = 1.2f
		};
		public Dictionary<string, float> closingTimeOffset2 = new Dictionary<string, float>
		{
			["breath"] = 1.2f
		};
		
		protected int waveState;
		protected bool isBellyActive;
		
		
		protected FreeControllerV3 pelvis;
		
		private float lastTime;
		
		public Belly(EBATCore core){
			atom = core.atom;
			this.core = core;
		}
		
		public void LoadSettings(){
			try{
				targetStrength["chestControl"] = core.settings["General"]["Breath Strength"].AsFloat;
				
				pelvis = Utility.GetFreeControllerV3(core.atom, "pelvisControl");
			}catch (Exception e){
				core.LoggerE(e);
			}

		}
		
		public void ApplySettings(){
			SetControllerAndProgress();
		}
		
		public void Maintenance(){

		}

		public void SetControllerAndProgress(){
			try{
				ctrl   = new Dictionary<string, FreeControllerV3>();
				progress = new Dictionary<string, float>();
				timing   = new Dictionary<string, CycleManager>();
				isActive = new Dictionary<string, bool>();
				backupRotationState = new Dictionary<string, string>();

				recoveryQuat = new Dictionary<string, Quaternion>();
				targetQuat = new Dictionary<string, Quaternion>();
				
				string[] keys = new string[keyBreath.Length + keyOrgasm.Length];
				Array.Copy(keyBreath, 0, keys, 0, keyBreath.Length);
				Array.Copy(keyOrgasm, 0, keys, keyBreath.Length, keyOrgasm.Length);
				
				for(int i = 0; i < keys.Length; i++){
					ctrl[keys[i]] = atom.GetStorableByID(keys[i]) as FreeControllerV3;
					progress[keys[i]] = 0;
				}
				isActive.Add("breath", false);
				timing.Add("breath", new CycleManager());					
				
			}catch (Exception e){
				core.LoggerE(e);
			}
		}

		public void UpdatePosition(float time, string[] keys, string position){
			try{
				if(isActive[position]){
					return;
				}else{
					if(core.trunk == null){
						return;
					}
					if(core.GetStage() >= 2){
						if(core.isEmbarassed){
							core.expression2.Update();
						}
						core.sound.PlaySound("breath", Time.time);
					}

					ChangeIsActive(true);
					float time1 = time + closingTimeOffset1[position] * UnityEngine.Random.Range(0.7f,1.0f) * AdjustInterval();
					float time2 = time1 + closingTimeOffset2[position] * UnityEngine.Random.Range(0.7f,1.0f) * AdjustInterval();
					timing[position].SetTimer(time, time1, time2);

					float tmpTargetStrength;
					FreeControllerV3 baseController = GetBaseController(position);
					for(int i = 0; i < keys.Length; i++){
						tmpTargetStrength  = targetStrength[keys[i]] * UnityEngine.Random.Range(0.8f,1.0f) * AdjustStrength();
						recoveryQuat[keys[i]] = Quaternion.AngleAxis(-0.5f * tmpTargetStrength, core.trunk.transform.right) * core.trunk.transform.rotation;
						targetQuat[keys[i]]   = Quaternion.AngleAxis( 0.5f * tmpTargetStrength, core.trunk.transform.right) * core.trunk.transform.rotation;
					}
				}
			}catch (Exception e){
				core.LoggerE(e);
			}
		}

		public FreeControllerV3 GetBaseController(string position){
			FreeControllerV3 result;
			if(position == "breath"){
				result = ctrl["chestControl"];
			}else{
				result = null;
			}
			return result;
		}
		
		public void CalcProgress(float time, string position){
			try{
				progress[position] = timing[position].GetProgress(time);
				
				if(timing[position].GetState(time) == 0){
					ChangeIsActive(false);
				}
			}catch (Exception e){
				core.LoggerE(e);
			}
		}
		
		public void ChangeJoint(float time, string[] keys, string position){
			try{
				if(!timing.ContainsKey(position)){
					core.Info("ERROR: ChangeJoint: key:" + position + " not found.");
				}
				int state = timing[position].GetState(time);
				for(int i = 0; i < keys.Length; i++){
					if(state != 0){
						if(!ctrl.ContainsKey(keys[i])){
							core.Info("ERROR: ChangeJoint: ctrl key:" + keys[i] + " not found.");
						}
						if(!targetQuat.ContainsKey(keys[i])){
							core.Info("ERROR: ChangeJoint: targetQuat key:" + keys[i] + " not found.");
						}
						if(!progress.ContainsKey(position)){
							core.Info("ERROR: ChangeJoint: progress key:" + position + " not found.");
						}
						ctrl[keys[i]].transform.rotation = Quaternion.Slerp(recoveryQuat[keys[i]], targetQuat[keys[i]], progress[position]);
					}
				}
			}catch (Exception e){
				core.LoggerE(e);
			}
		}

		public float AdjustInterval(){
			float stageCoefficient = (float)core.GetStage() * 0.1f;
			return (1 / (1 + stageCoefficient));
		}
		
		public float AdjustStrength(){
			float stageCoefficient = (float)core.GetStage() * 0.2f;
			return 1 + stageCoefficient;
		}
		
		public void Manage(float time){
			try{
				if(!core.IsLoaded()){ return; }
				
				if(time == lastTime){
					time += 0.001f;
				}else if(time < lastTime){
					time = lastTime + 0.001f;
				}
				lastTime = time;

				if(core.GetStage() < 1){
					ChangeIsActive(false);
					return;
				}
				if(isActive["breath"] == false){
					UpdatePosition(time, keyBreath, "breath");
					core.eyesControl.Update();
					return;
				}

				CalcProgress(time, "breath");
				ChangeJoint(time, keyBreath, "breath");
			}catch (Exception e){
				core.LoggerE(e);
			}
		}
		
		public void ChangeIsActive(bool isOn){
			if(isOn){
				isActive["breath"] = true;
				foreach(var key in keyBreath){
					string rotationState = ctrl[key].currentRotationState.ToString();
					if(rotationState != "On"){
						if(backupRotationState.ContainsKey(key)){
							backupRotationState[key] = rotationState;
						}else{
							backupRotationState.Add(key, rotationState);
						}
//						backupRotationState.TryAdd(key, rotationState);
						ctrl[key].SetRotationStateFromString("On");
					}
				}
			}else{
				isActive["breath"] = false;
				foreach(var key in keyBreath){
					if(backupRotationState.ContainsKey(key)){
						string rotationState = backupRotationState[key];
						ctrl[key].SetRotationStateFromString(rotationState);
					}
				}
			}
			
		}
		
		public void BumpTest(){
			core.StartCoroutine(BellyShiveringCoroutine(3, 8, 0.03f, 0.01f, 0.03f, 0.02f, 0.2f, "Lock"));
		}
		
		public IEnumerator BellyShiveringCoroutine(int waveCycleCount = 3, int waveSmallCountPerCycle = 4, float strengthBig = 0.05f, float strengthSmall = 0.02f, float durationBump = 0.05f, float durationBumpInterval = 0.05f, float durationDentInterval = 0.2f, string type = "On"){
			for(int i = 0; i < waveCycleCount; i++){
				core.StartCoroutine(BellyShiveringWaveCoroutine(strengthBig, durationBump, type));
				yield return new WaitForSeconds(durationBumpInterval);
				for(int j = 0; j < waveSmallCountPerCycle; j++){
					core.StartCoroutine(BellyShiveringWaveCoroutine(strengthSmall, durationBump, type));
					yield return new WaitForSeconds(durationBumpInterval);
				}
				yield return new WaitForSeconds(durationDentInterval);
			}
		}
		
		public void SetPositionState(ref string positionState, string type = "On", string bodyName = "pelvisControl"){
			FreeControllerV3 fc = ctrl[bodyName];
			if(fc.currentPositionState.ToString() != type){
				positionState = fc.currentPositionState.ToString();
				fc.SetPositionStateFromString(type);
			}
		}

		public void ResetPositionState(ref string positionState, string bodyName = "pelvisControl"){
			FreeControllerV3 fc = ctrl[bodyName];
			fc.SetPositionStateFromString(positionState);
		}
		
		public void SetRotationState(ref string rotationState, string type = "On", string bodyName = "pelvisControl"){
			FreeControllerV3 fc = ctrl[bodyName];
			if(fc.currentRotationState.ToString() != type){
				rotationState = fc.currentPositionState.ToString();
				fc.SetRotationStateFromString(type);
			}
		}

		public void ResetRotationState(ref string rotationState, string bodyName = "pelvisControl"){
			FreeControllerV3 fc = ctrl[bodyName];
			fc.SetRotationStateFromString(rotationState);
		}
		
		public IEnumerator BellyShiveringWaveCoroutine(float strength, float duration, string type = "On"){
			Vector3 origPos = Utility.GetPosition(pelvis.transform);
			Vector3 fromPos = Utility.GetPositionFromTargetTransformWithOffset(pelvis.transform, forwardOffset:strength * 0.5f);
			Vector3 toPos = Utility.GetPositionFromTargetTransformWithOffset(pelvis.transform, forwardOffset:strength * -1f);
			Quaternion rot = pelvis.transform.rotation;

			string positionState = pelvis.currentPositionState.ToString();
			SetPositionState(ref positionState, type);
			
			pelvis.transform.SetPositionAndRotation(toPos, rot);
			yield return new WaitForSeconds(duration);
			pelvis.transform.SetPositionAndRotation(origPos, rot);
			ResetPositionState(ref positionState);
		}
		
		public void BellyBump(float strength){
			try{
				if(!core.bellyBoneOn){ return; }
				for(int i = 0; i < keyOrgasm.Length; i++){
					ChangeRotation(keyOrgasm[i], targetStrength[keyOrgasm[i]] * strength);
				}
			}catch (Exception e){
				core.LoggerE(e);
			}
		}

		public void ChangeRotation(string bodyName, float strength){
			FreeControllerV3 fc = ctrl[bodyName];
			string rotationState = fc.currentRotationState.ToString();
			SetRotationState(ref rotationState, bodyName);
			
			if(bodyName.StartsWith("l")){
				fc.jointRotationDriveZTarget = strength * -1;
			}else if(bodyName.StartsWith("r")){
				fc.jointRotationDriveZTarget = strength;						
			}else{
				fc.jointRotationDriveXTarget = strength;
			}
			ResetRotationState(ref rotationState, bodyName);
		}
		
		public void ChangePosition(string bodyName, float strength){
			FreeControllerV3 fc = ctrl[bodyName];

			Vector3 origPos = Utility.GetPosition(fc.transform);
			Vector3 fromPos = Utility.GetPositionFromTargetTransformWithOffset(fc.transform, forwardOffset:strength * 1f);
			Vector3 toPos = Utility.GetPositionFromTargetTransformWithOffset(fc.transform, forwardOffset:strength * -1f);
			Quaternion rot = fc.transform.rotation;

			string positionState = fc.currentPositionState.ToString();
			SetPositionState(ref positionState, bodyName);
			
			fc.transform.SetPositionAndRotation(toPos, rot);
			
			ResetPositionState(ref positionState, bodyName);
		}

		public void OnDestroy(){
			RemoveX(timing);
			ctrl.Clear();
			targetQuat.Clear();
			recoveryQuat.Clear();
			progress.Clear();
			isActive.Clear();
			targetStrength.Clear();
			closingTimeOffset1.Clear();
			closingTimeOffset2.Clear();
			backupRotationState.Clear();
			keyBreath = null;
			keyOrgasm = null;
		}
		
		public void RemoveX(Dictionary<string, CycleManager> dic){
			List<string> keys = new List<string>(dic.Keys);
			foreach(var key in keys)
			{
				if(dic.ContainsKey(key)) dic[key] = null;
			}
			dic.Clear();
		}
	}
}