#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
#endregion

namespace Cotyounoyume{
	public class BodyMorph
	{
		public Atom atom;
		public float bodyMorphClosingTimeOffset = 10f;
		public EBATCore core;
		public bool disableLog;
		public bool nippleErectionOnGlobal = true;
		public bool clitorisErectionOnGlobal = true;
		public bool isBodyMorphActive = false;
		
		public Dictionary<string, string> otherMorphsMap = new Dictionary<string,string>(){
			{"Anus",       "CTY_ANUS Genital"},
			// {"Fingers",    "CTY_FINGERS"},
			{"Shivering1", "CTY_Shivering"},
			{"Shivering2", "CTY_Shivering2"},
			{"ShiveringP", "CTY_ShiveringPlus"},
		};
		
		public Dictionary<string, bool> otherMorphsStatus = new Dictionary<string, bool>(){
			{"Anus",       true},
			// {"Fingers",    true},
			{"Shivering1", true},
			{"Shivering2", true},
			{"ShiveringP", true},			
		};
		
		public float strengthFingerHighMin = 0.3f;
		public float strengthFingerHighMax = 0.5f;
		public float strengthFingerLowMin = 1.0f;
		public float strengthFingerLowMax = 0.8f;
		public float strengthShiveringLowMin = -0.1f;
		public float strengthShiveringLowMax = -0.2f;
		public float strengthShiveringHighMin = 0.5f;
		public float strengthShiveringHighMax = 0.7f;
		public float strengthShiveringIntervalMin =  0.0f;
		public float strengthShiveringIntervalMax =  0.1f;
		
		protected Dictionary<string, DAZMorph> Bmorph;
		protected float bodyMorphClosingTime;
		
		public BodyMorph(EBATCore core){
			try{
				atom = core.atom;
				isBodyMorphActive = false;
				this.core = core;
			}catch(Exception e){
				core.LoggerE(e);
			}
		}
		
		public void LoadSettings(){		
			try{
				nippleErectionOnGlobal = core.settings["General"]["Nipple Erection On"].AsBool;
				clitorisErectionOnGlobal = core.settings["General"]["Clitoris Erection On"].AsBool;
				bodyMorphClosingTimeOffset = core.settings["General"]["BodyMorph Duration Time"].AsFloat;

				SetOtherMorphs();
				
				if(!core.settings.HasKey("BodyMorphOnOff")){
					foreach(var kvp in otherMorphsMap){
						core.settings["BodyMorphOnOff"][kvp.Key] = "true";
					}
				}
				UpdateOtherMorphsStatus();

			}catch(Exception e){
				core.LoggerE(e);
			}
		}

		public void SetOtherMorphs(){
			JSONNode jsBmorph = core.settings["BodyMorph"];
			Bmorph = new Dictionary<string, DAZMorph>();
				
			JSONStorable geometry = atom.GetStorableByID("geometry");
			DAZCharacterSelector character = geometry as DAZCharacterSelector;
			GenerateDAZMorphsControlUI morphControl = character.morphsControlUI;
			
			Bmorph["Nipple"] = morphControl.GetMorphByDisplayName(jsBmorph["Nipple"]);
			Bmorph["Clitoris"] = morphControl.GetMorphByDisplayName(jsBmorph["Clitoris"]);
			foreach(var kvp in otherMorphsMap){
				Bmorph[kvp.Key] = morphControl.GetMorphByDisplayName(kvp.Value);
			}
		}
		
		public void UpdateOtherMorphsStatus(){
			foreach(var kvp in otherMorphsMap){
				if(core.settings["BodyMorphOnOff"].AsObject.HasKey(kvp.Key)){
					if(otherMorphsStatus.ContainsKey(kvp.Key)){
						otherMorphsStatus[kvp.Key] = core.settings["BodyMorphOnOff"][kvp.Key].AsBool;
					}else{
						otherMorphsStatus.Add(kvp.Key, core.settings["BodyMorphOnOff"][kvp.Key].AsBool);
					}
				}
			}
		}
		
		public void ApplySettings(){

		}
		
		public void Maintenance(){

		}

		public void Update(float time){
			try{
				if(core.gender == "Male"){
					return;
				}
				ApplyBodyMorph();
				bodyMorphClosingTime = time + bodyMorphClosingTimeOffset;
			}catch(Exception e){
				core.LoggerE(e);
			}
		}
		
		public void SetMorphIfValid(string name, float strength){
			if(Bmorph != null && Bmorph.ContainsKey(name) && Bmorph[name] != null){
				Bmorph[name].SetValue(strength);
			}
		}
		
		public void ApplyBodyMorphByName(string name, float strength){
			try{
				if(!core.bellyMorphOn){ return; }
				if(!otherMorphsStatus.ContainsKey(name) || !otherMorphsStatus[name]){ return; }
				SetMorphIfValid(name, strength);
			}catch(Exception e){
				core.LoggerE(e);
			}
		}
		
		public void StopBodyMorphByName(string name){
			SetMorphIfValid(name, 0.000001f);
			SetMorphIfValid(name, 0f);
		}
		
		public void ApplyBodyMorph(){
			try{
				isBodyMorphActive = true;
				int stage = core.GetStage();
				float ratio = (stage > 5) ? 1.0f : stage * 0.2f;
				if(nippleErectionOnGlobal){
					SetMorphIfValid("Nipple", ratio);
				}
				if(clitorisErectionOnGlobal){
					SetMorphIfValid("Clitoris", ratio);
				}
				// Utility.FistHand(atom, ratio/2);
				core.FistHand(ratio/2);
				// SetMorphIfValid("Fingers", ratio/2);
			}catch(Exception e){
				core.LoggerE(e);
			}
		}
		
		public void StopBodyMorph(){
			try{
				if(core.gender == "Male"){
					return;
				}
				string[] names = {
					"Nipple", "Clitoris", "Anus", "Shivering1", "Shivering2", "ShiveringP", "Fingers",
				};
				foreach(var name in names){
					if(name == "Fingers"){
						// Utility.FistHand(atom, 0.000001f);
						// Utility.FistHand(atom, 0.0f);
						core.FistHand(0.000001f);
						core.FistHand(0.0f);
					}else{
						SetMorphIfValid(name, 0.000001f);
						SetMorphIfValid(name, 0.0f);
					}
				}
			}catch(Exception e){
				core.LoggerE(e);
			}
		}
		
		public void ShiveringBump(bool isBump){
			float strengthFingerHigh = UnityEngine.Random.Range(strengthFingerHighMin, strengthFingerHighMax);
			float strengthFingerLow  = UnityEngine.Random.Range(strengthFingerLowMin, strengthFingerLowMax);
			float strengthShiveringHigh = UnityEngine.Random.Range(strengthShiveringHighMin, strengthShiveringHighMax);
			float strengthShiveringLow  = UnityEngine.Random.Range(strengthShiveringLowMin, strengthShiveringLowMax);
			
			float morphStrengthBody   = (isBump) ? strengthShiveringHigh : strengthShiveringLow;
			float morphStrengthFinger = (isBump) ? strengthFingerHigh    : strengthFingerLow;

			ApplyBodyMorphByName("Shivering1", morphStrengthBody);
			ApplyBodyMorphByName("Shivering2", morphStrengthBody);
			ApplyBodyMorphByName("ShiveringP", morphStrengthBody);
			ApplyBodyMorphByName("Anus", morphStrengthBody);
			// Utility.FistHand(atom, morphStrengthFinger);
			core.FistHand(morphStrengthFinger);
			// ApplyBodyMorphByName("Fingers", morphStrengthFinger);
		}

		public void ShiveringRandom(){
			float strengthShivering = UnityEngine.Random.Range(strengthShiveringIntervalMin, strengthShiveringIntervalMax);

			ApplyBodyMorphByName("Shivering1", strengthShivering);
			ApplyBodyMorphByName("Shivering2", strengthShivering);
			ApplyBodyMorphByName("ShiveringP", strengthShivering);
			ApplyBodyMorphByName("Anus", strengthShivering);
		}

		public void Manage(float time){
			try{
				if(core.gender == "Male"){
					return;
				}
				if(isBodyMorphActive && bodyMorphClosingTime < time && core.isDuringOrgasm == false){
					core.Logger("bodyMorph effect stop.");
					isBodyMorphActive = false;
					StopBodyMorph();
				}
			}catch(Exception e){
				core.LoggerE(e);
			}
		}
		
		public void OnDestroy(){
			StopBodyMorph();
		}
	}
}