using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;

namespace Cotyounoyume{
    public struct Offset{
        public float Up;
        public float Forward;
        public float Right;

        public Offset(float Up, float Forward, float Right){
            this.Up = Up;
            this.Forward = Forward;
            this.Right = Right;
        }
    };

    public class SphereColliderMemory
    {
        public Dictionary<string, SphereCollider> SphereColliderDic = new Dictionary<string, SphereCollider>();
        public Dictionary<string, Vector3> LocalPositionDic = new Dictionary<string, Vector3>();
        public Dictionary<string, Quaternion> LocalRotationDic = new Dictionary<string, Quaternion>();
        public Dictionary<string, bool> TriggerNameMap = new Dictionary<string, bool>(){
            {"LabiaTrigger", true}, {"VaginaTrigger", true}, {"DeepVaginaTrigger", true}, {"DeeperVaginaTrigger", true}, 
        };
        public List<string> TriggerNames = new List<string>();

        public SphereColliderMemory(Atom atom){
            Init(atom);
        }

        public void Init(Atom atom){
			SphereCollider[] cols = atom.GetComponentsInChildren<SphereCollider>();
            foreach(var col in cols){
                if(!TriggerNameMap.ContainsKey(col.name)) continue;
                TriggerNames.Add(col.name);
                SphereColliderDic.Add(col.name, col);
                LocalPositionDic.Add(col.name, col.transform.localPosition);
                LocalRotationDic.Add(col.name, col.transform.localRotation);
            }
        }

        public void CheckSphereCollidersSkew(bool force = false){
            TriggerNames.ForEach(name => {
                var col = SphereColliderDic[name];
                var dist = Vector3.Distance(col.transform.localPosition, LocalPositionDic[name]);
                var dot  = Quaternion.Dot(col.transform.localRotation, LocalRotationDic[name]);
                if(force){
                    // Utility.Debug($"name:{name}, dist:{dist}, dot:{dot}");
                }

                if(dist > 0.001f){
                    RecoverOriginPositionAndRotation(name);
                    // Utility.Debug($"Recovery: name:{name}");
                }
            });
        }

        public void RecoverOriginPositionAndRotation(string name){
            if(!SphereColliderDic.ContainsKey(name)) return;
            var col = SphereColliderDic[name];
            col.transform.localPosition = LocalPositionDic[name];
            col.transform.localRotation = LocalRotationDic[name];
        }

    }

	public class BodyPartLocator
	{
        public EBATCore core;
        public Atom atom;
        public SphereColliderMemory scMemory;
        public Dictionary<string, Offset> BodyPartPositionOffset;
        public Dictionary<string, Quaternion> BodyPartRotation;
        public List<string> BodyPartNames = new List<string>()
        {
            "rNipple", "lNipple", "LabiaTrigger", "Anus", "lShin", "rShin",
        };
        public float fluctuation = 0.0f;
        public Dictionary<string, Rigidbody> RigidbodyDic = new Dictionary<string, Rigidbody>();

        public Dictionary<string, FreeControllerV3> ControllerDic = new Dictionary<string, FreeControllerV3>();
        public float tUp = 0.00f;
        public float tForward = 0.00f;
        public BodyPartLocator(EBATCore core)
        {
            this.core = core;
            this.atom = core.atom;
            BodyPartPositionOffset = new Dictionary<string, Offset>(){
                {"LabiaTriggerForHandCovering",     new Offset( 0.03f,  0.07f, 0f)},
                {"rNippleForHandCovering",          new Offset( 0.02f,  0.05f, -0.02f)},
                {"lNippleForHandCovering",          new Offset( 0.02f,  0.05f,  0.02f)},
                {"pelvis",           new Offset(-0.2f,  0.05f, 0f)},
                {"LabiaTrigger",     new Offset(-0.05f,  0.05f, 0f)},
                {"Anus",             new Offset( 0.00f, -0.05f, 0f)},
                {"rNipple",          new Offset( 0.02f,  0.05f, -0.02f)},
                {"lNipple",          new Offset( 0.02f,  0.05f,  0.02f)},
                {"lShin",            new Offset( 0.05f,  0.05f, 0f)},
                {"rShin",            new Offset( 0.05f,  0.05f, 0f)},
                {"HandL_from_chest", new Offset(-0.20f,  1.0f, -0.20f)},
                {"HandR_from_chest", new Offset(-0.20f,  1.0f,  0.20f)},
                {"HandLR",           new Offset(-0.01f, -0.01f, 0f)},
                {"HandLR1",          new Offset(-0.01f, -0.01f, 0f)},
                {"TwinDildoR",       new Offset(-0.028f, -0.005f, 0f)},
                {"TwinDildoV",       new Offset(0.3f, 0.3f, -0.6f)},
                // {"WMNZ",             new Offset(-0.02f, 0.00f, 0f)},
                {"WMNZ",             new Offset(-0.035f, -0.008f, 0f)},
                {"Tentacle1",        new Offset(0.08f, 0f, 0f)},
                {"HandL",            new Offset(-0.02f, -0.015f, 0f)},
                {"HandL2",           new Offset(-0.00f, -0.05f, 0f)},
                {"HandR",            new Offset(-0.035f, -0.005f, 0.0f)},
                {"Mouth1L",          new Offset(0.013f,  0.00f, 0.01f)},
                {"Mouth1R",          new Offset(0.013f,  0.01f, 0.005f)},
                {"Mouth1C",          new Offset(-0.032f, 0.020f, -0.015f)},
                {"Lip",              new Offset(0f, 0f, 0f)},
            };
            BodyPartRotation = new Dictionary<string, Quaternion>(){
                {"LabiaTriggerForHandCovering",     Quaternion.identity},
                {"rNippleForHandCovering",          Quaternion.identity},
                {"lNippleForHandCovering",          Quaternion.identity},
                {"pelvis",           Quaternion.identity},
                {"LabiaTrigger",     Quaternion.identity},
                {"Anus",             Quaternion.identity},
                {"rNipple",          Quaternion.identity},
                {"lNipple",          Quaternion.identity},
                {"lShin",            Quaternion.identity},
                {"rShin",            Quaternion.identity},
                {"HandL_from_chest", Quaternion.identity},
                {"HandR_from_chest", Quaternion.identity},
                {"HandLR",           Quaternion.identity},
                {"HandLR1",          Quaternion.identity},
                {"TwinDildoR",       Quaternion.identity},
                {"TwinDildoV",       Quaternion.identity},
                // {"WMNZ",             new Offset(-0.02f, 0.00f, 0f)},
                {"WMNZ",             Quaternion.identity},
                {"Tentacle1",        Quaternion.identity},
                {"HandL",            Quaternion.AngleAxis(-10f, Vector3.right)},
                {"HandL2",           Quaternion.AngleAxis(  0f, Vector3.right)},
                {"HandR",            Quaternion.identity},
                {"Mouth1L",          Quaternion.identity},
                {"Mouth1R",          Quaternion.identity},
                {"Mouth1C",          Quaternion.AngleAxis( 25f, Vector3.right) *Quaternion.AngleAxis( 90f, Vector3.up) * Quaternion.AngleAxis(60f, Vector3.forward)},
                {"Lip",              Quaternion.identity},
            };
            ControllerList.ForEach((name) => ControllerDic.Add(name, this.atom.GetStorableByID(name) as FreeControllerV3));
            RigidbodyList.ForEach((name) => RigidbodyDic.Add(name, Utility.FindRigidBody(this.atom, name)));
            scMemory = new SphereColliderMemory(atom);
        }

        public void CheckSphereCollidersSkew(bool force = false){
            scMemory.CheckSphereCollidersSkew(force);
        }

        public FreeControllerV3 GetController(string name)
        {
            if(!ControllerDic.ContainsKey(name)) return null;
            return ControllerDic[name];
        }

        public FreeControllerV3 GetRelativeController(string name)
        {
            string altName = "";
            if(name == "lHandControl") altName = "lElbowControl";
            else if(name == "rHandControl") altName = "rElbowControl";
            else if(name == "rKneeControl") altName = "lKneeControl";
            else if(name == "lKneeControl") altName = "rKneeControl";

            if(!ControllerDic.ContainsKey(altName)) return null;
            return ControllerDic[altName];
        }

        public Rigidbody GetRigidbody(string name)
        {
            if(!RigidbodyDic.ContainsKey(name)) return null;
            return RigidbodyDic[name];
        }

        public Offset GetOffset(string name)
        {
            if(!BodyPartPositionOffset.ContainsKey(name)) {
                return new Offset(0f, 0f, 0f);
            }
            return BodyPartPositionOffset[name];
        }

        public Vector3 GetPosition(string name, bool isBothHandsOn = false, string nameForCovering = "")
        {
            try{
                if(nameForCovering == "") nameForCovering = name;
                var bothHandsOffset = isBothHandsOn ? 0.05f : 0.0f;
                Offset offset = GetOffset(nameForCovering);
                // float up = offset.Up;
                float up = offset.Up + tUp;
                // float forward = offset.Forward;
                float forward = offset.Forward + tForward;
                float right = offset.Right;
                return Utility.GetPositionFromTargetBodynameWithOffsetAndRotationArrangement(atom, name, core.trunk.transform, up, forward + bothHandsOffset, right, fluctuation);
            }catch(Exception e){
                Utility.LoggerE(e);
                return new Vector3();
            }
        }

        public Quaternion GetRotation(string name){
            if(!BodyPartRotation.ContainsKey(name)) return Quaternion.identity;
            var rot = BodyPartRotation[name];
            return rot;
        }

        public void OnDestroy()
        {
            BodyPartPositionOffset.Clear();
            BodyPartRotation.Clear();
        }

		public void SetControllersAndRigidbodies(){
				
		}

        public List<string> ControllerList = new List<string>(){
            "control",
            "eyeTargetControl",
            "headControl",
            "neckControl",
            "chestControl",
            "rShoulderControl",
            "lShoulderControl",
            "rArmControl",
            "lArmControl",
            "rElbowControl",
            "lElbowControl",
            "rHandControl",
            "lHandControl",
            "abdomenControl",
            "abdomen2Control",
            "hipControl",
            "pelvisControl",
            "rThighControl",
            "lThighControl",
            "rKneeControl",
            "lKneeControl",
            "rFootControl",
            "lFootControl",
        };

        public List<string> RigidbodyList = new List<string>(){
            "head",
            "neck",
            "chest",
            "lCollar",
            "rCollar",
            "lShldr",
            "rShldr",
            "lPectoral",
            "rPectoral",
            "lNipple",
            "rNipple",
            "lForeArm",
            "rForeArm",
            "lHand",
            "rHand",
            "abdomen",
            "abdomen2",
            "hip",
            "pelvis",
            "lThigh",
            "rThigh",
            "lShin",
            "rShin",
            "lFoot",
            "rFoot",
            "LabiaTrigger",
        };
    }
}