#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
#endregion

namespace Cotyounoyume{
	public class ClothesManager
	{
		public ClothesStorage[] clothesStorages;
		public Atom atom;
		public DAZCharacterSelector geometryForClothes;
		
		public ClothesManager(){
			clothesStorages = new ClothesStorage[6];
			for(int i = 0; i < 6; i++){
				clothesStorages[i] = new ClothesStorage();
			}
		}
		
		public void Initialize(Atom atom){
			this.atom = atom;
			JSONStorable geometryStorable = this.atom.GetStorableByID("geometry");
			geometryForClothes = geometryStorable as DAZCharacterSelector;
		}
		
		public void EquipClothes(int stage){
			try{
				if(stage < 0 || stage > 5){
					return;
				}
				for(int i = 0; i < clothesStorages[stage].clothes.Count; i++){
					geometryForClothes.SetActiveClothingItem(clothesStorages[stage].clothes[i], active: true);
				}
			}catch(Exception e){
				LoggerE(e);
			}
		}
		
		public void UnequipClothes(int stage){
			try{
				if(stage < 0 || stage > 5){
					return;
				}
				for(int i = 0; i < clothesStorages[stage].clothes.Count; i++){
					geometryForClothes.SetActiveClothingItem(clothesStorages[stage].clothes[i], active: false);
				}				
			}catch(Exception e){
				LoggerE(e);
			}
		}
		
		public List<DAZClothingItem> GetClothes(){
			List<DAZClothingItem> clothes = geometryForClothes.clothingItems.ToList();
			List<DAZClothingItem> myClothes = new List<DAZClothingItem>();
			foreach (DAZClothingItem dAZClothingItem in clothes){
				if(dAZClothingItem.active){
					myClothes.Add(dAZClothingItem);
				}
			}
			
			return myClothes;
		}
		
		public List<string> GetClothUids(int stage){
			List<string> myClothes = Clothes2uids(clothesStorages[stage].clothes);			
			return myClothes;
		}
		
		public void SetClothes(int stage){
			try{
				List<DAZClothingItem> tmpClohtes = GetClothes();
				clothesStorages[stage].SetClothes(tmpClohtes);				
			}catch(Exception e){
				LoggerE(e);
			}
		}
		
		public void ShowClothes(int stage){
			try{
				clothesStorages[stage].ShowClothes();
			}catch(Exception e){
				LoggerE(e);
			}
		}
		
		public void SetClothesByUids(int stage, List<string> uids){
			try{
				List<DAZClothingItem> tmpClohtes = this.Uids2clothes(uids);
				clothesStorages[stage].SetClothes(tmpClohtes);
			}catch(Exception e){
				LoggerE(e);
			}
		}
		
		public void ClearClothes(int stage){
			clothesStorages[stage].ClearClothes();
		}
		
		public void UndressAll(){
			geometryForClothes.RemoveAllClothing();
		}
		
		public List<string> Clothes2uids(List<DAZClothingItem> clothes){
			int num = clothes.Count;
			List<string> uids = new List<string>();
			for(int i = 0; i < num; i++){
				uids.Add(clothes[i].uid);
			}
			return uids;
		}
		
		public List<DAZClothingItem> Uids2clothes(List<string> uids){
			Dictionary<string, bool> uidsList;
			uidsList = new Dictionary<string, bool>();			
			for(int i = 0; i < uids.Count; i++){
				uidsList[uids[i]] = true;
			}
			
			List<DAZClothingItem> clothes = new List<DAZClothingItem>();

			List<DAZClothingItem> clothesAll = geometryForClothes.clothingItems.ToList();
			for(int i = 0; i < clothesAll.Count; i++){
				if(uidsList.ContainsKey(clothesAll[i].uid)){
					clothes.Add(clothesAll[i]);
				}
			}
			
			return clothes;
		}
		
		public void Debug(string str){
			SuperController.LogMessage("debug: " + str);
		}
		
		public void LoggerE(Exception e){
			SuperController.LogMessage("*** Exception:" + e);
		}
		
		public void OnDestroy(){
			RemoveClothesStorage();
		}
		
		public void RemoveClothesStorage(){
			for(int i = 0; i < 6; i++){
				clothesStorages[i] = null;
			}
			clothesStorages = null;			
		}
	}
	
	public class ClothesStorage
	{
		public List<DAZClothingItem> clothes;
		
		public ClothesStorage(){
			clothes = new List<DAZClothingItem>();
		}
		
		public void SetCurrentClothes(Atom atom){
			JSONStorable geometryStorable = atom.GetStorableByID("geometry");
			DAZCharacterSelector geometryForClothes = geometryStorable as DAZCharacterSelector;
			List<DAZClothingItem> clothes = geometryForClothes.clothingItems.ToList();
			for (int i = 0; i < clothes.Count; i++){
				DAZClothingItem c = clothes[i];
				if(c.active){
					
				}
			}
		}
		
		public void SetClothes(List<DAZClothingItem> clothes){
			this.clothes = new List<DAZClothingItem>();
			for(int i = 0; i < clothes.Count; i++){
				this.clothes.Add(clothes[i]);
			}
		}
		
		public void ShowClothes(){
			if(this.clothes.Count == 0){
				SuperController.LogMessage("  clothes not found.");				
			}
			for(int i = 0; i < this.clothes.Count; i++){
				SuperController.LogMessage("  Name: " + this.clothes[i].displayName);
//				SuperController.LogMessage("  Uid : " + this.clothes[i].uid);
			}
		}
		
		public void ClearClothes(){
			this.clothes = new List<DAZClothingItem>();
		}
		
		public void Debug(string str){
			SuperController.LogMessage("debug: " + str);
		}
		
		public void LoggerE(Exception e){
			SuperController.LogMessage("*** Exception:" + e);
		}
	}
}