#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
#endregion

namespace Cotyounoyume{
	public class SnapBackupMorphValue{
		public Atom atom;
		public GenerateDAZMorphsControlUI morphControl;
		public List<DAZMorph> backups;
		public List<float> values;
		public bool backupDone = false;
		
		public SnapBackupMorphValue(){
			
		}
		
		public void Backup(List<DAZMorph> morphs){
			Delete();

			backups = new List<DAZMorph>();
			values = new List<float>();
			
			for(int i = 0; i < morphs.Count; i++){
				backups.Add(morphs[i]);
				values.Add(morphs[i].morphValue);
			}
			backupDone = true;
		}
		
		public void Delete(){
			if(backups == null){
				return;
			}
			if(backups.Count > 0){
				backups.Clear();
				backups = null;
				values.Clear();
				values = null;
			}			
		}
		
		public void Recover(){
			if(!backupDone) return;
			
			for(int i = 0; i < backups.Count; i++){
				backups[i].SetValue(values[i] + 0.001f);
			}
		}
	}
	
	public class SnapBackup{
		public Atom atom;
		public string[] keys;
		public List<ControllerBackup> backups;
		public List<FreeControllerV3> ctrls;
		public List<Rigidbody> rbs;
		public bool isRigidbody = false;
		public bool backupDone = false;
		
		public SnapBackup(Atom atom, string[] keys, bool isRigidbody = false){
			this.atom = atom;
			this.isRigidbody = isRigidbody;
			Delete();
			BackupTemp(keys);
			if(!isRigidbody){
				AllJointOn();
				RecoverStatus();
				Delete();
				BackupTemp(keys);
			}
			backupDone = true;
		}
		
		public void BackupTemp(string[] keys){
			this.keys = keys;
			backups = new List<ControllerBackup>();
			ctrls = new List<FreeControllerV3>();
			rbs = new List<Rigidbody>();
			
			if(isRigidbody){
				for(int i = 0; i < keys.Length; i++){
					Rigidbody rb = Utility.FindRigidBody(atom, keys[i]);
					rbs.Add(rb);
					backups.Add(new ControllerBackup());
					backups[i].BackUp(rb);
				}
			}else{
				for(int i = 0; i < keys.Length; i++){
					FreeControllerV3 fc = null;
					try{
					fc = Utility.GetFreeControllerV3(atom, keys[i]);
					}catch(Exception e){
						Utility.LoggerE(e);
					}
					ctrls.Add(fc);
					backups.Add(new ControllerBackup());
					backups[i].BackUp(fc);
				}
			}
		}
					
		public void AllJointOn(){
			if(!backupDone) return;

			for(int i = 0; i < keys.Length; i++){
				FreeControllerV3 fc = Utility.GetFreeControllerV3(atom, keys[i]);
				Utility.SwitchPositionState(atom, keys, "On");
				Utility.SwitchRotationState(atom, keys, "On");
			}
		}
		
		public void Delete(){
			if(backups != null){
				if(backups.Count > 0){
					backups.Clear();
					ctrls.Clear();
					rbs.Clear();
					keys = null;
				}			
			}
		}
		
		public void Recover(){
			AllJointOn();
			RecoverPosition();
			RecoverRotation();
			RecoverStatus();
		}
		
		public void RecoverTemp(){
			if(!backupDone) return;

			for(int i = 0; i < keys.Length; i++){
				if(isRigidbody){
					rbs[i] = backups[i].Recovery(rbs[i]);
				}else{
					ctrls[i] = backups[i].Recovery(ctrls[i]);
				}
			}
		}
		
		public void RecoverPosition(){
			if(!backupDone) return;

			for(int i = 0; i < keys.Length; i++){
				ctrls[i] = backups[i].RecoveryPosition(ctrls[i]);
			}
		}
		
		public void RecoverRotation(){
			if(!backupDone) return;

			for(int i = 0; i < keys.Length; i++){
				ctrls[i] = backups[i].RecoveryRotation(ctrls[i]);
			}			
		}
		
		public void RecoverStatus(){
			if(!backupDone) return;

			for(int i = 0; i < keys.Length; i++){
				ctrls[i] = backups[i].RecoveryStatus(ctrls[i]);
			}						
		}
	}
	
	
	
	public class ControllerBackup{
		public Vector3 v3pos;
		public Quaternion v3rot;
		public string positionState;
		public string rotationState;
		
		public ControllerBackup(){
			v3pos = new Vector3(0f, 0f, 0f);
			v3rot = new Quaternion(0f, 0f, 0f, 0f);
			positionState = "Off";
			rotationState = "Off";
		}
		
		public void BackUp(FreeControllerV3 target){
			try{
				v3pos = Utility.GetPosition(target);
				v3rot = Utility.GetRotation(target);
				positionState = target.currentPositionState.ToString();
				rotationState = target.currentRotationState.ToString();
			}catch(Exception e){
				SuperController.LogMessage("error in ControllerBackup.backUp:" + e);
			}
		}
		
		public void BackUp(Rigidbody target){
			try{
				v3pos = Utility.GetPosition(target.position);
				v3rot = Utility.GetRotation(target.rotation);
			}catch(Exception e){
				SuperController.LogMessage("error in ControllerBackup.backUp:" + e);
			}
		}
		
		public FreeControllerV3 Recovery(FreeControllerV3 target){
			try{
				target.SetPositionStateFromString(positionState);
				target.SetRotationStateFromString(rotationState);
				target.transform.SetPositionAndRotation(v3pos, v3rot);
				return target;
			}catch(Exception e){
				SuperController.LogMessage("error in ControllerBackup.recovery:" + e);
				return null;
			}
		}

		public Rigidbody Recovery(Rigidbody target){
			try{
				target.position = v3pos;
				target.rotation = v3rot;
				return target;
			}catch(Exception e){
				SuperController.LogMessage("error in ControllerBackup.recovery:" + e);
				return null;
			}
		}

		public FreeControllerV3 RecoveryStatus(FreeControllerV3 target){
			try{
				target.SetPositionStateFromString(positionState);
				target.SetRotationStateFromString(rotationState);
				return target;
			}catch(Exception e){
				SuperController.LogMessage("error in ControllerBackup.recovery:" + e);
				return null;
			}
		}

		public FreeControllerV3 RecoveryPosition(FreeControllerV3 target){
			try{
				Quaternion tarRot = Utility.GetRotation(target);
				target.transform.SetPositionAndRotation(v3pos, tarRot);
				target.SetPositionStateFromString(positionState);
				return target;
			}catch(Exception e){
				SuperController.LogMessage("error in ControllerBackup.recovery:" + e);
				return null;
			}
		}

		public FreeControllerV3 RecoveryRotation(FreeControllerV3 target){
			try{
				Vector3 tarPos = Utility.GetPosition(target);
				target.transform.SetPositionAndRotation(tarPos, v3rot);
				target.SetRotationStateFromString(rotationState);
				return target;
			}catch(Exception e){
				SuperController.LogMessage("error in ControllerBackup.recovery:" + e);
				return null;
			}
		}
	}
}