#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
#endregion

namespace Cotyounoyume{
	public class Expression2
	{
		public Atom atom;
		public EBATCore core;
		public FaceMorphManager fmm;
		public bool expressionOnGlobal = true;
		public bool isExpressionActive = false;
		public bool inDuringSpecialTest = false;
		public List<string> specialExpressions = new List<string>();
		public Dictionary<string, bool> dSpecialExpressions = new Dictionary<string, bool>();
		public Dictionary<string, Dictionary<string, bool>> dLvlVsExprTable; 
		public Dictionary<string, float> dLvlVsRatioTable = new Dictionary<string, float>();
		public Dictionary<string, string> dEyesDirection = new Dictionary<string, string>();
		public string[] LvList = {"Lv1", "Lv2", "Lv3", "Lv4", "Lv5", "Lv6", "Lv7", };
		public string[] directionList = {"Neutral", "Up", "Down", "Right", "Left"};
		public Dictionary<string, List<string>> LevelVsExpression;

		private string[] _defaultSpecial = 
		{
			"CTY_Special_01Alt", "CTY_Special_02","CTY_Special_03",
		};
		private Dictionary<string, string> _defaultDirection = new Dictionary<string, string>()
		{
			{"CTY_Special_01Alt" , "Up"},
			{"CTY_Special_02" , "Up"},
			{"CTY_Special_03" , "Down"},
		};
		private Dictionary<string, float> _defaultRatio = new Dictionary<string, float>()
		{
			{"Lv1",   0f}, {"Lv2",   0f},
			{"Lv3", 0.1f}, {"Lv4", 0.1f},
			{"Lv5", 0.2f}, {"Lv6", 0.3f}, {"Lv7", 0.5f},
		};

		public Expression2(EBATCore core){
			this.atom = core.atom;
			this.core = core;
			
			fmm = new FaceMorphManager(this.atom);
		}

		public void LoadSettings(){		
			try{
				expressionOnGlobal = core.settings["General"]["Expression On"].AsBool;
				fmm.baseMultiplier = core.settings["General"]["Expression Strength"].AsFloat;
				LoadSpecialExpression();
				MakeSPLevelTable();
				MakeEyesDirectionTable();
				MakeSPRatioTable();
			}catch (Exception e){
				core.LoggerE(e);
			}
		}
		
		public void LoadSpecialExpression(){
			specialExpressions = new List<string>();
			dSpecialExpressions = new Dictionary<string, bool>();
			if (!core.settingsRoot["Common"].AsObject.HasKey("SpecialExpression")) { SetDefaultValue(); }

			JSONArray jsa = core.settingsRoot["Common"]["SpecialExpression"]["Entry"].AsArray;

			foreach(string key in jsa.Childs){
				dSpecialExpressions.Add(key, true);
				specialExpressions.Add(key);
			}
			specialExpressions.Sort();
		}

		public void SetDefaultValue()
		{
			foreach(var expr in _defaultSpecial)
			{
				core.settingsRoot["Common"]["SpecialExpression"]["Entry"].Add(expr);
			}

			foreach(var lv in LvList)
			{
				foreach(var expr in _defaultSpecial)
				{
					bool flag = !(lv == "Lv1" || lv == "Lv2");
					core.customUi.SetValueToSettings(flag.ToString(), "Common", "SpecialExpression", "LevelVsExpression", lv, expr);
				}
			}

			foreach(var expr in _defaultDirection.Keys.ToList<string>())
			{
				core.customUi.SetValueToSettings(_defaultDirection[expr], "Common", "SpecialExpression", "EyesDirection", expr);
			}

			foreach (var lv in _defaultRatio.Keys.ToList<string>())
			{
				core.customUi.SetValueToSettings(_defaultRatio[lv].ToString(), "Common", "SpecialExpression", "LevelVsRatio", lv);
			}
		}


		public void MakeSPLevelTable(){
			dLvlVsExprTable = new Dictionary<string, Dictionary<string, bool>>();
			List<string> exps = GetSpecialExpressions();
			bool isFirst = (core.settingsRoot["Common"]["SpecialExpression"]["LevelVsExpression"] == null);
			JSONClass jsc = isFirst ? null : core.settingsRoot["Common"]["SpecialExpression"]["LevelVsExpression"].AsObject;
			
			foreach(var lvl in LvList){
				dLvlVsExprTable.Add(lvl, new Dictionary<string, bool>());
				foreach(var exp in exps){
					bool isOn = isFirst ? false : jsc[lvl][exp].AsBool;
					if(isFirst){
						core.customUi.SetValueToSettings(isOn.ToString(), "Common", "SpecialExpression", "LevelVsExpression", lvl, exp);
					}
					dLvlVsExprTable[lvl].Add(exp, isOn);
				}
			}
			MakeLevelVsExpression();			
		}
		
		public void UpdateSPLevelTable(){
			JSONClass jsc = core.settingsRoot["Common"]["SpecialExpression"]["LevelVsExpression"][LvList[0]].AsObject;
			List<string> expsConfig = jsc.Keys.ToList();
			jsc = core.settingsRoot["Common"]["SpecialExpression"]["LevelVsExpression"].AsObject;
			List<string> expsMemory = GetSpecialExpressions();
			foreach(var name in expsConfig){
				if(!expsMemory.Contains(name)){
					foreach(var lvl in LvList){
						if(dLvlVsExprTable[lvl].ContainsKey(name)){
							dLvlVsExprTable[lvl].Remove(name);
							jsc[lvl].Remove(name);
						}
					}
				}
			}
			foreach(var name in expsMemory){
				if(!expsConfig.Contains(name)){
					foreach(var lvl in LvList){
						if(!dLvlVsExprTable[lvl].ContainsKey(name)){
							dLvlVsExprTable[lvl].Add(name, false);
							jsc[lvl][name] = "False";
						}
					}
				}
			}
			MakeLevelVsExpression();
		}
		
		public void MakeLevelVsExpression(){
			LevelVsExpression = new Dictionary<string, List<string>>();
			foreach(var level in LvList){
				List<string> expressions = new List<string>(dLvlVsExprTable[level].Keys);
				foreach(var expression in expressions){
					if(dLvlVsExprTable[level][expression]){
						if(!LevelVsExpression.ContainsKey(level)){ LevelVsExpression.Add(level, new List<string>()); }
						LevelVsExpression[level].Add(expression);
					}
				}
			}
			LevelVsExpression.Add("All", GetSpecialExpressions());
			fmm.UpdateLevelVsExpressions(LevelVsExpression);
		}
				
		public void MakeSPRatioTable(){
			bool isFirst = (core.settingsRoot["Common"]["SpecialExpression"]["LevelVsRatio"] == null);
			JSONClass jsc = isFirst ? null : core.settingsRoot["Common"]["SpecialExpression"]["LevelVsRatio"].AsObject;
			dLvlVsRatioTable = new Dictionary<string, float>();
			foreach(var lvl in LvList){
				float value = isFirst ? 0f : jsc[lvl].AsFloat;
				dLvlVsRatioTable.Add(lvl, value);
				if(isFirst){
					core.customUi.SetValueToSettings("0", "Common", "SpecialExpression", "LevelVsRatio", lvl);
				}
			}
			fmm.UpdateLevelVsFrequency(dLvlVsRatioTable);			
		}
		
		public void AddSpecialExpression(string name){
			if(!dSpecialExpressions.ContainsKey(name)){
				core.customUi.SetValueToSettings("True", "Common", "SpecialExpression", "Entry", name);
				dSpecialExpressions.Add(name, true);
				specialExpressions.Add(name);
				specialExpressions.Sort();
			}
		}
		
		public void RemoveSpecialExpression(string name){
			if(dSpecialExpressions.ContainsKey(name)){
				core.customUi.UnsetValueToSettings(name, "Common", "SpecialExpression", "Entry", "");
				dSpecialExpressions.Remove(name);
				specialExpressions.Remove(name);
			}
		}
		
		public void MakeEyesDirectionTable(){
			dEyesDirection = new Dictionary<string, string>();
			List<string> exps = GetSpecialExpressions();
			bool isFirst = (core.settingsRoot["Common"]["SpecialExpression"]["EyesDirection"] == null);
			JSONClass jsc = isFirst ? null : core.settingsRoot["Common"]["SpecialExpression"]["EyesDirection"].AsObject;
			
			foreach(var exp in exps){
				string direction = isFirst ? "Random" : (jsc[exp].Value);
				core.customUi.SetValueToSettings(direction, "Common", "SpecialExpression", "EyesDirection", exp);
				dEyesDirection.Add(exp, direction);
			}
			fmm.UpdateEyesDirection(dEyesDirection);
		}
		
		public void AddEyesDirection(string name, string direction)
		{
			if(!dEyesDirection.ContainsKey(name))
			{
				core.customUi.SetValueToSettings(direction, "Common", "SpecialExpression", "EyesDirection", name);
				dEyesDirection.Add(name, direction);
			}
		}
		
		public void RemoveEyesDirection(string name){
			if(dEyesDirection.ContainsKey(name))
			{
				core.customUi.UnsetValueToSettings(name, "Common", "SpecialExpression", "EyesDirection", "");
				dEyesDirection.Remove(name);
			}
		}
		
		public void StartTest()
		{
			inDuringSpecialTest = true;
			core.RefreshEBATVisual();
			core.DoDelayedAction(FinishTest, 3.0f);
		}

		public void FinishTest()
		{
			inDuringSpecialTest = false;
			core.RefreshEBATVisual();
		}

		
		public bool IsUsingSpecial()
		{
			if(IsFmmWrong()) return false;
			
			return fmm.isUsingSpecial;
		}
		
		public string GetCurrentEyesDirection()
		{
			string direction = "Random";
			if(!IsUsingSpecial()){ return direction; }
			string currentMorph = fmm.GetCurrentSpecialExpression();
			if(dEyesDirection.ContainsKey(currentMorph)){
				direction = dEyesDirection[currentMorph];
			}

			return (direction == "Random") ? GetRandomDirection() : direction;
		}

		public string GetRandomDirection()
		{
			int num = UnityEngine.Random.Range(0, directionList.Length);
			return directionList[num];
		}
		
		public string GetEyesDirection(string name)
		{
			if(dEyesDirection.ContainsKey(name))
			{
				return dEyesDirection[name];
			}
			else
			{
				return "";
			}
		}
		
		public List<string> GetSpecialExpressions(){
			return specialExpressions;
		}
		
		public void ClearContinuityCount()
		{
			if(IsFmmWrong()) return;
			
			fmm.ClearContinuityCount();
		}
		
		public void TestSpecialExpression(string name){
			if(IsFmmWrong()) return;
			
			fmm.ForceEnableSpecial();
			fmm.PickExpressionByMorphName(
				core.currentTime, "special", name, 
				5, false, true, "Lv7");
			core.DoDelayedAction(core.eyesControl.Update, 0.1f);
		}

		public void TestExpressionByLevel(string stageLevelLocal, string name = ""){
			if(IsFmmWrong()) return;

			fmm.PickExpression(core.currentTime, 5, false, isStrongWave:true, stageLevel:stageLevelLocal);
			core.DoDelayedAction(core.eyesControl.Update, 0.1f);
		}

		public bool IsFmmWrong()
		{
			if(fmm != null) return false;

			// Utility.Debug("UnknownError: fmm is null: remake fmm.");
			// new FaceMorphManager(this.atom);
			return true;
		}

		public void Update(){
			try{
				if(!expressionOnGlobal){return;}
				if(IsFmmWrong()) return;

				fmm.PickExpression(core.currentTime, core.GetStage(), core.isDuringAfterglow, isStrongWave:core.isFirstTerm, stageLevel:core.stageLevel);
				core.DoDelayedAction(core.eyesControl.Update, 0.1f);
			}catch (Exception e){
				core.LoggerE(e);
			}
		}
		
		public void Manage(){
			try{
				if(!expressionOnGlobal){return;}
				if(IsFmmWrong()) return;

				this.isExpressionActive = fmm.isExpressionActive;
				fmm.ManageStrength(core.currentTime);
			}catch (Exception e){
				core.LoggerE(e);
			}
		}
		
		public void OpenMouth(float duration){
			if(IsFmmWrong()) return;
			fmm.OpenMouth(core.currentTime, duration, core.curStage, core.isDuringAfterglow, core.isFirstTerm, core.stageHeat);
		}

		public void StopExpression(){
			if(IsFmmWrong()) return;
			fmm.StopExpression();
		}

		public void ResetExpression(){
			try{
				if(IsFmmWrong()) return;

				fmm.StopExpression();
				fmm.StopAllExpressionInPose();
			}catch(Exception e){
				Utility.LoggerE(e);
			}
		}
		
		public void StopAllExpressionInPose(){
			if(IsFmmWrong()) return;

			fmm.StopAllExpressionInPose();
		}
		
		public void OnDestroy(){
			if(IsFmmWrong()) return;

			fmm.OnDestroy();
			fmm = null;
		}
	}
}