using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;

namespace Cotyounoyume{
	public class ExternalLinkage
	{
        EBATCore core;
        public ColortoneSub colortoneSub;
        public PostMagicSub postMagicSub;
        public VAMMoanSub   vamMoanSub;
        public BrowserAssistSub browserAssistSub;

        public bool InDLing_Colortone = false;
        public bool InDLing_PostMagic = false;
        public bool InDLing_VAMMoan = false;
        public bool InDLing_BrowserAssist = false;


        public ExternalLinkage(EBATCore core){
            this.core = core;
            colortoneSub = new ColortoneSub(this.core);
            postMagicSub = new PostMagicSub(this.core);
            vamMoanSub   = new VAMMoanSub(this.core);
            browserAssistSub = new BrowserAssistSub(this.core);
        }

        public void OnLoadDone(){
            postMagicSub.OnLoadDone();
            vamMoanSub.OnLoadDone();
            core.worldUi.ClearReloadStatus();
        }

        public bool IsReloadNeeded(string type){
            if(type == "PostMagic"){
                return postMagicSub.IsReloadNeeded();
            }else if(type == "VAMMoan"){
                return vamMoanSub.IsReloadNeeded();
            }else if(type == "BrowserAssist"){
                return browserAssistSub.IsReloadNeeded();
            }else{
                return false;
            }
        }

		public void SetColorCorrect(bool isOn){ colortoneSub.SetColorCorrect(isOn); }
        public bool IsColortoneOn(){ return colortoneSub.IsColortoneOn(); }
        public bool ExistsColortoneInTheGame(){ return colortoneSub.ExistsColortoneInTheGame(); }

        public bool IsPostMagicOn(bool isDoFOn){ return postMagicSub.IsPostMagicOn(isDoFOn); }
		public void SetPostMagic(bool isOn){ postMagicSub.SetResource(isOn); }
            
        public bool ExistPostMagicInTheGame(){ return postMagicSub.ExistResourceInTheGame(); }
        public bool IsDoFOn(){ return postMagicSub.IsDoFOn();}
        public void DoFOnOff(bool isOn){ postMagicSub.DoFOnOff(isOn); }
    
        public bool IsVAMMoanOn(){ return vamMoanSub.IsVAMMoanOn(); }
            
		public void SetVAMMoan(bool isOn){ vamMoanSub.SetResource(isOn); }
            
        public bool ExistVAMMoanInTheGame(){ return vamMoanSub.ExistResourceInTheGame(); }

        public bool IsBrowserAssistOn() { return browserAssistSub.ExistResourceInTheScene(); }
		public void SetBrowserAssist(bool isOn){ browserAssistSub.SetResource(isOn); }
        public void LaunchBrowserAssist() { browserAssistSub.LaunchBrowserAssist(); }
    }

    public class VAMMoanSub: SubBase{
        public VAMMoanSub(EBATCore core) : base(core){
            base.Init(
                "VAMMoan", "VAMMoanPlugin.VAMMoan", "Custom\\Scripts\\VAMMoan\\VAMMoan.cslist", "*vammoan*", 97, 
                "hazmhox.vammoan.22.var", "3134", "Person", 
                OnOffActionImpl, CheckResourceInTheSceneFuncImpl, OnLoadActionImpl
            );
        }

        public void OnOffActionImpl(bool isOn){
            Utility.ToggleOnOffPlugin(core.atom, ShortName, isOn);
        }

        public bool CheckResourceInTheSceneFuncImpl(){
            return Utility.GetAtomWithPluginNamePattern(ShortName, core.atom) != null;
        }

        public void OnLoadActionImpl(){
            core.externalLinkage.InDLing_VAMMoan = false;
        }

        public bool IsVAMMoanOn(){
            return Utility.GetOnOffPlugin(ShortName);
        }
    }

    public class PostMagicSub : SubBase{
        public string PostMagicLabelName = "MacGruber_PostMagic";
        public string PostMagicPrefix = "MacGruber.PostMagic.";
        public string PostMagicManagerName = "MacGruber.PostMagic.Manager";
        public string PostMagicScriptPath = "Custom\\Scripts\\MacGruber\\PostMagic\\MacGruber_PostMagic.cslist"; 
        public string PostMagicSearchPattern = "*PostMagic*";
        public string DoFName = "MacGruber.PostMagic.DepthOfField";
        public bool DoFOn = false;

        public PostMagicSub(EBATCore core) : base(core){
            base.Init(
                "PostMagic", "MacGruber.PostMagic", "Custom\\Scripts\\MacGruber\\PostMagic\\MacGruber_PostMagic.cslist", "*PostMagic*", 96, 
                "MacGruber.PostMagic.3.var", "161", "Person", 
                OnOffActionImpl, CheckResourceInTheSceneFuncImpl, OnLoadActionImpl
            );
        }

        public void OnOffActionImpl(bool isOn){
            Utility.ToggleOnOffPlugin(PostMagicManagerName, isOn);
            Utility.ToggleOnOffPlugin(DoFName, DoFOn);
        }

        public bool CheckResourceInTheSceneFuncImpl(){
            return Utility.GetAtomWithPluginNamePattern(PostMagicManagerName) != null;
        }

        public void OnLoadActionImpl(){
            var result = Utility.HasThisPlugin(core.atom, PostMagicLabelName);

            if(Utility.HasThisPlugin(core.atom, PostMagicLabelName)){
                if(CheckPostMagicLoadStatusIsDefault()) ChangePostMagicLoadStatus();
            }
            core.externalLinkage.InDLing_PostMagic = false;
        }

        public Dictionary<string, bool> PostMagicOnMap = new Dictionary<string, bool>(){
            // {"MacGruber.PostMagic.Manager", false},
            {"MacGruber.PostMagic.AntiAliasing", false}, 
            {"MacGruber.PostMagic.Bloom", false},
            {"MacGruber.PostMagic.Grain", false},
            {"MacGruber.PostMagic.ChromaticAberration", true},
            {"MacGruber.PostMagic.DepthOfField", true},
            {"MacGruber.PostMagic.MotionBlur", true},
            {"MacGruber.PostMagic.UserLUT", true},
            {"MacGruber.PostMagic.Vignette", true},
        };

        public bool CheckPostMagicLoadStatusIsDefault(){
            var list = Utility.GetPluginsByPattern(core.atom, PostMagicPrefix);
            var totalResult = false;
            foreach(var name in list){
                if(name.Contains(PostMagicManagerName)) continue;
                if(name.Contains(DoFName)) continue;
                var result = Utility.GetOnOffPlugin(core.atom, name);
                totalResult = totalResult | result;
                var normalized = Utility.NormalizePluginName(name);
            }
            totalResult = !totalResult;

            return totalResult;
        }

        public void ChangePostMagicLoadStatus(){
            var list = Utility.GetPluginsByPattern(core.atom, PostMagicPrefix);
            foreach(var name in list){
                var normalized = Utility.NormalizePluginName(name);
                if(normalized == PostMagicManagerName) continue;
                if(!PostMagicOnMap.ContainsKey(normalized)) continue;
                var isOn = PostMagicOnMap[normalized];
                if(normalized == DoFName){
                    isOn = DoFOn;
                }
                Utility.ToggleOnOffPlugin(core.atom, name, isOn);
            }
        }

        public bool IsPostMagicOn(bool isDoFOn){
            return DoFOn == isDoFOn && Utility.GetOnOffPlugin(PostMagicManagerName);
        }

        public void DoFOnOff(bool isOn){
            DoFOn = isOn;
        }

        public bool IsDoFOn(){
            return DoFOn;
        }
    }

    public class BrowserAssistSub : SubBase{
        public string browserAssistScriptPathFree = "Custom\\Scripts\\JayJayWon\\BrowserAssist\\BrowserAssist(Free).cslist";
        public string browserAssistScriptPathFull = "Custom\\Scripts\\JayJayWon\\BrowserAssist\\BrowserAssist.cslist";
        public string browserAssistScriptPath     = "";
        public string browserAssistSearchPatternFree = "*BrowserAssist*";
        public string browserAssistSearchPatternFull = "*BrowserAssist.*";
        public string browserAssistSearchPattern = "";
        public UIDynamicButton launchButton;

        public BrowserAssistSub(EBATCore core) : base(core){
            SetVariables();
            base.Init(
                "BrowserAssist", "JayJayWon.BrowserAssist", browserAssistScriptPath, browserAssistSearchPattern, 98, 
                "JayJayWon.BrowserAssist(Free).19.var", "20134", "Session",
                OnOffActionImpl, CheckResourceInTheSceneFuncImpl, OnLoadActionImpl
            );
        }

        public void OnOffActionImpl(bool isOn){
            if(!isOn){
                Utility.RemoveButtonClickOnSessionPlugin("BrowserAssist");
            }
        }

        public void OnLoadActionImpl(){

        }

        public bool CheckResourceInTheSceneFuncImpl(){
            return Utility.HasThisPlugin(atom:null, pattern:LabelName);
        }

        public void SetVariables(){
            bool isFull = HasFullVersion();
            browserAssistScriptPath = isFull ? browserAssistScriptPathFull : browserAssistScriptPathFree;
            browserAssistSearchPattern = isFull ? browserAssistSearchPatternFull : browserAssistSearchPatternFree;
        }

        public bool HasFullVersion(){
            return Utility.GetLongestFilename(browserAssistSearchPatternFull) != "";
        }

        public void GetLaunchButton(){
            var buttons = SuperController.singleton.mainHUD.GetComponentsInChildren<UIDynamicButton>();
            foreach(var button in buttons){
                var name = button.buttonImage?.mainTexture?.name?.ToLower() ?? "";
                if(!name.Contains("browserassist") && !name.Contains("uisprite")) continue;
                launchButton = button;
                break;
            }
        }

        public void LaunchBrowserAssist(){
            GetLaunchButton();
            if(launchButton == null) return;
            launchButton.button.onClick.Invoke();
        }
    }


    public class ColortoneSub{
        public string ColortoneCUAName = "EBAT_Colortone";
        public string ColortoneAssetNamePath = "Assets/Scenes/TonemapOnly.unity";
        public string ColortoneAssetNameOnly = "TonemapOnly.unity";
        public string ColortoneAssetURLPath = "Custom/Assets/colorcorrect.assetbundle"; 
        public string ColortoneAssetFilename = "colorcorrect.assetbundle"; 
        public string ColortoneSearchPattern = "*TiSeb.Colortone*";
        public EBATCore core;
        public ColortoneSub(EBATCore core){
            this.core = core;
        }

        public bool IsColortoneOn(){
			var cua = GetColoCorrectAtom();
            if(cua == null) return false;
            return Utility.HasAssetName(cua, ColortoneAssetNameOnly);
        }

		public string GetColorcorrectURL(){
            var filename = Utility.GetLongestFilename(ColortoneSearchPattern);
            if(filename == "") return "";
            var varname = Utility.CorrectVarname(filename);
			return varname + ":/" + ColortoneAssetURLPath;
		}

		public void SetColorCorrect(bool isOn){
			if(!isOn) {
                SetColorCorrectOnOff(false);
            }else{
                if(ExistsColortoneInTheScene()){
                    SetColorCorrectOnOff(true);
                }else if(ExistsColortoneInTheGame()){
                    MakeCUAForColortone();
                }else{
                    core.externalLinkage.InDLing_Colortone = true;
                    core.worldUi.DLGuideCommon("Colortone", "TiSeb.Colortone.1.var", "8764");
                }
            }
		}

         public void MakeCUAForColortone(){
            var url = GetColorcorrectURL();
            core.customUnityAssetItem.SummonAtomMarker(ColortoneCUAName, "pelvisControl", 0f, -0.2f, 0f, bSetParent:true, debugOn:false);
            core.StartCoroutine(SetColorCorrectCoroutine(url));
        }

        public IEnumerator SetColorCorrectCoroutine(string url){
            Atom tmpatom = null;
            for(int i = 0; i < 200; i++){
                tmpatom = core.customUnityAssetItem.GetCUAAtom(ColortoneCUAName);
                if(tmpatom == null) yield return null;
            }
            if(tmpatom == null) yield break;

            Utility.SetAssetBundleUrl(tmpatom, url);
            Utility.ResetAssetBundleName(tmpatom, ColortoneAssetNamePath);
        }

        public bool ExistsColortoneInTheScene(){
            var cua = GetColoCorrectAtom();
            return cua != null;
        }

        public bool ExistsColortoneInTheGame(){
            var url = GetColorcorrectURL();
            return url != "";
        }

        public Atom GetColoCorrectAtom(){
            // return Utility.GetAssetIncludeAtom(ColortoneAssetURLPath);
            return Utility.GetAssetIncludeAtom(ColortoneAssetFilename);
        }

		public void SetColorCorrectOnOff(bool isOn){
			var cua = GetColoCorrectAtom();
             if(isOn){
    			Utility.ResetAssetBundleName(cua, ColortoneAssetNamePath);
            }else{
                if(cua == null) return;
                Utility.ResetAssetBundleName(cua, "");		
            }
		}
    }

    public class SubBase{
        public string LabelName = ""; //"PostMagic";
        public string CslistName = ""; //"MacGruber_PostMagic";
        public string ShortName = ""; //"MacGruber.PostMagic";

        public string ScriptPath = ""; //"Custom\\Scripts\\MacGruber\\PostMagic\\MacGruber_PostMagic.cslist";
        public string SearchPattern = ""; //"*PostMagic*";
        public int pluginNum = 99; //96;

        public string HubResourceName = ""; //"MacGruber.PostMagic.3.var";
        public string HubResourceNumber = ""; //"161";
        public string pluginType = "Person";
        public bool InDLing = false;
        public Action<bool> OnOffAction = null;
        public Func<bool> CheckResourceInTheSceneFunc = null;
        public Action OnLoadAction = null;
        public EBATCore core;

        public SubBase(EBATCore core){
            this.core = core;
        }

        public void Init(
            string LabelName, string ShortName, string ScriptPath, string SearchPattern, int pluginNum, 
            string HubResourceName, string HubResourceNumber, string pluginType, 
            Action<bool> OnOffAction, Func<bool> CheckResourceInTheSceneFunc, Action OnLoadAction
        ){
            this.LabelName = LabelName;
            this.ShortName = ShortName;
            this.ScriptPath = ScriptPath;
            this.SearchPattern = SearchPattern;
            this.pluginNum = pluginNum;
            this.pluginType = pluginType;
            this.HubResourceName = HubResourceName;
            this.HubResourceNumber = HubResourceNumber;
            this.OnOffAction = OnOffAction;
            this.CheckResourceInTheSceneFunc = CheckResourceInTheSceneFunc;
            this.OnLoadAction = OnLoadAction;
        }

        public void OnLoadDone(){
            if(pluginType == "Person"){
                if(Utility.HasThisPlugin(core.atom, CslistName)){
                    OnLoadAction();
                }
            }else if(pluginType == "Session"){
                if(Utility.HasThisPlugin(atom:null, pattern:CslistName)){
                    OnLoadAction();
                }
            }
        }

        public void OnMenuRecoveringDone(){
            if(!IsInDLing() || !ExistResourceInTheGame()) return;
            InDLing = false;
            LoadResource();
        }

        public bool IsInDLing(){
            return InDLing;
        }

        public bool IsReloadNeeded(){
            return ExistResourceInTheGame() && !ExistResourceInTheScene();
        }

        public void SetResource(bool isOn){
            // Utility.Debug($"route1: LabelName:{LabelName}, OnOffAction:{OnOffAction}, isOn:{isOn}");
            if(!isOn){
                // Utility.Debug($"route2: labelname:{LabelName}");
                SetResourceOnOff(false);
            }else if(ExistResourceInTheScene()){
                // Utility.Debug($"route3: labelname:{LabelName}");
                SetResourceOnOff(true);
            }else if(ExistResourceInTheGame()){
                // Utility.Debug($"route4: labelname:{LabelName}");
                LoadResource();
            }else{
                // Utility.Debug($"route5: labelname:{LabelName}");
                InDLing = true;
                core.worldUi.DLGuideCommon(LabelName, HubResourceName, HubResourceNumber);
            }
            // Utility.Debug($"route6: labelname:{LabelName}");
        }

        public bool ExistResourceInTheScene(){
            // Utility.Debug($"CheckResourceInTheSceneFunc:{CheckResourceInTheSceneFunc}");
            var result = CheckResourceInTheSceneFunc();
            // Utility.Debug($"LabelName:{LabelName}, result:{result}");
            return result;
        }

        public bool ExistResourceInTheGame(){
            InDLing = false;
            var url = GetResourceURL();
            return url != "";
        }

        public void SetResourceOnOff(bool isOn){
            // Utility.Debug($"LabelName:{LabelName}, OnOffAction:{OnOffAction}, isOn:{isOn}");
            OnOffAction(isOn);
        }

        public string GetResourceURL(){
			var url = Utility.GetPluginURL(SearchPattern, ScriptPath);
            return url;
        }

        public void LoadResource(){
			var url = GetResourceURL();
            Atom atom = pluginType == "Person" ? core.atom : null;
            // Utility.Debug($"LoadResource: atom:{atom}, url:{url}");

			Utility.LoadPluginIfNeeded(atom, url, LabelName, num:pluginNum);
            if(pluginType == "Session"){
                core.SetMemoryBool("Reload:" + LabelName, false);
            }
        }
    }
}