#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
#endregion

namespace Cotyounoyume{
	public class EyesControl
	{
		#region common or UI Properties
		public Atom atom;
		public EBATCore core;
		public bool disableLog;
		
		public bool isEyecontrolActive = false;
		public bool eyecontrolOnGlobal = true;
		public bool eyecontrolRollupOnGlobal = true;
		public bool lookModeGlobal = true;	
		public float eyecontrolClosingTimeInterval = 2f;
		#endregion
		
		#region eyeControl
		protected Transform eyeControlOrig;
		protected FreeControllerV3 headControl;
		protected FreeControllerV3 eyeControl;
		protected float eyecontrolDegree;
		protected float baseDegree = 15f;
		protected float eyecontrolClosingTime;
		protected float eyecontrolOpeningTime;
		protected float eyecontrolRandomOffset = 0.2f;
		#endregion
		
		public EyesControl(EBATCore core){
			atom = core.atom;
			isEyecontrolActive = false;
			this.core = core;
		}
		
		public void LoadSettings(){		
			try{
				lookModeGlobal = core.settings["General"]["LookMode"].AsBool;
				eyecontrolOnGlobal = core.settings["General"]["Eyes Control On"].AsBool;
				eyecontrolRollupOnGlobal = core.settings["General"]["Eyes Rollup On"].AsBool;
				eyecontrolClosingTimeInterval = core.settings["General"]["Eyes Control interval Time"].AsFloat;
			}catch (Exception e){
				core.LoggerE(e);
			}
		}
		
		public void ApplySettings(){
			try{
				atom.GetStorableByID("Eyes").SetStringChooserParamValue("lookMode", core.settings["General"]["LookMode"]);
			}catch (Exception e){
				core.LoggerE(e);
			}
		}
		
		public void Maintenance(){
			try{
				#region initialize eyeControl
				headControl = atom.GetStorableByID("headControl") as FreeControllerV3;
				eyeControl = atom.GetStorableByID("eyeTargetControl") as FreeControllerV3;			
				SetInitialTransformOfEyeControl();
				#endregion
			}catch (Exception e){
				core.LoggerE(e);
			}
		}

		public void SetInitialTransformOfEyeControl(){
			if(!core.IsLoaded()){ return; }
			
			try{
				eyeControl.transform.position = Utility.GetPosition(headControl);
				eyeControl.transform.rotation = Utility.GetRotation(headControl);
				eyeControl.transform.Translate(0f, 0f, 5f);

			}catch (Exception e){
				core.LoggerE(e);
			}
		}
		
		public void ClearEyeControlClosingTime()
		{
			eyecontrolClosingTime = 0;
		}
		
		public void Update(){
			if(!core.IsLoaded()){ return; }
			if(core.expression2 == null) 
			{
				Utility.Debug("expression2 is null.");
			}
			try{
				if(eyecontrolOnGlobal == false){
					return;
				}


				float delta = core.currentTime - eyecontrolOpeningTime;
				// Utility.Debug($"delta:{delta}, eyecontrolClosingTime:{eyecontrolClosingTime}");
				if(core.expression2 == null) return;
				if(!core.expression2.IsUsingSpecial() && delta < eyecontrolClosingTime){
					return;
				}else{
				}
				float randomOffset = UnityEngine.Random.Range(0.1f, eyecontrolRandomOffset);
				eyecontrolClosingTime = eyecontrolClosingTimeInterval + randomOffset;
				isEyecontrolActive = true;
				
				eyecontrolOpeningTime = core.currentTime;
				string direction = CalcEyeDirection();
				string[] directions = direction.Split(':');
				string upDown = directions[0];
				string rightLeft = directions[1];
				float eyecontrolStrength = CalcEyeStrength();
				if(upDown == "neutral" && rightLeft == "neutral"){
					EyeControlBaseSwitch("Player");
				}else{
					EyeControlBaseSwitch("Target");
//					eyeControlBaseSwitch("None");
					ChangeEyeControl(upDown, rightLeft, eyecontrolStrength);
//					changeEyeControl(upDown, rightLeft, eyecontrolStrength);
				}
			}catch (Exception e){
				core.LoggerE(e);
			}
		}

		public float CalcEyeStrength(){
			try{
				if(core.expression2.IsUsingSpecial())
				{
					return UnityEngine.Random.Range(0.7f, 1.5f);
				}
				
				float strength = 0;
				int stage;
				stage = core.GetStage();
				
				if(stage == 0){
					strength = UnityEngine.Random.Range(0.2f, 0.5f);
				}else if(stage == 1){
					strength = UnityEngine.Random.Range(0.2f, 0.5f);
				}else if(stage == 2){
					strength = UnityEngine.Random.Range(0.2f, 0.5f);
				}else if(stage == 3){
					strength = UnityEngine.Random.Range(0.2f, 0.5f);
				}else if(stage == 4){
					if(eyecontrolRollupOnGlobal){
						strength = UnityEngine.Random.Range(0.7f, 1.5f);
					}else{
						strength = UnityEngine.Random.Range(0.2f, 0.5f);
					}				
				}else{
					if(eyecontrolRollupOnGlobal){
						strength = UnityEngine.Random.Range(0.7f, 1.5f);
					}else{
						strength = UnityEngine.Random.Range(0.2f, 0.5f);
					}
				}
				return strength;
			}catch (Exception e){
				core.LoggerE(e);
				return 0f;
			}
		}
		
		public string CalcEyeDirection(){
			try{
				if(core.expression2.IsUsingSpecial()){
					string specialDirection = core.expression2.GetCurrentEyesDirection();
					string resultDirection = "neutral:neutral";
					
					if(specialDirection == "Neutral")	{ resultDirection = "neutral:neutral"; }
					if(specialDirection == "Up")		{ resultDirection = "up:neutral"; }
					if(specialDirection == "Down")		{ resultDirection = "down:neutral"; }
					if(specialDirection == "Right")		{ resultDirection = "neutral:right"; }
					if(specialDirection == "Left")		{ resultDirection = "neutral:left"; }
					return resultDirection;
				}
				
				string direction = "neutral:neutral";
				
				float rand1 = UnityEngine.Random.Range(0f, 1.0f);
				float rand2 = UnityEngine.Random.Range(0f, 1.0f);
				float rand3 = UnityEngine.Random.Range(0f, 1.0f);
				int stage;
				stage = core.GetStage();
				if(core.isFirstTerm){
					if(rand1 > 0.1f){
						direction = "up:neutral";			
					}else{
						direction = "neutral:neutral";
					}
					
				}else if(core.isEmbarassed){
					if(rand1 > 0.1f && stage >= 1){
						direction = (rand2 <= 0.5f) ? "down:right" : "down:left";						
					}else{
						direction = "neutral:neutral";
					}
				}else if(stage == 0){
					direction = "neutral:neutral";
				}else if(stage == 1){
					if(rand1 < 0.2f){
						direction = (rand2 <= 0.5f) ? "down:right" : "down:left";
					}else{
						direction = "neutral:neutral";
					}
				}else if(stage == 2){
					if(rand1 < 0.5f){
						direction = (rand2 <= 0.5f) ? "down:right" : "down:left";
					}else{
						direction = "neutral:neutral";
					}
				}else if(stage == 3){
					if(rand1 < 0.8f){
						direction = (rand2 <= 0.5f) ? "down:right" : "down:left";
					}else{
						direction = "neutral:neutral";					
					}
				}else if(stage == 4){
					if(rand1 > 0.0f && rand1 <= 0.3f){
						direction = "neutral:neutral";											
					}else if(rand1 > 0.3f && rand1 <= 0.7f){
						direction = (rand2 <= 0.5f) ? "down:right" : "down:left";						
					}else if(rand1 > 0.7f){
						if(eyecontrolRollupOnGlobal){
							direction = "up:neutral";
						}else{
							direction = (rand2 <= 0.5f) ? "down:right" : "down:left";						
						}
					}else{
						direction = "neutral:neutral";					
					}
				}else{
					if(rand1 > 0.0f && rand1 <= 0.2f){
						direction = "neutral:neutral";											
					}else if(rand1 > 0.2f && rand1 <= 0.6f){
						direction = (rand2 <= 0.5f) ? "down:right" : "down:left";						
					}else if(rand1 > 0.6f){
						if(eyecontrolRollupOnGlobal){
							direction = "up:neutral";
						}else{
							direction = (rand2 <= 0.5f) ? "down:right" : "down:left";						
						}
					}else{
						direction = "neutral:neutral";					
					}
				}
				return direction;
			}catch (Exception e){
				core.LoggerE(e);
				return "neutral:neutral";
			}
		}
		
		public void EyeControlBaseSwitch(string type){
			try{
				atom.GetStorableByID("Eyes").SetStringChooserParamValue("lookMode", type);
				SetInitialTransformOfEyeControl();
				/*
				if(type == "Player"){
					setInitialTransformOfEyeControl();
				}
				*/
			}catch (Exception e){
				core.LoggerE(e);
			}
		}
		
		public void Manage(float time){
			try{
				if(!core.IsLoaded()){ return; }
				
				float delta = time - eyecontrolOpeningTime;
				if(core.expression2.IsUsingSpecial()){ return; }
				
				if(isEyecontrolActive){
					if(core.GetStage() < 2 || delta > eyecontrolClosingTime && core.isEmbarassed == false && core.isDuringOrgasm == false){
						isEyecontrolActive = false;
						EyeControlBaseSwitch("Player");
					}
				}
			}catch (Exception e){
				core.LoggerE(e);
			}
		}

		public void ChangeEyeControl(string upDown, string rightLeft, float strength){
			try{
				SetInitialTransformOfEyeControl();
				float degreeUpDown = 0f;
				float degreeRightLeft = 0f;
				if(upDown != "neutral"){
					degreeUpDown = (upDown == "up") ? -1 * baseDegree : baseDegree/2;
					degreeUpDown *= strength;
					eyeControl.transform.RotateAround(
						headControl.transform.position, 
						headControl.transform.right, degreeUpDown); 
				}
				if(rightLeft != "neutral"){
					degreeRightLeft = (rightLeft == "left") ? -1 * baseDegree : baseDegree;	
					degreeRightLeft *= strength;
					eyeControl.transform.RotateAround(
						headControl.transform.position, 
						headControl.transform.up, degreeRightLeft); 
				}
			}catch (Exception e){
				core.LoggerE(e);
			}
		}
		
		public void ChangeEyeControl_new(string upDown, string leftRight, float strength){
			try{
				float upDownF = 0f;
				float leftRightF = 0f;
				float maxStrength = 30f;
				if(upDown != "neutral"){
					upDownF = (upDown == "up") ? -1.0f * strength * maxStrength : 1.0f * strength * maxStrength;
				}
				if(leftRight != "neutral"){
					leftRightF = (leftRight == "left") ? -1.0f * strength * maxStrength : 1.0f * strength * maxStrength;
				}
				core.atom.GetStorableByID("Eyes").SetFloatParamValue("upDownAngleAdjust", upDownF);
				core.atom.GetStorableByID("Eyes").SetFloatParamValue("leftRightAngleAdjust", leftRightF);
			}catch (Exception e){
				core.LoggerE(e);
			}
		}
		
		public void EyesControlOff(){
			EyeControlBaseSwitch("Player");
			SetInitialTransformOfEyeControl();
			isEyecontrolActive = false;
		}
		
		public void EyesControlOn(){
			Update();
			isEyecontrolActive = true;
		}
	}
}