#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
#endregion

namespace Cotyounoyume{
	
	public class FaceMorphManager{
		public Atom atom;

		public string[] listMorphnameBrow = {"CTY_Brow1", "CTY_BROW2b", "CTY_BROW3", "CTY_BROW4"};
		public string[] listMorphnameEyes = {"CTY_EYES_CLOSE", "CTY_EYES_CLOSE_3QUARTER", "CTY_EYES_CLOSE_HALF", "CTY_EYES_CLOSE_QUARTER", "CTY_EYES_CLOSE2", "CTY_EYES2"};
		public string[] listMorphnameMouth = {"CTY_MOUTH", "CTY_MOUTH2", "CTY_MOUTH3"};
		
		public string[] listMorphnameBrowBL = {"Brow Squeeze", "Pain", "Sad"};
		public string[] listMorphnameEyesBL = {"AAeyeclosebrowsqueeze2", "Eyes Closed", "Eyelids Bottom Up Left", "Eyelids Top Up Left", "Pain", "Sad"};
		public string[] listMorphnameMouthBL = {"Mouth Open", "Mouth Open Wide"};
		public string[] listTypeAll    = {"brow", "eyes", "mouth", "special"};
		public string[] listType       = {"brow", "eyes", "mouth"};
		public string[] listTypeSp     = {"special"};
		public Dictionary<string, float> startTime;
		public Dictionary<string, float> peakTime;
		public Dictionary<string, float> finishTime;
		public Dictionary<string, float> initialStrength;
		public Dictionary<string, float> targetStrength;
		public Dictionary<string, float> currentStrength;
		public Dictionary<string, string[]> dicListMorphs;
		public Dictionary<string, string[]> dicListBLMorphs;
		public Dictionary<string, float> BLStrength;

		public Dictionary<string, string> currentMorph;
		public Dictionary<string, DAZMorph> morphMap;
		public float peakTimeOffset   = 0.2f;
		public float finishTimeOffset = 2.8f;
		public float randomOffsetMin  = 0.0f;
		public float randomOffsetMax	= 0.3f;
		public float randomStrengthOffsetMin = 0.8f;
		public float randomStrengthOffsetMax = 1.2f;
		
		public float baseMultiplier = 1.0f;
		public float baseStrength = 0.3f;
		public float stageMax = 5.0f;
		public float lastUpdateInterval = 0.6f;
		public float lastUpdateTime = 0.0f;

		public bool specialOn = false;
		public bool isUsingSpecial = false;
		public bool expressionContinuityOn = true;
		public int continuityCount = 0;
		public int continuityCountMax = 4;
		public int continuityCountMin = 2;
		public Dictionary<string, string> lastMorph;
		
		public bool isExpressionActive = false;
		public Dictionary<string, bool> isExpressionActivePartial;
		
		public Dictionary<string, List<string>> levelVsExpressions;
		public Dictionary<string, float> levelVsFrequency;
		public Dictionary<string, string> eyesDirection;
		protected GenerateDAZMorphsControlUI morphControl;
		
		public FaceMorphManager(Atom atom){
			this.atom = atom;
			dicListMorphs = new Dictionary<string, string[]>();
			currentMorph = new Dictionary<string, string>();
			startTime = new Dictionary<string, float>();
			peakTime = new Dictionary<string, float>();
			finishTime = new Dictionary<string, float>();
			initialStrength = new Dictionary<string, float>();
			targetStrength = new Dictionary<string, float>();
			currentStrength = new Dictionary<string, float>();
			isExpressionActivePartial = new Dictionary<string, bool>();
			morphMap = new Dictionary<string, DAZMorph>();
			lastMorph = new Dictionary<string, string>();
			
			dicListBLMorphs = new Dictionary<string, string[]>();
			BLStrength = new Dictionary<string, float>();
			
			dicListMorphs.Add("brow",  listMorphnameBrow);
			dicListMorphs.Add("eyes",  listMorphnameEyes);
			dicListMorphs.Add("mouth", listMorphnameMouth);
			dicListMorphs.Add("special", new List<string>().ToArray());
			
			dicListBLMorphs.Add("brow",  listMorphnameBrowBL);
			dicListBLMorphs.Add("eyes",  listMorphnameEyesBL);
			dicListBLMorphs.Add("mouth", listMorphnameMouthBL);
			dicListBLMorphs.Add("special", new List<string>().ToArray());
			
			foreach(var type in listTypeAll){
				currentMorph.Add(type,"");
				initialStrength.Add(type, 0f);
				targetStrength.Add(type, 0f);
				currentStrength.Add(type, 0f);
				startTime.Add(type, 0f);
				peakTime.Add(type, 0f);
				finishTime.Add(type, 0f);
				BLStrength.Add(type, 0f);
				lastMorph.Add(type, "");
			}
			SetFlags(false);
			
			JSONStorable geometry = atom.GetStorableByID("geometry");
			DAZCharacterSelector character = geometry as DAZCharacterSelector;
			morphControl = character.morphsControlUI;
			continuityCount = 0;
			
			foreach(var type in listType){
				UpdateMorphs(type, false);
				UpdateMorphs(type, true);
			}
		}
		
		public void ClearContinuityCount()
		{
			continuityCount = 0;
		}
		
		public string GetCurrentSpecialExpression()
		{
			return currentMorph.ContainsKey("special") ? currentMorph["special"] : "";
		}
		
		public void UpdateLevelVsExpressions(Dictionary<string, List<string>> dic){
			this.levelVsExpressions = dic;
			UpdateMorphs("special", false);
		}

		public void UpdateLevelVsFrequency(Dictionary<string, float> dic){
			this.levelVsFrequency = dic;
		}
		
		public void UpdateEyesDirection(Dictionary<string, string> dic)
		{
			this.eyesDirection = dic;
		}
		
		public void SetMorphControlIfNeeded(){
			if(morphControl == null){
				JSONStorable geometry = atom.GetStorableByID("geometry");
				DAZCharacterSelector character = geometry as DAZCharacterSelector;
				morphControl = character.morphsControlUI;
			}
		}
		
		public void UpdateMorphs(string type, bool isBL = false){
			SetMorphControlIfNeeded();
			Dictionary<string, string[]> dic = (!isBL) ? dicListMorphs : dicListBLMorphs;
			
			if(type == "special")
			{
				foreach(KeyValuePair<string, List<string>> kvp in levelVsExpressions){
					string key = kvp.Key;
					List<string> list = kvp.Value;
					foreach(var value in kvp.Value){
						foreach(var name in list)
						{
							DAZMorph morph = morphControl.GetMorphByDisplayName(name);
							if(!morphMap.ContainsKey(name))
							{
								morphMap.Add(name, morph);
							}
						}
					}
				}
			}
			else
			{
				if(!dic.ContainsKey(type)) return;
				foreach(var name in dic[type]){
					DAZMorph morph = morphControl.GetMorphByDisplayName(name);
					if(morph == null){
						continue;
					}
					if(!morphMap.ContainsKey(name)){
						morphMap.Add(name, morph);
					}
				}
			}
		}
		
		public void SetContinuityCount(){
			if(continuityCount == 0){
				continuityCount = UnityEngine.Random.Range(continuityCountMin, continuityCountMax);
			}
		}
		
		public bool PickSpecialOrNot(int stage, bool isDuringAfterglow, bool isStrongWave, string stageLevel){
			float rnd = UnityEngine.Random.Range(0, 1f);
			int expressionCount = (!levelVsExpressions.ContainsKey(stageLevel)) ? 0 : levelVsExpressions[stageLevel].Count;
			if (expressionCount == 0) { return false; }

			if(!levelVsExpressions.ContainsKey(stageLevel)) return false;
			if (rnd <= levelVsFrequency[stageLevel]){
				return true;
			}else{
				return false;
			}
		}
		
		public void ForceEnableSpecial()
		{
			SetContinuityCount();
			isUsingSpecial = true;
			SetSpecialMorphList("All");
		}
		
		public void PickExpression(float currentTime, int stage, bool isDuringAfterglow, bool isStrongWave = false, string stageLevel = "Lv1"){
			if(currentTime < lastUpdateInterval + lastUpdateTime){ return; }
			lastUpdateTime = currentTime;

			bool repeatRoute = false;
			if(continuityCount > 0){
				continuityCount--;
				repeatRoute = true;
			}else{
				SetContinuityCount();
				isUsingSpecial = PickSpecialOrNot(stage, isDuringAfterglow, isStrongWave, stageLevel);
			}

			string[] listTypeLocal = (isUsingSpecial) ? listTypeSp : listType;
			foreach(var type in listTypeLocal){
				if(repeatRoute){
					PickExpressionByMorphName(currentTime, type, lastMorph[type], stage, isDuringAfterglow, isStrongWave, stageLevel);
				}else{
					PickExpressionByStage(currentTime, type, stage, isDuringAfterglow, isStrongWave, stageLevel);
				}
			}
		}
		
		public void SetSpecialMorphList(string stageLevel = "Lv1")
		{
			if(!dicListMorphs.ContainsKey("special")) return;
			if(!levelVsExpressions.ContainsKey(stageLevel)) return;
			dicListMorphs["special"] = levelVsExpressions[stageLevel].ToArray();
		}
		
		public string GetMorphByType(string type, int stage, bool isDuringAfterglow, bool isStrongWave = false, string stageLevel = "Lv1"){
			string result = null;
			
			try{

			if(type == "special"){
				SetSpecialMorphList(stageLevel);
				int num = UnityEngine.Random.Range(0, dicListMorphs["special"].Length);
				result = dicListMorphs["special"][num];
			}else{
				string[] list = dicListMorphs[type];
				int num = UnityEngine.Random.Range(0, list.Length);
				if(stageLevel == "Lv5" || stageLevel == "Lv6" || stageLevel == "Lv7"){
					if(type == "mouth"){
						num = (UnityEngine.Random.Range(0, 1.0f) < 0.25f) ? 2 : 1; //CTY_MOUTH3 or CTY_MOUTH2 
					}else if(type == "eyes"){
						num = (UnityEngine.Random.Range(0, 1.0f) < 0.25f) ? 0 : num; //CTY_EYES_CLOSE
					}
				}
				result = list[num];
			}
			return result;
			}catch(Exception e){
				Utility.Debug($"ERROR: type:{type}, stage{stage}, isDuringAfterglow:{isDuringAfterglow}, isStrongWave:{isStrongWave}, stageLevel:{stageLevel}");
				Utility.LoggerE(e);
				return "";
			}
		}

		public void PickExpressionByStage(float currentTime, string type, int stage, bool isDuringAfterglow, bool isStrongWave = false, string stageLevel = "Lv1"){
			string morphName = GetMorphByType(type, stage, isDuringAfterglow, isStrongWave, stageLevel);
			lastMorph[type] = morphName;
			PickExpressionByMorphName(currentTime, type, morphName, stage, isDuringAfterglow, isStrongWave, stageLevel);
		}
		
		public void PickExpressionByMorphName(float currentTime, string type, string morphName, int stage, bool isDuringAfterglow, bool isStrongWave = false, string stageLevel = "Lv1", float duration = 0.0f){
			lastMorph[type] = morphName;
			SetMorph(type, morphName, stage, isDuringAfterglow);
			SetTargetStrengthAndTime(currentTime, type, stage, isDuringAfterglow, isStrongWave, duration, stageLevel);
		}
		
		
		public void OpenMouth(float currentTime, float duration, int stage, bool isDuringAfterglow, bool isStrongWave = false, string stageLevel = "Lv1"){
			if(continuityCount > 0)
			{
				if(lastMorph["mouth"] == "CTY_MOUTH3" || isUsingSpecial)
				{
					return;
				}
			}
			
			PickExpressionByMorphName(currentTime, "mouth", "CTY_MOUTH2", stage, isDuringAfterglow, isStrongWave, stageLevel, duration);
		}
		
		public void SetTargetStrengthAndTime(float currentTime, string type, int stage, bool isDuringAfterglow, bool isStrongWave = false, float duration = 0.0f, string stageLevel = "Lv1"){
			isExpressionActivePartial[type] = true;
			isExpressionActive = true;
			SetTarGetTime(currentTime, type, stage, isDuringAfterglow, isStrongWave, duration);
			SetTargetStrength(currentTime, type, stage, isDuringAfterglow, isStrongWave, stageLevel);
		}

		public void SetTarGetTime(float currentTime, string type, int stage, bool isDuringAfterglow, bool isStrongWave = false, float duration = 0.0f){
			float afterglowRatio = (isDuringAfterglow) ? 2.0f : 1.0f;
			startTime[type] = currentTime;
			if(duration == 0.0f){
				peakTime[type] = startTime[type] + peakTimeOffset;
				finishTime[type] = peakTime[type] + finishTimeOffset * afterglowRatio;				
			}else{
				peakTime[type] = startTime[type] + peakTimeOffset;
				finishTime[type] = startTime[type] + duration;
			}
		}

		public void SetTargetStrength(float currentTime, string type, int stage, bool isDuringAfterglow, bool isStrongWave = false, string stageLevel = "Lv1"){
			if(stageLevel == "Lv5" || stageLevel == "Lv6" || stageLevel == "Lv7"){			
				isDuringAfterglow = false;
				isStrongWave = true;
			}
			initialStrength[type] = currentStrength[type];
			float randomRatio = UnityEngine.Random.Range(randomStrengthOffsetMin, randomStrengthOffsetMax);
			targetStrength[type] = (baseStrength + (1 - baseStrength) * (float)stage/stageMax) * randomRatio;
			targetStrength[type] *= baseMultiplier * ((isDuringAfterglow) ? 0.5f : 1.0f);
			if(targetStrength[type] < currentStrength[type]){
				targetStrength[type] = currentStrength[type];
			}
			if(isStrongWave && !isDuringAfterglow){
				if(type == "brow"){
					//nop
				}else if(type == "mouth"){
					targetStrength[type] *= 2.0f;
					if(targetStrength[type] > 1.3f){
						targetStrength[type] = 1.3f;
					}
				}else if(type == "eyes"){
					targetStrength[type] *= 2.0f;
					if(targetStrength[type] > 1.0f){
						targetStrength[type] = 1.0f;
					}
				}else if(type == "special"){
					targetStrength[type] *= 2.0f;
					if(targetStrength[type] > 1.0f){
						targetStrength[type] = 1.0f;
					}
				}else{
					targetStrength[type] *= 1.5f;
					if(targetStrength[type] > 1.0f){
						targetStrength[type] = 1.0f;
					}
				}
			}
		}

		public void SetMorph(string type, string morphName, int stage, bool isDuringAfterglow){
			if(type.StartsWith("special")){
				foreach(var typeTmp in listType){
					currentMorph[typeTmp] = "";
				}
			}else{
				if(type != "mouth"){
					currentMorph["special"] = "";
				}
			}
			currentMorph[type] = morphName;			
		}
		
		public void ManageStrength(float currentTime){
			try{
			if(!isExpressionActive){ return; }
			FlagUpdate(currentTime);
			if(!isExpressionActive){
				SetFlags(false);
				ClearCurrentMorph();
				return;
			}
			
			foreach(var type in listTypeAll){
				if(isExpressionActivePartial[type]){
					CalculateBLStrength(currentTime, type);				
					CalculateStrength(currentTime, type);
				}
			}
			foreach(var type in listTypeAll){
				if(isExpressionActivePartial.ContainsKey(type) && isExpressionActivePartial[type]){
					if(currentMorph.ContainsKey(type) && currentStrength.ContainsKey(type) && morphMap.ContainsKey(currentMorph[type])){
						morphMap[currentMorph[type]].SetValue(currentStrength[type]);
					}
				}
			}
			ReduceExcludedMorphs();	
			
			}catch(Exception e){
				Utility.LoggerE(e);
			}
		}
		
		public void CalculateStrength(float currentTime, string type){
			try{
				bool isBeforePeak = (currentTime < peakTime[type]);
				float fromStrength = isBeforePeak ? initialStrength[type] : targetStrength[type];
				float toStrength = isBeforePeak ? targetStrength[type] : 0.0f;
				currentStrength[type] = Utility.CalculateProgress(fromStrength, toStrength, startTime[type], finishTime[type], currentTime);
				currentStrength[type] *= (1 - BLStrength[type]);
			}catch (Exception e){
				Utility.LoggerE(e);
			}
		}

		public void CalculateBLStrength(float currentTime, string type){
			BLStrength[type] = 0f;
			foreach(var name in dicListBLMorphs[type]){
				if(morphMap.ContainsKey(name)){
					if(BLStrength[type] < morphMap[name].morphValue){
						BLStrength[type] = morphMap[name].morphValue;
					}
				}
			}
		}
		
		public void SetFlags(bool value){
			isExpressionActive = value;
			foreach(var type in listTypeAll){
				isExpressionActivePartial[type] = value;
			}
		}
		
		public void ClearCurrentMorph(){
			foreach(var type in listTypeAll){
				currentMorph[type] = "";
			}
			SetDefaultValue();
			isUsingSpecial = false;
		}
		
		public void ReduceExcludedMorphs(){
			string targetType = "";
			string nameType = "";
			try{
				foreach(var type in listTypeAll){
					targetType = type;
					foreach(var name in dicListMorphs[type]){
						nameType = name;
						if(!currentMorph.ContainsKey(type)) 
						{
							//�ʏ킠�蓾�Ȃ�(type���̂����݂��Ȃ��̂ŉ����d�l�ύX���Ȃ��ƋN���Ȃ�)
							Utility.Info("ERROR: currentMorph doesn't have this type:" + type);
							continue;
						}
						if(!morphMap.ContainsKey(name)) 
						{
							//�ʏ킠�蓾�Ȃ�(morphMap�ɑ��݂��Ȃ��̂ŉ�����������������)
							Utility.Info("ERROR: morphMap[" + type + "] doesn't have this name:" + name);
							continue;
						}
						if (name == currentMorph[type])
						{
							//currentMorph[type] == name �Ȃ�Reduce���Ȃ��Ă悢�̂ŃX�L�b�v
							continue;
						}

						if(morphMap[name].morphValue <= 0.000001f){
							morphMap[name].morphValue = 0.0f;
						}else if(morphMap[name].morphValue < 0.01f){
							morphMap[name].SetValue(0.000001f);
						}else if(morphMap[name].morphValue >= 0.01f){
							morphMap[name].SetValue(morphMap[name].morphValue * 0.8f);
						}
					}
				}			
			}catch(Exception e){
				Utility.Info("type:" + targetType + ", name:" + nameType );
				Utility.LoggerE(e);
			}
		}
		
		public void SetDefaultValue(){
			foreach(var type in listTypeAll){
				if(!dicListMorphs.ContainsKey(type)) continue;
				foreach(var name in dicListMorphs[type]){
					if(morphMap.ContainsKey(name)){
						morphMap[name].SetValue(0.000001f);
					}
				}
			}
		}
		
		public void SetZeroValue(){
			foreach(var type in listTypeAll){
				if(!dicListMorphs.ContainsKey(type)) continue;
				foreach(var name in dicListMorphs[type]){
					if(morphMap.ContainsKey(name)){
						morphMap[name].SetValue(0.0f);
					}
				}
			}
		}
		
		public void FlagUpdate(float currentTime){
			bool result = false;
			foreach(var type in listTypeAll){
				if(!finishTime.ContainsKey(type)) continue;

				if(currentTime < finishTime[type]){
					if(isExpressionActivePartial.ContainsKey(type)) isExpressionActivePartial[type] = true;
					result = true;
				}else{
					if(isExpressionActivePartial.ContainsKey(type)) isExpressionActivePartial[type] = false;
					isExpressionActivePartial[type] = false;					
				}
			}
			isExpressionActive = result;
		}
		
		
		public bool IsExpired(float currentTime){
			bool result = true;
			foreach(var type in listTypeAll){
				if(currentTime < finishTime[type]){
					result = false;
				}
			}
			return result;
		}
		
		public void StopExpression(){
			SetDefaultValue();
			SetZeroValue();
			SetFlags(false);
			ClearCurrentMorph();
			isUsingSpecial = false;
		}
		
		public void StopAllExpressionInPose(){
			try{
				List<DAZMorph> othersMorphs = new List<DAZMorph>();
				
				morphControl.onlyShowActive = true;
				morphControl.showType = GenerateDAZMorphsControlUI.ShowType.Pose;
				List<DAZMorph> morphs = morphControl.GetMorphs();
				for(int i = 0; i < morphs.Count; i++){
					if(morphs[i].resolvedRegionName.StartsWith("Pose/Head") && morphs[i].active){
						othersMorphs.Add(morphs[i]);
					}
				}
				
				for(int i = 0; i < othersMorphs.Count; i++){
					othersMorphs[i].SetDefaultValue();
				}			
			}catch (Exception e){
				Utility.LoggerE(e);
			}
		}
		
		public void OnDestroy(){
			
		}
		
	}
}