#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
#endregion

namespace Cotyounoyume{
	public class ActionAndRatio{
		public Dictionary<string, float> actionAndRatioMap;
		public List<string> actionList;
		public List<float> frequencyBin;
		
		public ActionAndRatio(){
			actionAndRatioMap = new Dictionary<string, float>();
			actionList = new List<string>();
			frequencyBin = new List<float>();
		}

		public void InitActionList(List<string> actionList){
			foreach(var name in actionList){
				this.actionList.Add(name);
			}
		}
		
		public void SetActionAndRatio(string actionName, float ratio){
			if(actionAndRatioMap.ContainsKey(actionName)){
				actionAndRatioMap[actionName] = ratio;
			}else{
				actionAndRatioMap.Add(actionName, ratio);
			}
		}
		
		public float GetActionAndRatio(string actionName){
			if(actionAndRatioMap.ContainsKey(actionName)){
				return actionAndRatioMap[actionName];
			}else{
				return 0f;
			}
		}
		
		public void MakeRatioTable(){
			float current = 0;
			frequencyBin.Clear();
			frequencyBin.Add(0f);
			foreach(var name in actionList){
				current += actionAndRatioMap[name];
				frequencyBin.Add(current);
			}
			for(int i = 0; i < frequencyBin.Count; i++){
				frequencyBin[i] = frequencyBin[i] / current;
			}
		}
		
		public string PickAction(){
			float rnd = UnityEngine.Random.Range(0.0f,1.0f);
			int num = 0;
			for(int i = 0; i < frequencyBin.Count; i++){
				if(rnd < frequencyBin[i]){
					num = i - 1;
					break;
				}
			}
			return actionList[num];
		}
		
		public void OnDestroy(){
			actionAndRatioMap.Clear();
			actionList.Clear();
			frequencyBin.Clear();
		}
		
	}
	
	public class StageAndEvent{
		public List<string> actionList;
		public Dictionary<string, float> fireRatioMap;
		public Dictionary<string, ActionAndRatio> actionAndRatioMap;
		
		public StageAndEvent(){
			actionList = new List<string>();
			fireRatioMap = new Dictionary<string, float>();
			actionAndRatioMap = new Dictionary<string, ActionAndRatio>();
		}
		
		public void InitActionList(string[] actionList){
			foreach(var name in actionList){
				this.actionList.Add(name);
			}
		}
		
		public List<string> ListStages(){
			List<string> list = new List<string>();
			foreach(KeyValuePair<string, ActionAndRatio> kvp in actionAndRatioMap){
				list.Add(kvp.Key);
			}
			return list;
		}
		
		public void InitFireRatio(Dictionary<string, float> fireRatio){
			this.fireRatioMap.Clear();
			foreach(KeyValuePair<string, float> kvp in fireRatio){
				this.fireRatioMap.Add(kvp.Key, kvp.Value);
			}
		}
		
		public void SetFireRatio(string stage, float ratio){
			if(fireRatioMap.ContainsKey(stage)){
				fireRatioMap[stage] = ratio;
			}else{
				fireRatioMap.Add(stage, ratio);
			}
		}
		
		public float GetFireRatio(int stage, int count){
			string name = Utility.ResolveStageName(stage, count);
			return fireRatioMap[name];
		}

		public float GetFireRatio(string stage){
			return fireRatioMap[stage];
		}
		
		public string PickAction(string stage, bool bForce = false){
			float rnd = UnityEngine.Random.Range(0.0f,1.0f);
			if(rnd < fireRatioMap[stage] || bForce){
				return actionAndRatioMap[stage].PickAction();
			}else{
				return "";
			}
		}
		
		public string PickAction(int stage, int count, bool bForce = false){
			string name = Utility.ResolveStageName(stage, count);
			return PickAction(name, bForce);
		}
		
		public void InitActionAndRatioByStage(string stage, Dictionary<string, float> actionAndRatio){
			if(!this.actionAndRatioMap.ContainsKey(stage)){
				this.actionAndRatioMap.Add(stage, new ActionAndRatio());
				this.actionAndRatioMap[stage].InitActionList(this.actionList);
			}
			foreach(KeyValuePair<string, float> kvp in actionAndRatio){
				this.actionAndRatioMap[stage].SetActionAndRatio(kvp.Key, kvp.Value);
			}
			this.actionAndRatioMap[stage].MakeRatioTable();
		}
		
		public void SetActionAndRatio(string stage, string actionName, float ratio){
			this.actionAndRatioMap[stage].SetActionAndRatio(actionName, ratio);
			this.actionAndRatioMap[stage].MakeRatioTable();
		}
		
		public float GetRatioByStage(string stage, string actionName){
			return actionAndRatioMap[stage].GetActionAndRatio(actionName);
		}		
		
		public void OnDestroy(){
			foreach(KeyValuePair<string, ActionAndRatio> kvp in actionAndRatioMap){
				kvp.Value.OnDestroy();
			}
			actionAndRatioMap.Clear();
			actionList.Clear();
			fireRatioMap.Clear();
		}
		
	}
	
	public class AxisAndAngles{
		public Dictionary<string, float> angleMap;
		
		public AxisAndAngles(){
			angleMap = new Dictionary<string, float>(){
				{"up", 0f},
				{"forward", 0f},
				{"right", 0f},
				{"forward2", 0f},
				{"right2", 0f},
			};
		}
		
		public void SetAngle(string axis, float degree){
			if(angleMap.ContainsKey(axis)){
				angleMap[axis] = degree;
			}else{
				angleMap.Add(axis, degree);
			}
		}
		
		public float GetAngle(string axis){
			if(angleMap.ContainsKey(axis)){
				return angleMap[axis];
			}else{
				return 0f;
			}
		}
		
		public void OnDestroy(){
			angleMap.Clear();
		}
	}
	
	public class HeadController
	{
		#region common or UI Properties
		public Atom atom;
		public EBATCore core;
		public bool disableLog;
		public bool headMotionOnGlobal = true;
		public bool headMotionStopDuringPossessionOnGlobal = true;
		#endregion
		
		#region bellyControl
		protected Quaternion initialQuat;
		protected Quaternion recoveryQuat;
		protected Quaternion targetQuat;
		protected Dictionary<string, float> progress;
		protected Dictionary<string, CycleManager> timing;
		protected Dictionary<string, bool> isActive;
		protected Dictionary<string, bool> isForce;
		
		protected Dictionary<string, float> lastActionTime;
		protected float actionIntervalRatioMin = 0.5f;
		protected float actionIntervalRatioMax = 2.0f;
		
		protected FreeControllerV3 headControl;
		protected FreeControllerV3 chestControl;
		protected FreeControllerV3 neckControl;
		
		protected string[] actionList = {
			"Turn Head Back 1",
			"Turn Head Back 2",
			"Glance",
			"Face Forward",
			"Face Sideways 1",
			"Face Sideways 2",
			"Face Sideways 3",
		};
		
		protected string[] eventList = {
			"Stage<3:OrgasmCount<99",
			"Stage<5:OrgasmCount<99",
			"Stage=5:OrgasmCount<01",
			"Stage=5:OrgasmCount<03",
			"Stage=5:OrgasmCount<06",
			"Stage=5:OrgasmCount<09",
			"Stage=5:OrgasmCount<99",
		};
		
		protected string[] axisList = {
			"up", "forward", "right", "forward2", "right2",
		};
		
		protected Dictionary<string, string> motionType = new Dictionary<string, string>{
			{"Turn Head Back 1", "RoundTrip"},
			{"Turn Head Back 2", "RoundTrip"},
			{"Face Sideways 1",  "OneWay"},
			{"Face Sideways 2",  "OneWay"},
			{"Face Sideways 3",  "OneWay"},
			{"Glance",           "OneWay"},
			{"Face Forward",     "OneWay"},
		};
		
		protected Dictionary<string, float> actionInterval = new Dictionary<string, float>
		{
			["Turn Head Back 1"] = 0.5f
			, ["Turn Head Back 2"] = 6.0f
			, ["Face Sideways 1"] = 4.0f
			, ["Face Sideways 2"] = 6.0f
			, ["Face Sideways 3"] = 8.0f
			, ["Glance"] = 4.0f
			, ["Face Forward"] = 3.0f
		};
		
		public Dictionary<string, float> closingTimeOffset1 = new Dictionary<string, float>
		{
			["Turn Head Back 1"] = 0.1f
			, ["Turn Head Back 2"] = 0.2f
			, ["Face Sideways 1"] = 0.3f
			, ["Face Sideways 2"] = 0.5f
			, ["Face Sideways 3"] = 0.5f
			, ["Glance"] = 0.4f
			, ["Face Forward"] = 0.2f
		};
		public Dictionary<string, float> closingTimeOffset2 = new Dictionary<string, float>
		{
			["Turn Head Back 1"] = 0.5f
			, ["Turn Head Back 2"] = 1.5f
			, ["Face Sideways 1"] = 0.3f
			, ["Face Sideways 2"] = 0.5f
			, ["Face Sideways 3"] = 1.0f
			, ["Glance"] = 0.4f
			, ["Face Forward"] = 0.2f
		};		
		
		protected Dictionary<string, AxisAndAngles> angleMap;
		protected StageAndEvent stageAndEventMap;
		#endregion
		
		#region breath
		#endregion
		
		#region head
		public float bendBackwardThreshold = 0.2f;
		public float bendBackwardStrongThreshold = 0.2f;
		public float rbendBackward = 0.3f;
		public float rBigMotion = 0.03f;
		public float rFaceSidewaysWeak = 0.2f;
		public float rFaceSideways     = 0.2f;
		public float rFaceSidewaysPeak = 0.2f;
		public float glanceThreshold = 0.2f;
		
		public float timeRandomMinRatio = 0.7f;
		public float timeRandomMaxRatio = 2.0f;
		#endregion
		
		private float lastTime;
		
		public HeadController(EBATCore core){
			atom = core.atom;
			this.core = core;
			headControl  = Utility.GetFreeControllerV3(atom, "headControl");
			neckControl  = Utility.GetFreeControllerV3(atom, "neckControl");
			chestControl = Utility.GetFreeControllerV3(atom, "chestControl");
			
			headMotionOnGlobal = core.settings["General"]["Head Motion On"].AsBool;
			headMotionStopDuringPossessionOnGlobal = core.settings["General"]["HeadMotionOff During Possession"].AsBool;
			
			angleMap = new Dictionary<string, AxisAndAngles>();
			foreach(var actionName in actionList){
				angleMap.Add(actionName, new AxisAndAngles());
			}
			UpdateAngle();
			
		}
		
		public void LoadSettings(){
			
			stageAndEventMap = new StageAndEvent();
			stageAndEventMap.InitActionList(actionList);
			
			UpdateFireRatio();
			UpdateRatioByStage();
		}
				
		public Dictionary<string, float> MakeDictionary(string type = "fireRatio", string stageStr = ""){
			Dictionary<string, float> tmpDic = new Dictionary<string, float>();
			if(type == "fireRatio"){
				foreach(var stageName in eventList){
					tmpDic.Add(stageName, core.settings["HeadControlRatio"]["EventFireRatio"][stageName].AsFloat);
				}
			}else if(type == "ratioByStage"){
				foreach(var actionName in actionList){
					tmpDic.Add(actionName, core.settings["HeadControlRatio"][stageStr][actionName].AsFloat);
				}				
			}
			return tmpDic;
		}
		
		public List<string> ListStages(){
			return stageAndEventMap.ListStages();
		}
		
		public string[] ListActions(){
			return actionList;
		}
		
		public float GetAngle(string actionName, string axis){
			return angleMap[actionName].GetAngle(axis);
		}
		
		public float GetFireRatio(int stage, int count){
			return stageAndEventMap.GetFireRatio(stage, count);
		}
		
		public float GetFireRatio(string stage){
			return stageAndEventMap.GetFireRatio(stage);
		}
		
		public float GetRatioByStage(string stage, string actionName){
			return stageAndEventMap.GetRatioByStage(stage, actionName);
		}
		
		public void UpdateAngle(){
			foreach(var actionName in actionList){
				foreach(var axis in axisList){
					angleMap[actionName].SetAngle(axis, core.settings["HeadControlAngle"][actionName][axis].AsFloat);
				}
			}
		}
		
		public void UpdateFireRatio(){
			stageAndEventMap.InitFireRatio(MakeDictionary("fireRatio"));
		}
		
		public void UpdateRatioByStage(){
			foreach(var stageName in eventList){
				stageAndEventMap.InitActionAndRatioByStage(stageName, MakeDictionary("ratioByStage", stageStr:stageName));
			}
		}
		
		public void ApplySettings(){
			SetControllerAndProgress();
		}
		
		public void Maintenance(){

		}

		public void SetControllerAndProgress(){
			progress = new Dictionary<string, float>();
			lastActionTime = new Dictionary<string, float>();
			timing   = new Dictionary<string, CycleManager>();
			isActive = new Dictionary<string, bool>();
			isForce = new Dictionary<string, bool>();
			
			foreach(var key in actionList){
				isActive[key] = false;
				isForce[key] = false;
				timing.Add(key, new CycleManager());
				lastActionTime.Add(key, 0.0f);
			}
		}

		public bool IsFacingForward(){
			try{
				return (core.guess.GetFaceAngleHorizontal() == 3);
			}catch(Exception e){
				core.LoggerE(e);
				return false;
			}
		}
		
		public float Rand(string type = "small"){
			if(type == "big"){
				return UnityEngine.Random.Range(0.5f,3.0f);				
			}else{
				return UnityEngine.Random.Range(0.8f,1.2f);				
			}
		}

		public float GetHeatMulti(){
			float heatMulti = 1.0f;
			if(core.GetStageHeat() == "veryHot"){
				heatMulti = 0.9f;
			}else if(core.GetStageHeat() == "nymphomania"){
				heatMulti = 0.7f;
			}
			return heatMulti;
		}
		
		public bool ReadyNextAction(string type){
			bool result = true;
			if(isActive["Face Sideways 1"] 
				|| isActive["Face Sideways 2"] 
				|| isActive["Face Sideways 3"] 
				|| isActive["Turn Head Back 2"] 
				|| isActive["Glance"]){
					result = false;
			}
			float offsetRatio = UnityEngine.Random.Range(actionIntervalRatioMin, actionIntervalRatioMax) * GetHeatMulti();
			if(core.currentTime < lastActionTime[type] + actionInterval[type] * offsetRatio){
				result = false;
			}
			
			return result;
		}
		
		public void UpdateMotion(string type, bool bForce = false){
			try{
				if(!headMotionOnGlobal){
					return;
				}else if(core.trunk == null){
					return;
				}else{
					if(!bForce && ReadyNextAction(type) == false){
						return;
					}
					if(headMotionStopDuringPossessionOnGlobal && headControl.possessed){
						return;
					}
					core.DoDelayedAction(() => core.DoAddonManager2Function("MoveHandToMouthIfNeeded"), 0.05f);
					core.DoDelayedAction(() => core.DoAddonManager2Function("MoveHandToMouthIfNeeded"), 0.2f);
					float time = core.currentTime;
					float cFacingForward = IsFacingForward() ? 0.5f : 1.0f;
					float time1 = time + closingTimeOffset1[type] * UnityEngine.Random.Range(timeRandomMinRatio, timeRandomMaxRatio) * AdjustInterval(bForce) * cFacingForward * GetHeatMulti();
					float time2 = time1 + closingTimeOffset2[type] * UnityEngine.Random.Range(timeRandomMinRatio, timeRandomMaxRatio) * AdjustInterval(bForce) * cFacingForward * GetHeatMulti();
					
					lastActionTime[type] = time;
					timing[type].SetTimer(time, time1, time2);

					Transform mainT = headControl.transform;
//					Transform mainT = (core.isUnderControl) ? core.trunkHead.transform : headControl.transform;
					isForce[type] = bForce;
					initialQuat = mainT.rotation;
					if(isActive[type] == false){
						isActive[type] = true;
//						recoveryQuat = headControl.transform.rotation;
						recoveryQuat = mainT.rotation;
					}
					if(type == "Turn Head Back 1"){
						targetQuat =   GetBendingBackRotation(bForce, type, mainT);
					}else if(type == "Turn Head Back 2" || type == "Face Forward"){
						recoveryQuat = GetBendingBackStrongRotationInit(bForce, type, mainT);
						targetQuat =   GetBendingBackStrongRotationTarget(bForce, type, mainT);
					}else{
						if(type.Contains("Sideways") || type == "Glance"){
							recoveryQuat = GetFaceAwayRotation(bForce, type, mainT);
							targetQuat =   GetFaceAwayRotation(bForce, type, mainT);
						}
					}
				}
			}catch (Exception e){
				core.LoggerE(e);
			}
		}		
		
		public Quaternion GetBendingBackRotation(bool bForce, string type, Transform mainT){
			float strength  = angleMap[type].angleMap["right"] * Rand() * AdjustStrength(bForce);
			return 
				Quaternion.AngleAxis(strength, mainT.right) * 
				mainT.rotation;
		}
		
		public Quaternion GetBendingBackStrongRotationTarget(bool bForce, string type, Transform mainT){
			float strength  = angleMap[type].angleMap["right"] * Rand() * AdjustStrength(bForce);
			strength = (type == "Face Forward") ? 10f : strength;
			return 
				Quaternion.AngleAxis(strength, core.trunk.transform.right) * 
				core.trunk.transform.rotation;
		}
		
		public Quaternion GetBendingBackStrongRotationInit(bool bForce, string type, Transform mainT){
			float strengthBase = 10f;
			float strength  = strengthBase * Rand() * AdjustStrength(bForce);
			strength = (type == "Face Forward") ? 10f : strength;
			return
				Quaternion.AngleAxis(strength, core.trunk.transform.right) * 
				core.trunk.transform.rotation;
		}
		
		public Quaternion GetFaceAwayRotation(bool bForce, string type, Transform mainT){
			int angleH = core.guess.GetFaceAngleHorizontal();
			int angleV = core.guess.GetFaceAngleVertical();
			
			float sign = 1.0f;
			float degreeUp       =  angleMap[type].angleMap["up"]       * Rand();
			float degreeForward  =  angleMap[type].angleMap["forward"]  * Rand();
			float degreeForward2 =  angleMap[type].angleMap["forward2"] * Rand();
			float degreeRight    =  angleMap[type].angleMap["right"]    * Rand();
			float degreeRight2   =  angleMap[type].angleMap["right2"]   * Rand();
			
			if(angleH == 3){ //"C"
				sign = (UnityEngine.Random.Range(0.0f,1.0f) > 0.5) ? 1.0f : -1.0f;
			}else{
				sign = (angleH > 3) ? -1.0f : 1.0f; // L:1,2 C:3 R:4,5
			}
			string arrow = (sign == -1.0f) ? "Left" : "Right";
			
			Quaternion rot =
					Quaternion.AngleAxis(degreeRight2, core.trunk.transform.right) *
					Quaternion.AngleAxis(degreeForward2 * sign, core.trunk.transform.forward) *
					Quaternion.AngleAxis(degreeForward * -1.0f * sign, core.trunk.transform.forward) *
					Quaternion.AngleAxis(degreeUp * sign, core.trunk.transform.up) *
					Quaternion.AngleAxis(degreeRight, core.trunk.transform.right) *
					core.trunk.transform.rotation;
			return rot;
		}

		public void CalcProgress(float time, string type){
			try{
				progress[type] = timing[type].GetProgress(time);
				
				if(timing[type].GetState(time) == 0){
					isActive[type] = false;
				}
			}catch (Exception e){
				core.LoggerE(e);
			}
		}
		
		public void ChangeJoint(float time, string type){
			if(!timing.ContainsKey(type)){
				core.Info("ERROR: ChangeJoint: key:" + type + " not found.");
			}
			int state = timing[type].GetState(time);
			if(state == 0 || (state == 2 && motionType[type] == "OneWay")){
					
			}else{
				if(state == 1){
					headControl.transform.rotation = 
						Quaternion.Slerp(initialQuat, targetQuat, progress[type]);						
					core.trunkHead.transform.rotation = headControl.transform.rotation;
				}else if(state == 2){
					headControl.transform.rotation = 
						Quaternion.Slerp(recoveryQuat, targetQuat, progress[type]);
					core.trunkHead.transform.rotation = headControl.transform.rotation;
				}
			}
		}

		public float AdjustInterval(bool bForce){
			float stageCoefficient = (bForce) ? 5.0f * 0.1f : (float)core.GetStage() * 0.1f;
			return (1 / (1 + stageCoefficient));
		}
		
		public float AdjustStrength(bool bForce){
			float stageCoefficient = (bForce) ? 5.0f * 0.1f : (float)core.GetStage() * 0.2f;
			return 1 + stageCoefficient;
		}
		
		public void Manage(float time){
			try{
				if(time == lastTime){
					time += 0.001f;
				}else if(time < lastTime){
					time = lastTime + 0.001f;
				}
				lastTime = time;

				int stage = core.GetStage();
				foreach(var key in actionList){
					if(isActive[key]){
						if(isForce[key] || stage > 0){
							CalcProgress(time, key);
							ChangeJoint(time, key);
						}else{
							isActive[key] = false;
						}
					}
				}
			}catch (Exception e){
				core.LoggerE(e);
			}
		}
		
		public void PickAction(string stage, bool bIgnoringFiringRatio = false, bool bForce = false){
			float rnd = UnityEngine.Random.Range(0f, 1f);
			float ratio = GetFireRatio(stage);
			bool flag = (bIgnoringFiringRatio || rnd < ratio);
			if(flag){
				string actionName = stageAndEventMap.PickAction(stage, true);
				UpdateMotion(actionName, bForce);
			}
		}
		
		public void PickAction(bool bIgnoringFiringRatio = false, bool bForce = false){
			PickAction(core.GetStage(), (int)core.customUi.jsOrgasmCount.val, bIgnoringFiringRatio);
		}
		
		public void PickAction(int stage, int count, bool bIgnoringFiringRatio = false, bool bForce = false){
			string name = Utility.ResolveStageName(stage, count);
			PickAction(name, bIgnoringFiringRatio);
		}
		
		public void HeadMotion(string type, bool bForce = false){
			switch(type){
				case "Turn Head Back 1":
					UpdateMotion("Turn Head Back 1", bForce); break;
				case "Turn Head Back 2":
					UpdateMotion("Turn Head Back 2", bForce); break;
				case "Face Forward":
					UpdateMotion("Face Forward", bForce); break;
				case "Face Sideways 1":
					UpdateMotion("Face Sideways 1", bForce); break;
				case "Face Sideways 2":
					UpdateMotion("Face Sideways 2", bForce); break;
				case "Face Sideways 3":
					UpdateMotion("Face Sideways 3", bForce); break;
				case "Glance":
					UpdateMotion("Glance", bForce); break;
			}
		}
		
		
		public void OnDestroy(){
			RemoveTiming();
			progress.Clear();
			isActive.Clear();
			isForce.Clear();
			lastActionTime.Clear();
		
			motionType.Clear();
			actionInterval.Clear();
			closingTimeOffset1.Clear();
			closingTimeOffset2.Clear();
			
			foreach(KeyValuePair<string, AxisAndAngles> kvp in angleMap){
				kvp.Value.OnDestroy();
			}
			angleMap.Clear();
			stageAndEventMap.OnDestroy();
		}
		
		public void RemoveTiming(){
			List<string> keys = new List<string>(timing.Keys);
			foreach(var key in keys)
			{
				if(timing[key] != null) timing[key] = null;
			}
			timing = null;			
		}
	}
}