using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;

namespace Cotyounoyume{
	public class ParentHolder
	{
		public EBATCore core;
		Quaternion diffRotation = Quaternion.identity;
		Vector3    diffPosition = new Vector3(0f, 0f, 0f);
		Transform parent = null;
		Transform child = null;
		string name = "noname";
		bool atOnce = true;
		
		public ParentHolder(EBATCore core){
			this.core = core;			
		}
		
		public void SetParent(Transform child, Transform parent, string name = "noname"){
			this.parent = parent;
			this.child = child;
			this.name = name;
			if(atOnce){
				atOnce = false;
				diffRotation = child.rotation * Quaternion.Inverse(parent.rotation);
				diffPosition = parent.position - child.position;
			}
		}
		
		public float GetAngle(){
			return Quaternion.Angle(parent.rotation, child.rotation);
		}
		
		public float GetDistance(){
			return Vector3.Distance(child.position, parent.position);
		}
		
		public void Debug(){
			core.Info("name:" + name + ", child:" + child + ", parent:" + parent);
			core.Info("  child.rotation :" + child.rotation.ToString("F3"));
			core.Info("  parent.rotation:" + parent.rotation.ToString("F3"));
			core.Info("  diffRot.rotation:" + diffRotation.ToString("F3"));
			Quaternion rot = diffRotation * parent.rotation;
			core.Info("  recoveryRotation:" + rot.ToString("F3"));
			core.Info("  angle :" + GetAngle());
			core.Info("  ----------");
			core.Info("  child.position :" + child.position.ToString("F3"));
			core.Info("  parent.position:" + parent.position.ToString("F3"));
			core.Info("  diffRot.position:" + diffPosition.ToString("F3"));
			Vector3 pos = parent.position - diffPosition;
			core.Info("  recoveryPosition:" + pos.ToString("F3"));
			float dist = Vector3.Distance(child.position, parent.position);
			core.Info("  dist  :" + GetDistance());
		}
		
		public void Recovery(bool isForce = false){
			if(parent == null || child == null){
				return;
			}
			if(GetAngle() > 10f){
				child.rotation = diffRotation * parent.rotation;
				child.position = parent.position - diffPosition;
			}
		}
		
		public void OnDestroy(){
			child = null;
			parent = null;
		}
	}
	
	public class ParentAngleChecker
	{
		public EBATCore core;
		public Dictionary<Transform, ParentHolder> parentHolderMap;
		
		public ParentAngleChecker(EBATCore core){
			this.core = core;
			parentHolderMap = new Dictionary<Transform, ParentHolder>();
		}
		
		public void RegisterPair(Transform child, Transform parent, string name){
			if(parentHolderMap.ContainsKey(child)){
				parentHolderMap[child] = new ParentHolder(core);
			}else{
				parentHolderMap.Add(child, new ParentHolder(core));
			}
			parentHolderMap[child].SetParent(child, parent, name);
		}
		
		public void CheckChilds(){
			foreach(KeyValuePair<Transform, ParentHolder> kvp in parentHolderMap){
				parentHolderMap[kvp.Key].Debug();
			}
		}
		
		public void Recovery(){
			foreach(KeyValuePair<Transform, ParentHolder> kvp in parentHolderMap){
				parentHolderMap[kvp.Key].Recovery();
			}
		}
		
		public void OnDestroy(){
			List<Transform> queue = new List<Transform>();
			foreach(KeyValuePair<Transform, ParentHolder> kvp in parentHolderMap){
				queue.Add(kvp.Key);
			}
			
			foreach(var item in queue){
				parentHolderMap[item].OnDestroy();
				parentHolderMap.Remove(item);
			}
			
			parentHolderMap.Clear();
			queue.Clear();
		}
	}
}
