#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
using MVR.FileManagementSecure;
#endregion

//namespace MVR.FileManagementSecure{
namespace Cotyounoyume{
	public class SoundStorage{
		public Dictionary<string, SoundPersonality> soundPersonality;
		public string[] types;
		public string baseDir;
		public bool isLoadFirst = false;
		public string lastSelection = "";
		
		public SoundStorage(string[] types, string baseDir){
			this.types = types;
			this.baseDir = baseDir;
			this.soundPersonality = new Dictionary<string, SoundPersonality>();
			CleaningAudioClip();
		}
		
		public void SetPersonality(string personality){
			if(lastSelection != personality){
				if(lastSelection != ""){
					soundPersonality[lastSelection].Destroy();
					soundPersonality.Remove(lastSelection);
				}
			}
			
			if(soundPersonality.ContainsKey(personality)){
				soundPersonality[personality].Destroy();
				soundPersonality.Remove(personality);
			}
			soundPersonality.Add(personality, new SoundPersonality(personality, types, baseDir));
			lastSelection = personality;
		}
		
		public void SetLoadSetting(bool isLoadFirst){
			this.isLoadFirst = isLoadFirst;
		}
		
		public NamedAudioClip PickRandomSound(string personality, string type){
			if(soundPersonality.ContainsKey(personality) == false) return null;
			if(soundPersonality[personality].soundType.ContainsKey(type) == false) return null;
			return soundPersonality[personality].soundType[type].PickRandomSound();
		}
		
		public void Destroy(){
			foreach(var kvp in soundPersonality){
				soundPersonality[kvp.Key].Destroy();
			}
			soundPersonality.Clear();
			soundPersonality = null;
			types = null;
		}
		
		public void CleaningAudioClip(){
			try{
				JSONClass jsc;
				jsc = URLAudioClipManager.singleton.GetJSON();
//				SuperController.LogMessage(jsc.ToString());
				List<string> uids = new List<string>();
				
				JSONNode jsn = jsc["clips"];
				foreach(var item in jsn.Childs){
					string name = item["url"];
					if(name.Contains("ExpressionBlushingAndTears")){
						uids.Add(item["displayName"]);
					}
				}
				foreach(var uid in uids){
//					SuperController.LogMessage("uid: " + uid);
					NamedAudioClip nac = URLAudioClipManager.singleton.GetClip(uid);
					URLAudioClipManager.singleton.RemoveClip(nac);
				}
				
//				jsc = URLAudioClipManager.singleton.GetJSON();
//				SuperController.LogMessage(jsc.ToString());
			}catch(Exception e){
				Utility.LoggerE(e);				
			}
		}
	}
	
	public class SoundPersonality{
		public string[] types;
		public Dictionary<string, SoundType> soundType;
		
		public SoundPersonality(string personality, string[] types, string baseDir){
			this.types = types;
			
			soundType = new Dictionary<string, SoundType>();
			foreach(var type in this.types){
				soundType.Add(type, new SoundType(personality, type, baseDir));
			}
		}
		
		public void Destroy(){
			foreach(var type in types){
				soundType[type].Destroy();
			}
			soundType.Clear();
			soundType = null;
			types = null;
		}
	}
		
	public class SoundType{
		public List<NamedAudioClip> soundList = new List<NamedAudioClip>();
		
		public SoundType(string personality, string type, string baseDir){
			try{
				string dirPath = baseDir + "\\" + personality + "\\" + type;
				if(FileManagerSecure.DirectoryExists(dirPath) == false) return;

				string[] files = FileManagerSecure.GetFiles(dirPath, "*.wav");
				foreach(var file in files){
					NamedAudioClip nac = LoadSound(file);
					if(nac != null) soundList.Add(nac);
				}
			}catch(Exception e){
				SuperController.LogMessage("*** Exception:" + e);
			}
		}
		
		public NamedAudioClip PickRandomSound(){
			if(soundList.Any() == false) return null;
			int rand = UnityEngine.Random.Range(0, soundList.Count);
			return soundList[rand];
		}

		protected NamedAudioClip LoadSound(string path){
			try{
				string localPath = SuperController.singleton.NormalizeLoadPath(path);
				NamedAudioClip existing = URLAudioClipManager.singleton.GetClip(localPath);
				if (existing != null) return existing;

				URLAudioClip clip = URLAudioClipManager.singleton.QueueClip(SuperController.singleton.NormalizeMediaPath(path));
				if (clip == null) return null;
				
				NamedAudioClip nac = URLAudioClipManager.singleton.GetClip(clip.uid);
				if (nac == null) return null;
				
				return nac;
			}catch (Exception e){
				SuperController.LogMessage("*** Exception:" + e);
				return null;
			}
		}
		
		public void Destroy(){
			UnloadAllClip();
			soundList.Clear();
			soundList = null;
		}
		
		public void UnloadAllClip(){
			foreach(var nac in soundList){
				URLAudioClipManager.singleton.RemoveClip(nac);
			}
		}
	}
	
	public class Sound
	{
		#region common or UI Properties
		public Atom atom;
		public EBATCore core;
		public bool disableLog;
		
		public string[] types = {"breath", "moan", "climax"};
		public SoundStorage ss;

		public bool soundOnGlobal = true;
		public AudioSourceControl headAudioSource;
		
		public string defaultName = "default";
		public string sampleName = "sample";
		public string soundDirRoot = "Custom\\Sounds\\ExpressionBlushingAndTears";

		public List<string> soundTypes;
		public string soundTypeDefault = "default";
		public float lastPlayTime;
		public Dictionary<string, float> playbackRate = new Dictionary<string, float>(){
			{"breath", 1.0f},
			{"moan", 1.0f},
			{"climax", 1.0f},
		};

		public bool isLoadFirst = false;
		#endregion
		
		public Sound(EBATCore core){
			try{
				atom = core.atom;
				this.core = core;
				Maintenance();
			}catch (Exception e){
				core.LoggerE(e);
			}
		}
		
		public void LoadSettings(){
			try{
				soundOnGlobal = core.settings["General"]["Sound On"].AsBool;
				headAudioSource = atom.GetStorableByID("HeadAudioSource") as AudioSourceControl;
				headAudioSource.pitch = core.settings["General"]["Sound Pitch"].AsFloat;
				headAudioSource.volume = core.settings["General"]["Sound Volume"].AsFloat;
				
				soundTypeDefault = core.settings["General"]["Sound Type(Default)"];

				SetFrequency("breath", core.settings["General"]["Playback Rate: Breath"].AsFloat);
				SetFrequency("moan",   core.settings["General"]["Playback Rate: Moan"].AsFloat);
				SetFrequency("climax", core.settings["General"]["Playback Rate: Climax"].AsFloat);

				RefreshSoundDirectory();
			}catch (Exception e){
				core.LoggerE(e);
			}
		}
		
		public void ApplySettings(){
			try{
				if(core.customUi.jsSoundTypePersonal.val == "" || core.customUi.jsSoundTypePersonal.val == null){
					core.customUi.jsSoundTypePersonal.val = soundTypeDefault;
				}
				lastPlayTime = 0f;
				SetupSounds();				
			}catch (Exception e){
				core.LoggerE(e);
			}
		}

		public void RefreshSoundDirectory(){
			try{
				soundTypes = new List<string>();
				
				string[] dirs = FileManagerSecure.GetDirectories(soundDirRoot);
				var res = dirs.Where(str => CheckSoundDir(str)).Select(str => str);
				foreach(var str in res){
					string[] strAry = str.ToString().Split('\\');
					var last = strAry.Last();
					soundTypes.Add(last);
				}
			}catch(Exception e){
				core.LoggerE(e);
			}
		}
		
		public bool CheckSoundDir(string path){
			if(FileManagerSecure.DirectoryExists(path) == false) return false;
			string[] dirs = {"breath", "moan", "climax"};
			foreach(var dir in dirs){
				string tmpPath = path + "\\" + dir;
				if(FileManagerSecure.DirectoryExists(tmpPath) == false) return false;
				if(FileManagerSecure.GetFiles(tmpPath, "*.wav").Any() == false) return false;
			}
			return true;
		}
		
		public void Maintenance(){
			try{
				headAudioSource = atom.GetStorableByID("HeadAudioSource") as AudioSourceControl;
				if(FileManagerSecure.DirectoryExists(soundDirRoot + "\\" + defaultName) == false){
					CopySound();
				}
			}catch (Exception e){
				core.LoggerE(e);
			}
		}

		public void UpdateSound(){
			RefreshSoundDirectory();
			SetupSounds();
		}
				
		public void SetupSounds(){
			try{
				if(ss != null){
					ss.Destroy();
				}
				ss = new SoundStorage(types, soundDirRoot);
				ss.SetLoadSetting(isLoadFirst);
				
				if(isLoadFirst){
					foreach(var personality in soundTypes){
						ss.SetPersonality(personality);
					}
				}else{
					ss.SetPersonality(core.customUi.jsSoundTypePersonal.val);
				}
			}catch (Exception e){
				core.LoggerE(e);
			}			
		}
		
		public void CopySound(){
			try{
				string srcDirBase1 = core.myPath.Replace("/", "\\") + "\\" + "Sounds\\default";
				string srcDirBase2 = core.myPath.Replace("/", "\\") + "\\" + "Sounds\\sample";
				string dstDirBase1 = soundDirRoot + "\\" + defaultName;
				string dstDirBase2 = soundDirRoot + "\\" + sampleName;
				CopySoundBase(srcDirBase1, dstDirBase1);
				CopySoundBase(srcDirBase2, dstDirBase2);
				SuperController.LogMessage("The sound files has been copied.\nFolder: " + dstDirBase1 + ", " + dstDirBase2);
				SuperController.LogMessage("(This message is displayed only the first time you use this plugin.");
				SuperController.LogMessage("This message will not be displayed next time.)");
			}catch (Exception e){
				core.LoggerE(e);
			}
		}

		public void CopySoundBase(string src, string dst){
			string[] targetDirs = {"breath", "moan", "climax"};
			foreach (string dir in targetDirs){
				string srcDir = src + "\\" + dir;
				string dstDir = dst + "\\" + dir;
				FileManagerSecure.CreateDirectory(dstDir);
				string[] srcFiles = FileManagerSecure.GetFiles(srcDir);
				for(int i = 0; i < srcFiles.Length; i++){
					int dirLength = srcDir.Length;
					string filename = srcFiles[i].Remove(0, dirLength+1);
					string dstFile = dstDir + "\\" + filename;
					FileManagerSecure.CopyFile(srcFiles[i], dstFile);
				}
			}			
		}
		
		public void SetFrequency(string type, float value){
			// Utility.Debug($"type:{type}, value:{value}");
			if(value == 0f) value = 1.0f;
			if(!playbackRate.ContainsKey(type)) playbackRate.Add(type, value);
			else playbackRate[type] = value;
		}

		public float GetFrequency(string type){
			if(!playbackRate.ContainsKey(type)) return 1.0f;
			return playbackRate[type];
		}

		public void PlaySound(string type, float time, bool force = false){
			try{
				if(!force && soundOnGlobal == false) return;
				if(!force && core.vammoanLinkOnGlobal) return;
				if(!force && type != "climax" && time < lastPlayTime) return;

				if(!playbackRate.ContainsKey(type)) return;
				var rand = UnityEngine.Random.Range(0f, 1f);
				if(!force && rand > playbackRate[type]) return;

				NamedAudioClip nac = ss.PickRandomSound(core.customUi.jsSoundTypePersonal.val, type);
				if(nac == null || nac.sourceClip == null){
					return;
				}
				float duration = nac.sourceClip.length;
				lastPlayTime = time + duration;
				headAudioSource.PlayNow(nac);
				core.expression2.OpenMouth(duration);
			}catch (Exception e){
				core.LoggerE(e);
			}
		}
		
		public void Update(float time){
			try{
			}catch (Exception e){
				core.LoggerE(e);
			}
		}
		
		public void Manage(float time){
			try{
			}catch (Exception e){
				core.LoggerE(e);
			}
		}
		
		public void OnDestroy(){
			ss.Destroy();
		}
	}
}