#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
#endregion

namespace Cotyounoyume{
	public class VariableTriggerBackup
	{
		public List<VariableTrigger> vts;
		public List<JSONClass> jscs;
		public GenerateDAZMorphsControlUI morphControl;
		public List<DAZMorph> morphs;
		public Atom atom;
		
		public VariableTriggerBackup(Dictionary<VariableTrigger, JSONClass> dic, Atom atom){
			this.atom = atom;
			JSONStorable geometry = atom.GetStorableByID("geometry");
			DAZCharacterSelector character = geometry as DAZCharacterSelector;
			morphControl = character.morphsControlUI;

			vts = new List<VariableTrigger>();
			jscs = new List<JSONClass>();
			
			foreach(KeyValuePair<VariableTrigger, JSONClass> kvp in dic){
				vts.Add(kvp.Key);
				JSONClass newJsc = CopyJSONClass(kvp.Value);
				jscs.Add(newJsc);
			}
		}

		public JSONClass CopyJSONClass(JSONClass jsc){
			string strCurrent = jsc.ToString();
			JSONNode jsn = JSONNode.Parse(strCurrent);
			JSONClass newJsc = jsn.AsObject;
			string strNew = newJsc.ToString();
			return newJsc;
		}
		
		public void ClearTargetMorphs(){
			try{
				for(int i = 0; i < vts.Count; i++){
					JSONClass newJsc = CopyJSONClass(jscs[i]);
					
					for(int j = 0; j < jscs[i]["transitionActions"].Count; j++){
						string atomName = jscs[i]["transitionActions"][j]["receiverAtom"];
						string receiveTargetName = jscs[i]["transitionActions"][j]["receiverTargetName"];
						string targetAtomName = atom.uid;
						if(atomName == atom.uid){
							DAZMorph morph = morphControl.GetMorphByDisplayName(receiveTargetName);
							if(morph != null){
								newJsc["transitionActions"][j]["receiverTargetName"] = "";
								morph.SetDefaultValue();
							}
						}
					}
					vts[i].trigger.RestoreFromJSON(newJsc);
				}
			}catch(Exception e){
				Utility.LoggerE(e);
			}
		}
		
		public void Recover(){
			for(int i = 0; i < vts.Count; i++){
				vts[i].trigger.RestoreFromJSON(jscs[i]);
			}
		}
	}

	public class CollisionTriggerBackup
	{
		public List<CollisionTrigger> cts;
		public List<JSONClass> jscs;
		public GenerateDAZMorphsControlUI morphControl;
		public List<DAZMorph> morphs;
		public Atom atom;
		
		public CollisionTriggerBackup(Dictionary<CollisionTrigger, JSONClass> dic, Atom atom){
			this.atom = atom;
			JSONStorable geometry = atom.GetStorableByID("geometry");
			DAZCharacterSelector character = geometry as DAZCharacterSelector;
			morphControl = character.morphsControlUI;

			cts = new List<CollisionTrigger>();
			jscs = new List<JSONClass>();
			
			foreach(KeyValuePair<CollisionTrigger, JSONClass> kvp in dic){
				cts.Add(kvp.Key);
				JSONClass newJsc = CopyJSONClass(kvp.Value);
				jscs.Add(newJsc);
			}
		}

		public JSONClass CopyJSONClass(JSONClass jsc){
			string strCurrent = jsc.ToString();
			JSONNode jsn = JSONNode.Parse(strCurrent);
			JSONClass newJsc = jsn.AsObject;
			string strNew = newJsc.ToString();
				
			return newJsc;
		}
		
		public void ClearTargetMorphs(){
			try{
				for(int i = 0; i < cts.Count; i++){
					JSONClass newJsc = CopyJSONClass(jscs[i]);
					
					for(int j = 0; j < jscs[i]["transitionActions"].Count; j++){
						string atomName = jscs[i]["transitionActions"][j]["receiverAtom"];
						string receiveTargetName = jscs[i]["transitionActions"][j]["receiverTargetName"];
						string targetAtomName = atom.uid;
						if(atomName == atom.uid){
							DAZMorph morph = morphControl.GetMorphByDisplayName(receiveTargetName);
							if(morph != null){
								newJsc["transitionActions"][j]["receiverTargetName"] = "";
								morph.SetDefaultValue();
							}
						}
					}
					cts[i].trigger.RestoreFromJSON(newJsc);
				}
			}catch(Exception e){
				Utility.LoggerE(e);
			}
		}
		
		public void Recover(){
			for(int i = 0; i < cts.Count; i++){
				cts[i].trigger.RestoreFromJSON(jscs[i]);
			}
		}
	}

	public class ReceiverBackup
	{
		public List<JSONNode> jsns;
		public List<DAZMorph> morphs;
		public GenerateDAZMorphsControlUI morphControl;
		
		public ReceiverBackup(List<JSONNode> jsns, Atom atom){
			this.jsns = jsns;
			JSONStorable geometry = atom.GetStorableByID("geometry");
			DAZCharacterSelector character = geometry as DAZCharacterSelector;
			morphControl = character.morphsControlUI;
			
			List<DAZMorph> morphs = new List<DAZMorph>();
			foreach(JSONNode jsn in this.jsns){
				string morphDisplayName = jsn["receiverTargetName"];
				DAZMorph morph = morphControl.GetMorphByDisplayName(morphDisplayName);
				morphs.Add(morph);
			}
		}
		
		public void ClearTargetMorphs(){
			foreach(JSONNode jsn in jsns){
				string morphDisplayName = jsn["receiverTargetName"];
				DAZMorph morph = morphControl.GetMorphByDisplayName(morphDisplayName);
				if(morph != null){
					jsn["receiverTargetName"] = null;
					morph.SetDefaultValue();
				}
			}			
		}
		
		public void TestTargetMorphs(){
			foreach(JSONNode jsn in jsns){
				Utility.Info("jsn[receiverAtom]: " + jsn["receiverAtom"]);
				Utility.Info("jsn[receiver]: " + jsn["receiver"]);
				Utility.Info("jsn[receiverTargetName]: " + jsn["receiverTargetName"]);
			}
		}
		
		public void RecoverTargetMorphs(){
			for(int i = 0; i < jsns.Count; i++){
				if(morphs[i] != null){
					jsns[i]["receiverTargetName"] = morphs[i].displayName; 
				}
			}
		}
	}
}