using System;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using Extensions;

namespace Cotyounoyume
{
    public class TriggersForExternalLinkage
    {
		#region Define vars
		public EBATCore core;
		public Dictionary<string, JSONStorableAction> jsas;
		#endregion
		
		#region common
        public TriggersForExternalLinkage(EBATCore core)
        {
			this.core = core;
			jsas = new Dictionary<string, JSONStorableAction>();
        }
		
		public void SetJSActions(){
			SetAndRegisterAction("▉▉▉▉▉▉▉▉ Start of Trigger for External Linkage ▉▉▉▉▉▉▉▉", Dummy);
			SetAndRegisterAction("ToggleMinimize"        , ToggleMinimize);
			SetAndRegisterAction("Suspend"               , Suspend);
			SetAndRegisterAction("Resume"                , Resume);
			SetAndRegisterAction("ToggleCompact"         , ToggleCompact);
			SetAndRegisterAction("CollapseAndExpand"     , CollapseAndExpand);
			SetAndRegisterAction("ToggleBarOnly"         , ToggleBarOnly);
			SetAndRegisterAction("CloseUI"               , CloseUI);
			SetAndRegisterAction("LookFront"             , LookFront);
			SetAndRegisterAction("ReloadPlugins"         , ReloadPlugins);
			SetAndRegisterAction("ReloadThisPluginOnly"  , ReloadThisPluginOnly);
			SetAndRegisterAction("ToggleSize"            , ToggleSize);
			SetAndRegisterAction("InstantCUAScaleLow"    , InstantCUAScaleLow);
			SetAndRegisterAction("InstantCUAScaleMiddle" , InstantCUAScaleMiddle);
			SetAndRegisterAction("InstantCUAScaleHigh"   , InstantCUAScaleHigh);

			if(core.orgasmModeOn){
				SetAndRegisterAction("MakeHerOrgasm", MakeHerOrgasm);
				SetAndRegisterAction("StopHerOrgasm", StopHerOrgasm);
				SetAndRegisterAction("SingleWave", SingleWave);
			}else{
				SetAndRegisterAction("MaximizeEcstasy", MakeHerOrgasm);
			}
			SetAndRegisterAction("CalmHerDown"  , CalmHerDown);
			SetAndRegisterAction("ResetExpression"  , ResetExpression);
			SetAndRegisterAction("StageUp"      , StageUp);
			SetAndRegisterAction("StageDown"    , StageDown);
			
			SetAndRegisterAction("FaceForward", FaceForward);
			SetAndRegisterAction("FaceSidewaysT", FaceSidewaysT);
			SetAndRegisterAction("FaceSidewaysS", FaceSidewaysS);
			SetAndRegisterAction("FaceSidewaysM", FaceSidewaysM);
			SetAndRegisterAction("FaceSidewaysL", FaceSidewaysL);
			SetAndRegisterAction("TurnHeadBackT", TurnHeadBackT);
			SetAndRegisterAction("TurnHeadBackL", TurnHeadBackL);
			
			SetAndRegisterAction("ToggleExpression"			,ToggleExpression);
			SetAndRegisterAction("ToggleAutoExpression"		,ToggleAutoExpression);
			SetAndRegisterAction("ToggleAutoBlink"			,ToggleAutoBlink);
			SetAndRegisterAction("ToggleBlushing"			,ToggleBlushing);
			SetAndRegisterAction("ToggleTears"				,ToggleTears);
			SetAndRegisterAction("ToggleDrool"				,ToggleDrool);
			SetAndRegisterAction("ToggleEyesControl"	   	,ToggleEyesControl);
			SetAndRegisterAction("ToggleEyesRollup"       	,ToggleEyesRollup);
			SetAndRegisterAction("ToggleSound"				,ToggleSound);
			SetAndRegisterAction("ToggleSquirt"				,ToggleSquirt);
			SetAndRegisterAction("ToggleNippleErection"		,ToggleNippleErection);
			SetAndRegisterAction("ToggleClitorisErection"	,ToggleClitorisErection);
			SetAndRegisterAction("ToggleHeadMotion"			,ToggleHeadMotion);
			SetAndRegisterAction("ToggleShiveringBone"		,ToggleShiveringBone);
			SetAndRegisterAction("ToggleShiveringMorph"		,ToggleShiveringMorph);
			// SetAndRegisterAction("ToggleSensitiveTrigger"	,ToggleSensitiveTrigger);
			SetAndRegisterAction("TogglePenisSupport"		,TogglePenisSupport);
			
			SetAndRegisterAction("ToggleAutoStopOtherPlugins"			,ToggleAutoStopOtherPlugins);
			SetAndRegisterAction("ToggleStopMorphControlByNonPlugins"	,ToggleStopMorphControlByNonPlugins);
			SetAndRegisterAction("ToggleStopHeadControlByNonPlugins"	,ToggleStopHeadControlByNonPlugins);
			SetAndRegisterAction("ToggleStopBodyControlByNonPlugins"	,ToggleStopBodyControlByNonPlugins);

			if(core.addon1ModeOn){

				SetAndRegisterAction("UndressRemove"                ,UndressRemove);
				SetAndRegisterAction("AltSkinWrap"                  ,AltSkinWrap);
				SetAndRegisterAction("UndressRecover"               ,UndressRecover);
				SetAndRegisterAction("ExposeBreasts"                ,ExposeBreasts);
				SetAndRegisterAction("TearOffShirt"                 ,TearOffShirt);
				SetAndRegisterAction("PullUpSkirt"                  ,PullUpSkirt);
				SetAndRegisterAction("StopExposingAndPullingUp"		,StopExposingAndPullingUp);
				
				SetAndRegisterAction("ToggleHoldClothesContainingNGWords" ,HoldClothesContainingNGWords);
				SetAndRegisterAction("ToggleHoldClothesWithClothSim"      ,HoldClothesWithClothSim);
				SetAndRegisterAction("ToggleHoldClothesWithAlterSkinWrap" ,HoldClothesWithAlterSkinWrap);
				SetAndRegisterAction("ToggleAlterSkinWrapToClothSim"      ,AlterSkinWraptoClothSim);

				foreach(var part in core.addonManager1.worldUIAddon1.parts){
					Action act = () => core.addonManager1.worldUIAddon1.UndressPartialCommon(part);
					SetAndRegisterAction("UndressPartial" + part.ToUpperLineHead(), act);
				}
				SetAndRegisterAction("UndressOverall", core.addonManager1.worldUIAddon1.OnUndressOverallClick);
				SetAndRegisterAction("UndressPartialOutfit", core.addonManager1.worldUIAddon1.OnUndressOutfitClick);
				SetAndRegisterAction("UndressPartialUnderwears", core.addonManager1.worldUIAddon1.OnUndressUnderwearClick);
				SetAndRegisterAction("UndressPartialMisc", core.addonManager1.worldUIAddon1.OnUndressMiscClick);
			}
			
			if(core.addon2ModeOn){
				SetAndRegisterAction("BackupPose"				,BackupPose);
				SetAndRegisterAction("RecoverPose"				,RecoverPose);
				SetAndRegisterAction("HoldMouth"				,HoldMouth);
				SetAndRegisterAction("DuctTape"				    ,DuctTape);
				SetAndRegisterAction("StopHoldMouth"			,StopHoldMouth);
				SetAndRegisterAction("HoldArms"					,HoldArms);
				SetAndRegisterAction("HoldArmL"					,HoldArmL);
				SetAndRegisterAction("HoldArmR"					,HoldArmR);
				SetAndRegisterAction("HoldArmsSide"				,HoldArmsSide);
				SetAndRegisterAction("HoldArmLSide"				,HoldArmLSide);
				SetAndRegisterAction("HoldArmRSide"				,HoldArmRSide);
				SetAndRegisterAction("HoldArmsBack"				,HoldArmsBack);
				SetAndRegisterAction("HoldArmLBack"				,HoldArmLBack);
				SetAndRegisterAction("HoldArmRBack"				,HoldArmRBack);
				SetAndRegisterAction("HoldArmsUp"				,HoldArmsUp);
				SetAndRegisterAction("HoldArmLUp"				,HoldArmLUp);
				SetAndRegisterAction("HoldArmRUp"				,HoldArmRUp);
				SetAndRegisterAction("StopHoldArms"				,StopHoldArms);
				SetAndRegisterAction("StopHoldArmL"				,StopHoldArmL);
				SetAndRegisterAction("StopHoldArmR"				,StopHoldArmR);
				SetAndRegisterAction("ToggleArmBinderByHand"	,ToggleArmBinderByHand);
				SetAndRegisterAction("ToggleArmBinderByBeam"	,ToggleArmBinderByBeam);
				SetAndRegisterAction("HoldLegs"					,HoldLegs);
				SetAndRegisterAction("SpreadLegsUp"				,SpreadLegsUp);
				SetAndRegisterAction("SpreadLegsWidely"			,SpreadLegsWidely);
				SetAndRegisterAction("SpreadLegsSlightly"		,SpreadLegsSlightly);
				SetAndRegisterAction("LiftUpKnee"				,LiftUpKnee);
				SetAndRegisterAction("LiftUpKnees"				,LiftUpKnees);
				SetAndRegisterAction("SplitLegs"				,SplitLegs);
				SetAndRegisterAction("StopHoldLegs"				,StopHoldLegs);
				SetAndRegisterAction("ToggleLegBinderByHand"	,ToggleLegBinderByHand);
				SetAndRegisterAction("ToggleLegBinderByBeam"	,ToggleLegBinderByBeam);
				SetAndRegisterAction("SummonHandR"				,SummonHandR);
				SetAndRegisterAction("SummonHandL"				,SummonHandL);
				SetAndRegisterAction("SummonTentacle1"			,SummonTentacle1);
				SetAndRegisterAction("SummonMouth"				,SummonMouth);
				SetAndRegisterAction("SummonClitSucker"			,SummonClitSucker);
				SetAndRegisterAction("SummonTwinDildo"			,SummonTwinDildo);
				SetAndRegisterAction("UnsummonAll"				,UnsummonAll);
				SetAndRegisterAction("ToggleCUACollision"		,ToggleCUACollision);

				SetAndRegisterAction("ToggleCoveringActions"	,ToggleCoveringActions);
				SetAndRegisterAction("ToggleCoverWithHands"		,ToggleCoverWithHands);
				SetAndRegisterAction("ToggleCloseLegs"			,ToggleCloseLegs);
				SetAndRegisterAction("ToggleCoverMouth"			,ToggleCoverMouth);

				SetAndRegisterAction("CoverActionWithHandsRandom"	,CoverActionWithHandsRandom);
				SetAndRegisterAction("CoverActionWithHandsUU"		,CoverActionWithHandsUU);
				SetAndRegisterAction("CoverActionWithHandsUL"		,CoverActionWithHandsUL);
				SetAndRegisterAction("CoverActionWithHandsLL"		,CoverActionWithHandsLL);
				SetAndRegisterAction("CoverActionCloseLegs"			,CoverActionCloseLegs);
				SetAndRegisterAction("CoverActionCoverMouth"		,CoverActionCoverMouth);
			}
			/*
			SetAndRegisterAction("TogglePenisSupport",     	TogglePenisSupport);
			SetAndRegisterAction("TogglePenisSupport",     	TogglePenisSupport);
			SetAndRegisterAction("TogglePenisSupport",     	TogglePenisSupport);
			SetAndRegisterAction("TogglePenisSupport",     	TogglePenisSupport);
			SetAndRegisterAction("TogglePenisSupport",     	TogglePenisSupport);
			SetAndRegisterAction("TogglePenisSupport",     	TogglePenisSupport);
			SetAndRegisterAction("TogglePenisSupport",     	TogglePenisSupport);
			SetAndRegisterAction("TogglePenisSupport",     	TogglePenisSupport);
			*/
			
//			SetAndRegisterAction("Make Her Orgasm", core.worldUi.OnOrgasmButtonClick);
			SetAndRegisterAction("▉▉▉▉▉▉▉▉ End of Trigger for External Linkage ▉▉▉▉▉▉▉▉", Dummy);
		}

		public void ToggleMinimize(){	core.worldUi.OnMinimizeClick(); }
		public void Suspend(){	core.worldUi.OnPauseStartClick(); }
		public void Resume(){	core.worldUi.OnPauseEndClick(); }
		public void ToggleCompact(){	core.worldUi.OnCompactClick(); }
		public void CollapseAndExpand(){	core.worldUi.OnCollapseExpandClick(); }
		public void ToggleBarOnly(){	core.worldUi.OnCollapseWOBarUIClick(); }
		public void CloseUI(){	core.worldUi.OnCloseUIClick(); }
		public void LookFront(){	core.worldUi.OnLookFrontClick(); }
		public void ReloadPlugins(){	core.worldUi.OnReloadPluginsClick(); }
		public void ReloadThisPluginOnly(){	core.worldUi.OnReloadThisPluginOnlyClick(); }
		public void ToggleSize(){	core.worldUi.OnChangeSizeClick(); }
		public void ToggleCategory(){	core.worldUi.OnCategoryOnClick(); }
		public void InstantCUAScaleLow(){	core.worldUi.OnInstantScaleLowClick(); }
		public void InstantCUAScaleMiddle(){	core.worldUi.OnInstantScaleMiddleClick(); }
		public void InstantCUAScaleHigh(){	core.worldUi.OnInstantScaleHighClick(); }

		public void ToggleColortone(){	core.worldUi.OnToggleColortoneClick(); }
		public void TogglePostMagic(){	core.worldUi.OnTogglePostMagicClick(false); }
		public void TogglePostMagicWithDoF(){	core.worldUi.OnTogglePostMagicClick(true); }
		public void ToggleVAMMoan(){	core.worldUi.OnToggleVAMMoanClick(); }

		public void MakeHerOrgasm(){   core.worldUi.StartOrgasm(); }
		public void StopHerOrgasm(){   core.worldUi.FinishOrgasm(); }
		public void SingleWave(){      core.worldUi.OnSingleWaveClick(); }
		public void MakeHerSquirt(){   core.worldUi.OnPeeButtonClick(); }
		public void CalmHerDown(){     core.worldUi.OnCalmDownButtonClick(); }
		public void StageUp(){         core.worldUi.OnStageUp(); }
		public void StageDown(){       core.worldUi.OnStageDown(); }
		public void ResetExpression(){ core.worldUi.OnResetExpressionButton(); }
		
		public void FaceForward()  { core.worldUi.OnFaceForwardClick(); }
		public void FaceSidewaysT(){ core.worldUi.OnLookGlancingClick(); }
		public void FaceSidewaysS(){ core.worldUi.OnFaceSidewaysClick(); }
		public void FaceSidewaysM(){ core.worldUi.OnFaceSidewaysAltClick(); }
		public void FaceSidewaysL(){ core.worldUi.OnFaceSidewaysMaxClick(); }
		public void TurnHeadBackT(){ core.worldUi.OnBackBendingWeakClick(); }
		public void TurnHeadBackL(){ core.worldUi.OnBackBendingStrongClick(); }
		

		public void ToggleExpression()			{ core.worldUi.OnConfigXClickCommon("C:Expression"); }
		public void ToggleBlushing()			{ core.worldUi.OnConfigXClickCommon("C:Blushing"); }
		public void ToggleTears()				{ core.worldUi.OnConfigXClickCommon("C:Tears"); }
		public void ToggleDrool()				{ core.worldUi.OnConfigXClickCommon("C:Drool"); }
		public void ToggleAutoExpression()		{ core.worldUi.OnConfigXClickCommon("C:AutoExprOFF"); }
		public void ToggleAutoBlink()			{ core.worldUi.OnConfigXClickCommon("C:AutoBlinkOFF"); }
		public void ToggleEyesControl()			{ core.worldUi.OnConfigXClickCommon("C:EyesControl"); }
		public void ToggleEyesRollup()			{ core.worldUi.OnConfigXClickCommon("C:EyesRollup"); }
		public void ToggleNippleErection()		{ core.worldUi.OnConfigXClickCommon("C:NippleErection"); }
		public void ToggleClitorisErection()	{ core.worldUi.OnConfigXClickCommon("C:ClitorisErection"); }
		public void ToggleHeadMotion()			{ core.worldUi.OnConfigXClickCommon("C:HeadMotion"); }
		// public void ToggleSensitiveTrigger()	{ core.worldUi.OnSensitiveTriggerButtonClick(); }
		public void ToggleCollisionFailureCheck()	{ core.worldUi.OnConfigXClickCommon("C:Collision Failure Check"); }
		public void TogglePenisSupport()		{ core.worldUi.OnConfigXClickCommon("C:Penis Support"); }
		public void ToggleShiveringBone()		{ core.worldUi.OnConfigXClickCommon("C:Shivering(Bone)"); }
		public void ToggleShiveringMorph()		{ core.worldUi.OnConfigXClickCommon("C:Shivering(Morph)"); }
		public void ToggleSquirt()				{ core.worldUi.OnConfigXClickCommon("C:Peeing"); }
		public void ToggleSound()				{ core.worldUi.OnConfigXClickCommon("C:Sound"); }		

		public void ToggleAutoStopOtherPlugins()			{ core.worldUi.OnConfigXClickCommon("C:AutoStop Plugins"); }		
		public void ToggleStopMorphControlByNonPlugins()	{ core.worldUi.OnConfigXClickCommon("C:AutoStop Morph"); }		
		public void ToggleStopHeadControlByNonPlugins()		{ core.worldUi.OnConfigXClickCommon("C:AutoStop Head"); }		
		public void ToggleStopBodyControlByNonPlugins()		{ core.worldUi.OnConfigXClickCommon("C:AutoStop Body"); }				
		
		// public void UndressAndRemove()				{ core.addonManager1.DoWorldUIAddon1Function("OnUndressAndRemoveClick"); }
		// public void UndressNormal()					{ core.addonManager1.DoWorldUIAddon1Function("OnUndressClick"); }
		// public void UndressForce()					{ core.addonManager1.DoWorldUIAddon1Function("OnUndressForceClick"); }
		public void UndressRemove()					{ core.addonManager1.DoWorldUIAddon1Function("OnUndressRemoveClick"); }
		// public void ReverseGravity()				{ core.addonManager1.DoWorldUIAddon1Function("OnReverseClick"); }
		// public void UndressAndReverse()				{ core.addonManager1.DoWorldUIAddon1Function("OnUndressAndReverseClick"); }
		// public void UndressRevStrong()  			{ core.addonManager1.DoWorldUIAddon1Function("OnUndressRevStrongClick"); }
		public void AltSkinWrap()					{ core.addonManager1.DoWorldUIAddon1Function("OnAltSkinWrapClick"); }
		// public void UndressStop()					{ core.addonManager1.DoWorldUIAddon1Function("OnUndressStopClick"); }
		public void UndressRecover()				{ core.addonManager1.DoWorldUIAddon1Function("OnUndressRecoverClick"); }
		public void ExposeBreasts()					{ core.addonManager1.DoWorldUIAddon1Function("OnExposeBreastsClick"); }
		public void TearOffShirt()					{ core.addonManager1.DoWorldUIAddon1Function("OnExposeBreastsStrongClick"); }
		public void PullUpSkirt()					{ core.addonManager1.DoWorldUIAddon1Function("OnPullUpSkirtClick"); }
		public void StopExposingAndPullingUp()		{ core.addonManager1.DoWorldUIAddon1Function("OnRemoveGameObjectsClick"); }
		public void HoldClothesContainingNGWords()	{ core.addonManager1.DoWorldUIAddon1Function("OnUndressOpt1Click"); }
		public void HoldClothesWithClothSim()		{ core.addonManager1.DoWorldUIAddon1Function("OnUndressOpt2Click"); }
		public void HoldClothesWithAlterSkinWrap()  { core.addonManager1.DoWorldUIAddon1Function("OnUndressOpt3Click"); }
		public void AlterSkinWraptoClothSim()		{ core.addonManager1.DoWorldUIAddon1Function("OnUndressOpt4Click"); }

		public void BackupPose()			{ core.DoAddonManager2Function("OnBackupPoseClick"); }
		public void RecoverPose()			{ core.DoAddonManager2Function("OnRecoverPoseClick"); }
		public void HoldMouth()				{ core.DoAddonManager2Function("OnHoldMouth1Click"); }
		public void DuctTape()				{ core.DoAddonManager2Function("OnHoldMouth2Click"); }
		public void HoldMouth2()			{ core.DoAddonManager2Function("OnHoldMouth2Click"); }
		public void StopHoldMouth()			{ core.DoAddonManager2Function("OnStopHoldMouthClick"); }
		public void HoldArms()				{ core.DoAddonManager2Function("OnHoldArmsClick"); }
		public void HoldArmL()				{ core.DoAddonManager2Function("OnHoldArmLClick"); }
		public void HoldArmR()				{ core.DoAddonManager2Function("OnHoldArmRClick"); }
		public void HoldArmsSide()			{ core.DoAddonManager2Function("OnLockArmsSideClick"); }
		public void HoldArmLSide()			{ core.DoAddonManager2Function("OnLockArmLSideClick"); }
		public void HoldArmRSide()			{ core.DoAddonManager2Function("OnLockArmRSideClick"); }
		public void HoldArmsBack()			{ core.DoAddonManager2Function("OnLockArmsBackClick"); }
		public void HoldArmLBack()			{ core.DoAddonManager2Function("OnLockArmLBackClick"); }
		public void HoldArmRBack()			{ core.DoAddonManager2Function("OnLockArmRBackClick"); }
		public void HoldArmsUp()			{ core.DoAddonManager2Function("OnLockArmsUpClick"); }
		public void HoldArmLUp()			{ core.DoAddonManager2Function("OnLockArmLUpClick"); }
		public void HoldArmRUp()			{ core.DoAddonManager2Function("OnLockArmRUpClick"); }
		public void StopHoldArms()			{ core.DoAddonManager2Function("OnStopLockArmsClick"); }
		public void StopHoldArmL()			{ core.DoAddonManager2Function("OnStopLockArmLClick"); }
		public void StopHoldArmR()			{ core.DoAddonManager2Function("OnStopLockArmRClick"); }
		public void ToggleArmBinderByHand()	{ core.DoAddonManager2Function("OnArmBinderByHandClick"); }
		public void ToggleArmBinderByBeam()	{ core.DoAddonManager2Function("OnArmBinderByBeamClick"); }
		public void HoldLegs()				{ core.DoAddonManager2Function("OnHoldLegsClick"); }
		public void SpreadLegsUp()			{ core.DoAddonManager2Function("OnOpenLegsUpClick"); }
		public void SpreadLegsWidely()		{ core.DoAddonManager2Function("OnOpenLegsWidelyClick"); }
		public void SpreadLegsSlightly()	{ core.DoAddonManager2Function("OnOpenLegsSlightlyClick"); }
		public void LiftUpKnee()			{ core.DoAddonManager2Function("OnLiftUpKneeClick"); }
		public void LiftUpKnees()			{ core.DoAddonManager2Function("OnLiftUpKneesClick"); }
		public void SplitLegs()				{ core.DoAddonManager2Function("OnLegSplitClick"); }
		public void StopHoldLegs()			{ core.DoAddonManager2Function("OnStopOpeningLegsClick"); }
		public void ToggleLegBinderByHand()	{ core.DoAddonManager2Function("OnLegBinderByHandClick"); }
		public void ToggleLegBinderByBeam()	{ core.DoAddonManager2Function("OnLegBinderByBeamClick"); }
		public void SummonHandR()			{ core.DoAddonManager2Function("OnSummonHandRv2Click"); }
		public void SummonHandL()			{ core.DoAddonManager2Function("OnSummonHandLv2Click"); }
		public void SummonTentacle1()		{ core.DoAddonManager2Function("OnSummonTentacle1v2Click"); }
		public void SummonMouth()			{ core.DoAddonManager2Function("OnSummonMouthClick"); }
		public void SummonClitSucker()		{ core.DoAddonManager2Function("OnSummonWMNZClick"); }
		public void SummonTwinDildo()		{ core.DoAddonManager2Function("OnSummonTwinDildoClick"); }
		public void UnsummonAll()			{ core.DoAddonManager2Function("OnStopAllClick"); }
		public void ToggleCUACollision()	{ core.DoAddonManager2Function("ChangeCollisionControls"); }

		public void CoverActionWithHandsRandom()	{ core.DoAddonManager2Function("CoverWithHands"); }
		public void CoverActionWithHandsUU()	{ core.DoAddonManager2Function("CoverWithHandsUU"); }
		public void CoverActionWithHandsUL()	{ core.DoAddonManager2Function("CoverWithHandsUL"); }
		public void CoverActionWithHandsLL()	{ core.DoAddonManager2Function("CoverWithHandsLL"); }
		public void CoverActionCloseLegs()	{ core.DoAddonManager2Function("CloseLegs"); }
		public void CoverActionCoverMouth()	{ core.DoAddonManager2Function("CoverMouth"); }
		public void ToggleCoveringActions() { core.DoAddonManager2Function("ToggleCoveringActions");}
		public void ToggleCoverWithHands() { core.DoAddonManager2Function("ToggleCoverWithHands");}
		public void ToggleCloseLegs() { core.DoAddonManager2Function("ToggleCloseLegs");}
		public void ToggleCoverMouth() { core.DoAddonManager2Function("ToggleCoverMouth");}


		/*
		public void ToggleSound()		{ core.worldUi.OnSoundOnButtonClick();	 }
		public void ToggleSound()		{ core.worldUi.OnSoundOnButtonClick();	 }
		public void ToggleSound()		{ core.worldUi.OnSoundOnButtonClick();	 }
		*/
		
		public void Dummy(){
			core.Info("This trigger is a dummy for marker purposes. It has no function other than to display this message.");
		}
		
		public void SetAndRegisterAction(string name, Action act){
			jsas.Add(name, new JSONStorableAction(name, () => { act();}));
			core.RegisterAction(jsas[name]);
		}
		#endregion

		// public string ToUpperLineHead(string str)
		// {
		// 	var _array = str.ToCharArray();
		// 	var up = char.ToUpper(_array[0]);
		// 	_array[0] = up;
		// 	Utility.Debug($"before: str:{str}, after:array:{new string(_array)}");
		// 	return new string(_array);
		// }
	}
}
