#region using
using System;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
using MVR.FileManagementSecure;
using MVR.FileManagement;
using GPUTools.Cloth.Scripts;
using GPUTools.Cloth.Scripts.Runtime.Data;
#endregion

namespace Cotyounoyume{
	public static class Utility
	{
		public static bool disableLog = true;
		
		public static Dictionary<string, bool> specificControlMap = new Dictionary<string, bool>(){
			{"headControl", true},
			{"headAnimation", true},
			{"pelvisControl", true},
			{"pelvisAnimation", true},
			/*
			{"neckControl", true},
			{"pelvisControl", true},
			{"abdomenControl", true},
			{"abdomen2Control", true},
			
			{"headAnimation", true},
			{"neckAnimation", true},
			{"pelvisAnimation", true},
			{"abdomenAnimation", true},
			{"abdomen2Animation", true},
			*/
		};		
		
		static public JSONStorable GetPluginStorableById(Atom atom, string id){
			try{
				string storableIdName = atom.GetStorableIDs().FirstOrDefault((string storeId) =>{
					if (string.IsNullOrEmpty(storeId)){
						return false;
					}
					return storeId.Contains(id);
				});
				if (storableIdName == null){
					return null;
				}
				return atom.GetStorableByID(storableIdName);
			}catch (Exception e){
				LoggerE(e);
				return null;
			}
		}
		
		public static MVRPluginManager GetMVRPluginManager(){
			var target = SuperController.singleton.GetComponentsInChildren<MVRPluginManager>().Where(mgr => mgr.name.ToLower().StartsWith("session")).First();
			return target;
		}

		public static MVRPluginManager GetMVRPluginManager(Atom atom){
			try{
				if(atom == null){ return null; }
				MVRPluginManager manager = atom.GetStorableByID("PluginManager") as MVRPluginManager;
				return manager;				
			}catch(Exception e){
				LoggerE(e);
				return null;
			}
		}
		
		public static List<MVRScript> GetMVRScripts(Atom atom){
			MVRPluginManager mvrpm = GetMVRPluginManager(atom);
			
			List<MVRScript> result = new List<MVRScript>();
			List<string> ids = atom.GetStorableIDs();
			foreach(var id in ids){
				if(id.StartsWith("plugin")){
					MVRScript script = atom.GetStorableByID(id) as MVRScript;
					result.Add(script);
				}
			}
				
			return result;
		}
		
		// Function taken from VAMDeluxe's code
		public static string GetPluginPath(MVRScript plugin, Atom atom){
            SuperController.singleton.currentSaveDir = SuperController.singleton.currentLoadDir;
            string pluginId = plugin.storeId.Split('_')[0];
            MVRPluginManager manager = GetMVRPluginManager(atom);
            string pathToScriptFile = manager.GetJSON(true, true)["plugins"][pluginId].Value;
            string pathToScriptFolder = pathToScriptFile.Substring(0, pathToScriptFile.LastIndexOfAny(new char[] { '/', '\\' }));
            return pathToScriptFolder;
		}

		public static string GetPluginNameFromAtom(Atom atom){
			JSONNode jsnLoad = SuperController.singleton.loadJson;
			JSONNode jsnPerson = GetStorableByID(jsnLoad["atoms"], atom.uid);
			if(jsnPerson == null){ return null; }
			
			JSONNode jsnStorables = jsnPerson["storables"];
			
			for(int i = 0; i < jsnStorables.Count; i++){
				string key = jsnStorables[i]["id"];
				if(key.EndsWith("EBATCore")){
					return key;
				}
			}
			return null;
		}
		
		public static JSONNode GetPluginJSONFromAtom(Atom atom){
			JSONNode jsnLoad = SuperController.singleton.loadJson;
			JSONNode jsnPerson = GetStorableByID(jsnLoad["atoms"], atom.uid);
//			if(jsnPerson == null){ return null; }
			
			JSONNode jsnStorables = jsnPerson["storables"];
			
			for(int i = 0; i < jsnStorables.Count; i++){
				string key = jsnStorables[i]["id"];
				if(key.EndsWith("EBATCore")){
					return jsnStorables[i];
				}
			}
			return null;
		}
		
		public static List<string> GetPluginNamesFromAtom(Atom atom){
			JSONNode jsnLoad = SuperController.singleton.loadJson;
			JSONNode jsnPerson = GetStorableByID(jsnLoad["atoms"], atom.uid);
//			if(jsnPerson == null){ return null; }
			
			JSONNode jsnStorables = jsnPerson["storables"];
			
			List<string> result = new List<string>();
			for(int i = 0; i < jsnStorables.Count; i++){
				string key = jsnStorables[i]["id"];
				if(key.StartsWith("plugin#")){
					result.Add(key);
				}
			}
			return result;
		}
		
		public static JSONNode GetStorableByID(JSONNode jsn, string key){
			for(int i = 0; jsn[i] != null; i++){
				string id = jsn[i]["id"];
				if(id == key){
					return jsn[i];
				} 
			}
			return null;
		}
		
		public static JSONStorable GetEBATJS(Atom atom){
			JSONStorable js = null;
			foreach(var id in atom.GetStorableIDs()){
				if(id.EndsWith("EBATCore")){
					js = atom.GetStorableByID(id);
				}
			}
			return js;
		}
		
		public static Dictionary<string, string> GetParameterInSave(Atom atom){
			Dictionary<string, string> result = new Dictionary<string, string>();
			
			string pluginName = Utility.GetPluginNameFromAtom(atom);
			if(pluginName == null || pluginName == ""){
				return null;
			}
			JSONClass loadJson = SuperController.singleton.loadJson.AsObject;
			JSONArray atoms = loadJson["atoms"].AsArray;
			JSONClass person = null;
			foreach(JSONNode jsn in atoms.Childs){
				JSONClass jsc = jsn.AsObject;
				string id = jsc["id"].ToString();
				string uid = "\"" + atom.uid + "\"";
				
				if(id == uid){
					person = jsc;
				}
			}
			if(person == null){ return null; }
			
			JSONArray storables = person["storables"].AsArray;
			JSONClass myStorage = null;
			foreach(JSONNode jsn in storables.Childs){
				JSONClass jsc = jsn.AsObject;
				string id = jsc["id"].ToString();
				string pname = "\"" + pluginName + "\"";
				
				if(id == pname){
					myStorage = jsc;
				}
			}
			
			foreach(string key in myStorage.Keys){
				if(!result.ContainsKey(key)){
					result.Add(key.ToString(), myStorage[key].ToString());
				}
			}
			return result;
		}		
		
		public static float CalculateProgress(float from, float to, float start, float finish, float current){
			float result;
			float distance = to - from;
			float denominator = finish - start;
			float numerator = current - start;
			result = from + distance * numerator / denominator;
			return result;
		}
		
		public static void ResetParent(Atom targetAtom){
			targetAtom.transform.SetParent(null);
		}
		
		public static void SetParent(Atom targetAtom, Atom srcAtom, string srcBodyName){
			try{
				int num = FindRigidBodyNumber(srcAtom, srcBodyName);
				if(num != -1){
					targetAtom.transform.SetParent(srcAtom.rigidbodies[num].transform, false);
					return;
				}
				FreeControllerV3 fc = srcAtom.GetStorableByID(srcBodyName) as FreeControllerV3;
				if(fc != null){
					targetAtom.transform.SetParent(fc.transform, false);
					return;
				}
				Logger("can't find " + srcBodyName + " in " + srcAtom);
			}catch(Exception e){
				LoggerE(e);
			}
		}

		public static List<Atom> GetLinkedCUAs(){
			var lists = new List<Atom>();
			SuperController.singleton.GetAtoms().ForEach(atom => {
				if(atom.type == "CustomUnityAsset"){
					var rbName = GetLinkToRB(atom);
					if(rbName != ""){
						lists.Add(atom);
					}
				}
			});
			return lists;
		}

		public static string GetLinkToRB(Atom cua){
			if(cua.type != "CustomUnityAsset") return "";
			var mc = cua.mainController;
			if(mc == null) return "";
			var linkto = mc.linkToRB;
			if(linkto == null) return "";

			return linkto.name;
		}

		public static void SetParent2(Atom targetAtom, Atom srcAtom, string srcBodyName){
			try{
				Rigidbody rb = Utility.FindRigidBody(srcAtom, srcBodyName);
				Utility.Debug($"rb:{rb}");
				targetAtom.mainController.SelectLinkToRigidbody(rb);
				// Utility.Debug($"masterController:{targetAtom.masterController}");
				// targetAtom.masterController.SelectLinkToRigidbody(rb);
			}catch(Exception e){
				LoggerE(e);
			}
		}
		
		public static int FindRigidBodyNumber(Atom srcAtom, string bodyName){
			int num = -1;
			for(int i = 0; i < srcAtom.rigidbodies.Length; i++){
				if(srcAtom.rigidbodies[i].name == bodyName){
					num = i;
					Logger("rigidbody: i=" + i + ", name=" + srcAtom.rigidbodies[i].name);
				}
			}
			return num;
		}
		
		public static Rigidbody FindRigidBody(Atom srcAtom, string bodyName){
			int num = -1;
			for(int i = 0; i < srcAtom.rigidbodies.Length; i++){
				if(srcAtom.rigidbodies[i].name == bodyName){
					num = i;
					Logger("rigidbody: i=" + i + ", name=" + srcAtom.rigidbodies[i].name);
				}
			}
			return srcAtom.rigidbodies[num];
		}
		
		public static void SwitchPositionState(Atom atom, string[] keys, string state){
			for(int i = 0; i < keys.Length; i++){
				FreeControllerV3 fc = GetFreeControllerV3(atom, keys[i]);
				fc.SetPositionStateFromString(state);
			}
		}
		
		public static void SwitchPositionState(Atom atom, string key, string state){
			FreeControllerV3 fc = GetFreeControllerV3(atom, key);
			fc.SetPositionStateFromString(state);
		}
		
		public static void SwitchRotationState(Atom atom, string[] keys, string state){
			for(int i = 0; i < keys.Length; i++){
				FreeControllerV3 fc = GetFreeControllerV3(atom, keys[i]);
				fc.SetRotationStateFromString(state);
			}
		}
		
		public static void SwitchRotationState(Atom atom, string key, string state){
			FreeControllerV3 fc = GetFreeControllerV3(atom, key);
			fc.SetRotationStateFromString(state);
		}
		
		public static FreeControllerV3 GetFreeControllerV3(Atom atom, string bodyName){
			return atom.GetStorableByID(bodyName) as FreeControllerV3;			
		}
		
		public static Vector3 GetPosition(FreeControllerV3 fc){
			return new Vector3(fc.transform.position.x, fc.transform.position.y, fc.transform.position.z);
		}
		
		public static Vector3 GetPosition(Transform t){
			return new Vector3(t.position.x, t.position.y, t.position.z);
		}
		
		public static Vector3 GetPosition(Rigidbody r){
			return new Vector3(r.position.x, r.position.y, r.position.z);
		}

		public static Vector3 GetPosition(Vector3 pos){
			return new Vector3(pos.x, pos.y, pos.z);
		}

		public static Quaternion GetRotation(FreeControllerV3 fc){
			return new Quaternion(fc.transform.rotation.x, fc.transform.rotation.y, fc.transform.rotation.z, fc.transform.rotation.w);
		}
		
		public static Quaternion GetRotation(Transform t){
			return new Quaternion(t.rotation.x, t.rotation.y, t.rotation.z, t.rotation.w);
		}
		
		public static Quaternion GetRotation(Rigidbody r){
			return new Quaternion(r.rotation.x, r.rotation.y, r.rotation.z, r.rotation.w);
		}		
		
		public static Quaternion GetRotation(Quaternion rot){
			return new Quaternion(rot.x, rot.y, rot.z, rot.w);
		}
		
		public static void PrintPosition(string str, Vector3 pos){
			Info(str + ": pos x/y/z: " + pos.x + "/" + pos.y + "/" + pos.z);
		}
		
		public static Transform GetTransformFromTargetBodyname(Atom atom, string bodyName){
			try{
				FreeControllerV3 fc = atom.GetStorableByID(bodyName) as FreeControllerV3;
				if(fc != null){
					return fc.transform;
				}
				int num = FindRigidBodyNumber(atom, bodyName);
				if(num != -1){
					return atom.rigidbodies[num].transform;
				}
				Logger("error: bodyName=" + bodyName + " is not found from " + atom);
				return null;
			}catch(Exception e){
				LoggerE(e);
				return null;
			}
		}
		
		public static Vector3 GetPositionFromTargetBodyname(Atom atom, string bodyName){
			try{
				Transform t = GetTransformFromTargetBodyname(atom, bodyName);
				if(t != null){
					return GetPosition(t);
				}else{
					return new Vector3(0f, 0f, 0f);
				}
			}catch(Exception e){
				LoggerE(e);
				return new Vector3(0f, 0f, 0f);
			}			
		}

		public static Quaternion GetRotationFromTargetBodyname(Atom atom, string bodyName){
			try{
				Transform t = GetTransformFromTargetBodyname(atom, bodyName);
				if(t != null){
					return GetRotation(t);
				}
				return new Quaternion(0f, 0f, 0f, 0f);
			}catch(Exception e){
				LoggerE(e);
				return new Quaternion(0f, 0f, 0f, 0f);
			}			
		}
		
		public static Vector3 GetPositionFromTargetTransformWithOffset(Transform target, float upOffset = 0f, float forwardOffset = 0f, float rightOffset = 0f, float fluctuation = 0f){
			try{
				Vector3 result = Vector3.zero;
				if(target == null){
					Info("target is null");
				}
				if(target != null){
					float fup      = UnityEngine.Random.Range(-1f * fluctuation, fluctuation);
					float fforward = UnityEngine.Random.Range(-1f * fluctuation, fluctuation);
					float fright   = UnityEngine.Random.Range(-1f * fluctuation, fluctuation);
					result = target.TransformPoint(target.up * (upOffset + fup) + target.forward * (forwardOffset + fforward) + target.right * (rightOffset + fright));
				}
				
				return result;
			}catch(Exception e){
				LoggerE(e);
				return Vector3.zero;
			}
		}
				
		public static List<Atom> CollectPersons(bool isFemale){
			List<Atom> atoms = new List<Atom>();
            foreach (var atom in SuperController.singleton.GetAtoms())
            {
				if(atom.type == "Person"){
					if(isFemale){
						if(Utility.IsFemale(atom)){
							atoms.Add(atom);
						}
					}else{
						if(!Utility.IsFemale(atom)){
							atoms.Add(atom);
						}
					}
				}
            }
			return atoms;
		}				
				
		public static Vector3 GetPositionFromTargetBodynameWithOffset(Atom atom, string bodyName, float upOffset = 0f, float forwardOffset = 0f, float rightOffset = 0f, float fluctuation = 0f){
			try{
				Vector3 result = new Vector3(0f, 0f, 0f);
				Transform t = GetTransformFromTargetBodyname(atom, bodyName);
				if(t == null){
					Info("t is null");
				}
				if(t != null){
					float fup      = UnityEngine.Random.Range(-1f * fluctuation, fluctuation);
					float fforward = UnityEngine.Random.Range(-1f * fluctuation, fluctuation);
					float fright   = UnityEngine.Random.Range(-1f * fluctuation, fluctuation);
					result = t.TransformPoint(Vector3.up * (upOffset + fup) + Vector3.forward * (forwardOffset + fforward) + Vector3.right * (rightOffset + fright));
				}
				
				return result;
			}catch(Exception e){
				LoggerE(e);
				return new Vector3(0f, 0f, 0f);
			}
		}
				
		public static Vector3 GetPositionFromTargetBodynameWithOffsetAndRotationArrangement(Atom atom, string bodyName, Transform transformForRotation, float upOffset = 0f, float forwardOffset = 0f, float rightOffset = 0f, float fluctuation = 0f){
			Transform t1 = GetTransformFromTargetBodyname(atom, bodyName);
			return GetPositionFromTargetBodynameWithOffsetAndRotationArrangement(atom, t1, transformForRotation, upOffset, forwardOffset, rightOffset, fluctuation);
		}
		
		public static Vector3 GetPositionFromTargetBodynameWithOffsetAndRotationArrangement(Atom atom, string bodyName, string bodyNameForRotation, float upOffset = 0f, float forwardOffset = 0f, float rightOffset = 0f, float fluctuation = 0f){
			Transform t1 = GetTransformFromTargetBodyname(atom, bodyName);
			Transform t2 = GetTransformFromTargetBodyname(atom, bodyNameForRotation);
			return GetPositionFromTargetBodynameWithOffsetAndRotationArrangement(atom, t1, t2, upOffset, forwardOffset, rightOffset, fluctuation);
		}
		
		public static Vector3 GetPositionFromTargetBodynameWithOffsetAndRotationArrangement(Atom atom, Transform t1, Transform t2, float upOffset = 0f, float forwardOffset = 0f, float rightOffset = 0f, float fluctuation = 0f){
			try{
				Vector3 result = new Vector3(0f, 0f, 0f);
				Vector3 diff = t1.position - t2.position;
				Vector3 bkup1 = GetPosition(t1);
				Vector3 bkup2 = GetPosition(t2);
				if(t1 == null || t2 == null){
					Info("t1 or t2 is null");
				}
				if(t1 != null && t2 != null){
					float fup      = UnityEngine.Random.Range(-1f * fluctuation, fluctuation);
					float fforward = UnityEngine.Random.Range(-1f * fluctuation, fluctuation);
					float fright   = UnityEngine.Random.Range(-1f * fluctuation, fluctuation);
					t1.position += t2.up * (upOffset + fup);
					t1.position += t2.forward * (forwardOffset + fforward);
					t1.position += t2.right * (rightOffset + fright);
					result = t1.position;
					t1.position = bkup1;
					t2.position = bkup2;
				}				
				return result;
			}catch(Exception e){
				LoggerE(e);
				return new Vector3(0f, 0f, 0f);
			}
		}
				
		public static Vector3 GetMiddlePosition(Vector3 pos1, Vector3 pos2){
			return new Vector3((pos1.x + pos2.x)/2, (pos1.y + pos2.y)/2, (pos1.z + pos2.z)/2);
		}

		public static void SetAssetBundleUrl(Atom cua, string url){
			JSONStorable js = cua.GetStorableByID("asset");
			JSONStorableUrl jsAssetUrl = js.GetUrlJSONParam("assetUrl");

			Logger("Before: jsAssetUrl.val: " + jsAssetUrl.val);
			js.SetUrlParamValue("assetUrl", url);
			Logger("After : jsAssetUrl.val: " + jsAssetUrl.val);
		}
		
		public static Atom GetAssetIncludeAtom(string assetUrl){
			var atoms = SuperController.singleton.GetAtoms();
			Atom cua = null;
			foreach(var atom in atoms){
				var result = HasAsset(atom, assetUrl);
				if(result){
					return atom;
				}
			}
			return cua;
		}


		public static bool HasAssetName(Atom atom, string pattern = ""){
			if(atom.type != "CustomUnityAsset") return false;
			JSONStorable js = atom.GetStorableByID("asset");
			if(js == null) return false;
			JSONStorableStringChooser jsAssetName = js.GetStringChooserJSONParam("assetName");
			if(jsAssetName == null) return false;
			var name = jsAssetName.val.ToLower();
			// Utility.Debug($"atom:{atom}, name:{name}, pattern:{pattern}");
			if(pattern == "") return true;
			pattern = pattern.ToLower();
			if(name.Contains(pattern)) return true;
			return false;
		}

		public static List<string> GetAssetUrlAndName(Atom atom){
			var list = new List<string>(){
				"", ""
			};
			if(atom.type != "CustomUnityAsset") return list;
			JSONStorable js = atom.GetStorableByID("asset");
			if(js == null) return list;
			JSONStorableUrl jsAssetUrl = js.GetUrlJSONParam("assetUrl");
			JSONStorableStringChooser jsAssetName = js.GetStringChooserJSONParam("assetName");
			if(jsAssetUrl != null) list[0] = jsAssetUrl.val;
			if(jsAssetName != null) list[1] = jsAssetName.val;
			return list;
		}

		public static bool HasAsset(Atom atom, string pattern = ""){
			if(atom.type != "CustomUnityAsset") return false;
			JSONStorable js = atom.GetStorableByID("asset");
			if(js == null) return false;
			JSONStorableUrl jsAssetUrl = js.GetUrlJSONParam("assetUrl");
			JSONStorableStringChooser jsAssetName = js.GetStringChooserJSONParam("assetName");
			if(jsAssetUrl == null) return false;
			if(jsAssetName == null) return false;
			var url  = jsAssetUrl.val.ToLower();
			var name = jsAssetName.val.ToLower();
			// Utility.Debug($"atom:{atom}, url:{url}, name:{name}, pattern:{pattern}");
			if(pattern == "") return true;
			pattern = pattern.ToLower();
			if(url.Contains(pattern)) return true;
			if(name.Contains(pattern)) return true;
			return false;
		}

		public static void ResetAssetBundleName(Atom cua, string name){
			JSONStorable js = cua.GetStorableByID("asset");
			JSONStorableStringChooser jsAssetName = js.GetStringChooserJSONParam("assetName");

			js.SetStringChooserParamValue("assetName", "None");
			js.SetStringChooserParamValue("assetName", name);
			Logger("After1: jsAssetName.val: " + name);
		}
		
		public static int GetIndexFromJSONArray(JSONArray jsa, string target){
			int result = -1;
			for(int i = 0; i < jsa.Count; i++){
				if(jsa[i].ToString().Trim('"') == target){
					result = i;
					break;
				}
			}
			return result;
		}
		
		public static void Logger(string str){
			if(disableLog == false){
				SuperController.LogMessage(str);
			}
		}
		
		public static void LoggerE(Exception e){
			SuperController.LogMessage("*** Exception:" + e);
		}

		public static void Debug(string str)
		{
			SuperController.LogMessage("debug: " + str);
		}

		public static void Info(string str)
		{
			SuperController.LogMessage("info: " + str);
		}

		public static UIDynamic CreateSpacer(MVRScript plugin, Canvas canvas, float width = 100, float height = 80)
		{
			try{
				Transform spacer = GameObject.Instantiate<Transform>(plugin.manager.configurableSpacerPrefab);
				ConfigureTransform(spacer, width, height);
				ParentToCanvas(spacer, canvas);
				
				UIDynamic uiSpacer = spacer.GetComponent<UIDynamic>();
				return uiSpacer;
			}catch (Exception e){
				LoggerE(e);
				return null;
			}
		}
		
		public static UIDynamicButton CreateButton(MVRScript plugin, Canvas canvas, string label, float width = 100, float height = 80, int fontSize = 28)
		{
			try{
				Transform button = GameObject.Instantiate<Transform>(plugin.manager.configurableButtonPrefab);
				ConfigureTransform(button, width, height);
				ParentToCanvas(button, canvas);
		
				UIDynamicButton uiButton = button.GetComponent<UIDynamicButton>();
				uiButton.label = label;
				uiButton.buttonText.fontSize = fontSize;
				return uiButton;
			}catch (Exception e){
				LoggerE(e);
				return null;
			}
		}

		public static UIDynamicTextField CreateTextField(MVRScript plugin, Canvas canvas, string label, float width = 100, float height = 80, int fontSize = 28)
		{
			try{
				Transform textField = GameObject.Instantiate<Transform>(plugin.manager.configurableTextFieldPrefab);
				ConfigureTransform(textField, width, height);
				ParentToCanvas(textField, canvas);
		
				UIDynamicTextField uiText = textField.GetComponent<UIDynamicTextField>();
				uiText.text = label;
				uiText.UItext.fontSize = fontSize;
				return uiText;
			}catch (Exception e){
				LoggerE(e);
				return null;
			}
		}
		
        public static UIDynamicToggle CreateToggle(MVRScript plugin, Canvas canvas, string name, float width = 100, float height = 80, int fontSize = 28)
        {
            Transform toggle = GameObject.Instantiate<Transform>(plugin.manager.configurableTogglePrefab);
            ConfigureTransform(toggle, width, height);
			ParentToCanvas(toggle, canvas);

            UIDynamicToggle uiToggle = toggle.GetComponent<UIDynamicToggle>();
            uiToggle.label = name;
			uiToggle.labelText.fontSize = fontSize;
            return uiToggle;
        }

        public static UIDynamicSlider CreateSlider(MVRScript plugin, Canvas canvas, string name, float width = 300, float height = 80, bool minimal = true, int fontSize = 28)
        {
            Transform slider = GameObject.Instantiate<Transform>(plugin.manager.configurableSliderPrefab);
            ConfigureTransform(slider, width, height);
            ParentToCanvas(slider, canvas);

            UIDynamicSlider sliderDynamic = slider.GetComponent<UIDynamicSlider>();
			
            if (minimal)
            {
                sliderDynamic.quickButtonsEnabled = false;
                sliderDynamic.rangeAdjustEnabled = false;
                sliderDynamic.defaultButtonEnabled = false;
            }
            sliderDynamic.label = name;
            sliderDynamic.labelText.fontSize = fontSize;

            return sliderDynamic;
        }

        public static UIDynamicSlider CreateSliderAlt(MVRScript plugin, Canvas canvas, string name, float width = 300, float height = 80, float min = 0, float max = 1, float defaultValue = 0, bool minimal = true, int fontSize = 28, float ratio = 1.0f)
        {
            Transform slider = GameObject.Instantiate<Transform>(plugin.manager.configurableSliderPrefab);
            ConfigureTransform(slider, width, height);
            ParentToCanvas(slider, canvas);
			
            UIDynamicSlider sliderDynamic = slider.GetComponent<UIDynamicSlider>();
			sliderDynamic.Configure(name, min, max, defaultValue, showRangeAdjust:true);
            if (minimal)
            {
                sliderDynamic.quickButtonsEnabled = false;
                sliderDynamic.rangeAdjustEnabled = false;
                sliderDynamic.defaultButtonEnabled = false;
            }
            sliderDynamic.label = name;
            sliderDynamic.labelText.fontSize = fontSize;
			sliderDynamic.sliderValueTextFromFloat.UIInputField.textComponent.fontSize = fontSize;

			if(ratio < 1.0f){
				float barHeight = 40f;
				ConfigureTransformHeight(sliderDynamic.slider.handleRect, barHeight * ratio, -1 * barHeight * ratio / 2);
				ConfigureTransformHeight(sliderDynamic.slider.fillRect, barHeight * ratio, -1 * barHeight * ratio / 2);				
			}

            return sliderDynamic;
        }

		public static void TranslateUI(UIDynamic ui, float locx, float locy){
			if(ui != null && ui.transform != null){
				ui.transform.Translate(locx, locy, 0, Space.Self);
			}
		}
		
		public static void ConfigureTransformHeight(RectTransform rt, float height, float heightOffset)
		{
			rt.anchoredPosition = new Vector2(rt.anchoredPosition.x, heightOffset);
			rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
		}
		
		public static void ConfigureTransform(Transform t, float width, float height)
		{
			t.transform.position = Vector3.zero;
			RectTransform rt = t.GetComponent<RectTransform>();
			rt.anchoredPosition = new Vector2(width / 2, height / 2);
			rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);
			rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
		}
		
        public static void ColorButton(UIDynamicButton button, Color textColor, Color buttonColor)
        {
            button.textColor = textColor;
            button.buttonColor = buttonColor;
        }

		public static void LookFront(Transform t){
			Transform cameraT = SuperController.singleton.lookCamera.transform;
			Vector3 endPos = cameraT.position + cameraT.forward * 10000000.0f;
			t.LookAt(endPos, cameraT.up);			
		}
		
		public static void ParentToCanvas(Transform t, Canvas canvas)
		{
			if(t != null && canvas != null)
				t.SetParent(canvas.transform, false);
		}
		
		public static void RemoveSliderForCanvas(UIDynamicSlider dslider){
//				sliderToJSONStorableFloat.Remove(dslider);
				Transform transform = dslider.transform;
//				rightUIElements.Remove(transform);
//				leftUIElements.Remove(transform);
//				value.slider = null;
				GameObject.Destroy(transform.gameObject);
		}
		
		public static void RemoveTextFieldForCanvas(UIDynamicTextField text){
				Transform transform = text.transform;
				GameObject.Destroy(transform.gameObject);
		}
		
		public static bool IsEditMode()
		{
			return SuperController.singleton.gameMode == SuperController.GameMode.Edit;
		}

		public static void ChangeGameModeIfNeeded()
		{
			if(IsEditMode()) return;
			SuperController.singleton.gameMode = SuperController.GameMode.Edit;
		}

		public static void OpenAtomUI(Atom atom){
			ChangeGameModeIfNeeded();
			SuperController.singleton.SelectController(atom.mainController);
			SuperController.singleton.ShowMainHUD();
		}		
		
		public static void CloseUI(){
			SuperController.singleton.HideMainHUD();
		}		
		
		public static List<DAZClothingItem> GetClothes(Atom atom){
			JSONStorable geometryStorable = atom.GetStorableByID("geometry");
			DAZCharacterSelector geometryForClothes = geometryStorable as DAZCharacterSelector;
			return geometryForClothes.clothingItems.ToList();
		}
		
		public static DAZClothingItem GetClothByDisplayName(Atom atom, string displayName){
			List<DAZClothingItem> clothes = GetClothes(atom);
			for (int i = 0; i < clothes.Count; i++){
				if(clothes[i].displayName == displayName && clothes[i].isLatestVersion){
					return clothes[i];
				}
			}
			return null;
		}

		public static void UnequipClothByDisplayName(Atom atom, string displayName){
			try{
				JSONStorable geometryStorable = atom.GetStorableByID("geometry");
				DAZCharacterSelector geometryForClothes = geometryStorable as DAZCharacterSelector;
			
				List<DAZClothingItem> clothes = geometryForClothes.clothingItems.ToList();
				for (int i = 0; i < clothes.Count; i++){
					if(clothes[i].displayName == displayName){
						geometryForClothes.SetActiveClothingItem(clothes[i], false);
					}
				}
			}catch(Exception e){
				LoggerE(e);
			}
		}
		
		public static DAZClothingItem EquipClothByDisplayName(Atom atom, string displayName, bool active = true, bool isLatestVersion = true){
			try{
				JSONStorable geometryStorable = atom.GetStorableByID("geometry");
				DAZCharacterSelector geometryForClothes = geometryStorable as DAZCharacterSelector;

				List<DAZClothingItem> clothes = geometryForClothes.clothingItems.ToList();
				for (int i = 0; i < clothes.Count; i++){
					if(clothes[i].displayName == displayName && clothes[i].isLatestVersion){
						geometryForClothes.SetActiveClothingItem(clothes[i], active);
						return clothes[i];
					}
				}
				Logger("can't find cloth: " + displayName);
				return null;
			}catch(Exception e){
				LoggerE(e);
				return null;
			}
		}

		public static Dictionary<string, VariableTrigger> CollectVariableTriggers(){
			Dictionary<string, VariableTrigger> result = new Dictionary<string, VariableTrigger>();
			foreach(var atom in SuperController.singleton.GetAtoms()){
				if(atom.type == "VariableTrigger"){
					VariableTrigger vt = atom.GetStorableByID("Trigger") as VariableTrigger;
					if(vt != null){
						result.Add(atom.uid, vt);
					}
				}
			}
			return result;
		}
		
		public static Dictionary<string, AnimationPattern> CollectAnimationPatterns(){
			Dictionary<string, AnimationPattern> result = new Dictionary<string, AnimationPattern>();
			foreach(var atom in SuperController.singleton.GetAtoms()){
				if(atom.type == "AnimationPattern"){
					AnimationPattern animationPattern = atom.GetStorableByID("AnimationPattern") as AnimationPattern;
					if(animationPattern != null){
						result.Add(atom.uid, animationPattern);
					}
				}
			}
			return result;
		}
		
		public static Dictionary<string, JSONStorable> CollectAnimatedObjects(){
			Dictionary<string, JSONStorable> result = new Dictionary<string, JSONStorable>();
			foreach(var atom in SuperController.singleton.GetAtoms()){
				if(atom.type == "AnimationPattern"){
					JSONStorable js = atom.GetStorableByID("AnimatedObject");
					result.Add(atom.uid, js);
				}
			}
			return result;
		}

		
		public static JSONClass MakeJSCWithAtomRelatedTriggerExcluded(Atom atom, VariableTrigger vTrigger, bool isMorph = true){
			JSONClass jscRoot = vTrigger.GetJSON();
			JSONClass jsc = jscRoot["trigger"].AsObject;
			return MakeJSCWithAtomRelatedActionExcluded(atom, jsc, isMorph);
		}
		
		public static JSONClass MakeJSCWithAtomRelatedTriggersExcluded(Atom atom, AnimationPattern animationPattern, bool isMorph = true){
			JSONClass jscRoot = animationPattern.GetJSON();
			JSONArray jsa     = jscRoot["triggers"].AsArray;

			for(int i = 0; i < jsa.Count; i++){
				JSONClass newJSC = MakeJSCWithAtomRelatedActionExcluded(atom, jsa[i].AsObject, isMorph);
				jsa[i] = newJSC;
			}
			return jscRoot;
		}
		
		public static JSONClass MakeJSCWithAtomRelatedActionExcluded(Atom atom, JSONClass jsc, bool isMorph = true){
			string[] actionNames = {"startActions", "transitionActions", "endActions"};
			foreach(var actionName in actionNames){
				int count = jsc[actionName].AsArray.Count;
				for(int i = count - 1; i >= 0; i--){
					if(ContainsAtomInActionJSN(atom, jsc[actionName][i], isMorph)){
						jsc[actionName].AsArray.Remove(i);
					}
				}
			}
			return jsc;			
		}
		
		public static JSONClass MakeJSCWithAtomRelatedReceiverExcluded(Atom atom, JSONStorable js, bool isSpecificControl = true){
			JSONClass jsc = js.GetJSON();
			if(ContainsAtomInAnimationTarget(atom, js, isSpecificControl)){
				jsc.Remove("receiver");
			}
			return jsc;
		}		
		
		public static bool ContainsAtomInActionJSN(Atom atom, JSONNode jsn, bool isMorph = true){
			if(jsn["receiverAtom"].ToString().Trim('"') == atom.uid){
				if(isMorph){
					DAZCharacterSelector character = atom.GetStorableByID("geometry") as DAZCharacterSelector;
					GenerateDAZMorphsControlUI morphControl = character.morphsControlUI;
					DAZMorph morph = morphControl.GetMorphByDisplayName(jsn["receiverTargetName"].ToString().Trim('"'));
					if(morph != null){
						return true;
					}
				}else{
					return true;
				}
			}
			return false;
		}
		
		public static bool ContainsAtomInTrigger(Atom atom, VariableTrigger vTrigger, bool isMorph = true){
			JSONClass jscRoot = vTrigger.GetJSON();
			JSONClass jsc = jscRoot["trigger"].AsObject;
			return ContainsAtomInTriggerJSC(atom, jsc, isMorph);
		}
		
		public static bool ContainsAtomInTriggers(Atom atom, AnimationPattern animationPattern, bool isMorph = true){
			string[] actionNames = {"startActions", "transitionActions", "endActions"};
			JSONClass jscRoot = animationPattern.GetJSON();
			JSONArray jsa     = jscRoot["triggers"].AsArray;
			foreach(JSONClass jsc in jsa){
				if(ContainsAtomInTriggerJSC(atom, jsc, isMorph)){
					return true;
				}
			}
			return false;
		}
		
		public static bool ContainsAtomInTriggerJSC(Atom atom, JSONClass jsc, bool isMorph = true){
			string[] actionNames = {"startActions", "transitionActions", "endActions"};
			foreach(var actionName in actionNames){
				foreach(JSONNode jsn in jsc[actionName].AsArray){
					if(ContainsAtomInActionJSN(atom, jsn, isMorph)){
						return true;
					}
				}
			}
			return false;
		}
	
		public static bool ContainsAtomInAnimationTarget(Atom atom, JSONStorable js, bool isSpecificControl = true){
			JSONClass jsc = js.GetJSON();
			if(jsc.HasKey("receiver")){
				string receiver = jsc["receiver"].ToString().Trim('"');
				string[] words = receiver.Split(':');
				string atomUid = words[0];
				string control = (words.Length > 1) ? words[1] : "";
				
				if(atomUid == atom.uid){
					if(!isSpecificControl){
						return true;
					}else if(CheckSpecificControl(control)){
						return true;						
					}
				}
			}
			return false;
		}
		
		public static bool CheckSpecificControl(string name){
			bool result = false;
			
			if(specificControlMap.ContainsKey(name) && specificControlMap[name]){
				result = true;
			}else{
				result = false;
			}
			return result;
		}
		
		public static List<DAZMorph> GetAllActivePose(Atom atom, string[] categories = null){
			
			JSONStorable geometry = atom.GetStorableByID("geometry");
			DAZCharacterSelector character = geometry as DAZCharacterSelector;
			GenerateDAZMorphsControlUI morphControl = character.morphsControlUI;
			
			Dictionary<string, bool> hash = new Dictionary<string, bool>();
			if(categories != null){
				foreach(string category in categories){
					hash.Add(category, true);
				}
			}

			List<DAZMorph> result;
			morphControl.onlyShowActive = true;
			
			if(categories != null){
				result = new List<DAZMorph>();
				foreach(string category in categories){
					List<DAZMorph> morphs = morphControl.GetMorphs();
					foreach(DAZMorph morph in morphs){
						if(hash.ContainsKey(morph.region)){
							result.Add(morph);
						}
					}
				}
			}else{
				result = morphControl.GetMorphs();
			}
			
			return result;
		}
		
		public static void FistHand(Atom atom, float ratio){
			FistHand(atom, ratio, "left");
			FistHand(atom, ratio, "right");
		}

		public static void FistHand(Atom atom, float ratio, string type, bool indexOn = false){
			// Utility.Debug($"type:{type}, ratio:{ratio}");
			JSONStorable geometry = atom.GetStorableByID("geometry");
			DAZCharacterSelector character = geometry as DAZCharacterSelector;
			GenerateDAZMorphsControlUI morphControl = character.morphsControlUI;
			var hand = type == "left" ? "Left Hand Fist" : "Right Hand Fist";
			var index = type == "left" ? "Left Pinky Finger Bend" : "Right Pinky Finger Bend";
			morphControl.GetMorphByDisplayName(hand).SetValue(ratio);
			morphControl.GetMorphByDisplayName(index).SetValue(ratio * -0.7f);
		}
		
		public static void ResetHandsPose(Atom atom){
			List<DAZMorph> morphs = CollectHandsPose(atom);
			foreach(DAZMorph morph in morphs){
				morph.SetValue(0.001f);
			}
		}
		
		public static DAZMorph GetMorphByDisplayName(Atom atom, string name){
			JSONStorable geometry = atom.GetStorableByID("geometry");
			DAZCharacterSelector character = geometry as DAZCharacterSelector;
			GenerateDAZMorphsControlUI morphControl = character.morphsControlUI;
			return morphControl.GetMorphByDisplayName(name);
		}
		
		public static List<DAZMorph> CollectHandsPose(Atom atom){
			string[] categories = {"Pose Controls/Hands/Left", "Pose Controls/Hands/Right",  "Pose/Hands/Left","Pose/Hands/Left/Fingers", "Pose/Hands/Right", "Pose/Hands/Right/Fingers"};
			List<DAZMorph> morphs = GetAllActivePose(atom, categories);
			return morphs;
		}
		
		//for Test
		public static List<string> GetAllReceiverTarget(Atom atom){
			try{
				List<string> pairs = GetAllReceiverAtomAndTargetNamePair();
				string atomName = atom.uid;
				List<string> result = new List<string>();
				
				foreach(string pair in pairs){
					string[] kvp = pair.Split(':');
					string key = kvp[0];
					string value = kvp[1];
					if(key == atomName){
						result.Add(value);
					}
				}
				
				pairs = null;
				return result;				
			}catch(Exception e){
				LoggerE(e);
				return null;
			}
		}

		public static List<JSONNode> GetAllReceiverReference(Atom atomForFilter){
			string targetAtomName = atomForFilter.uid;
			List<JSONNode> result = new List<JSONNode>();
			
			List<Atom> atoms = SuperController.singleton.GetAtoms();			
			foreach(Atom atom in atoms){
				JSONStorable js;
				JSONClass jsc;
				List<JSONNode> refs;
				
				string[] types = {"CollisionTrigger", "Step", "Trigger", "AnimationPattern"};
				foreach(string type in types){
					js = atom.GetStorableByID(type);
					if(js != null){
						jsc = js.GetJSON();
						refs = GetReceiverReference(jsc, type, targetAtomName);
						for(int i = 0; i < refs.Count; i++){
							result.Add(refs[i]);
						}
					}
				}
			}
			return result;			
		}
		
		public static List<JSONNode> GetReceiverReference(JSONClass jsc, string type, string atomName){
			if(type == "CollisionTrigger" || type == "Trigger" || type == "Step"){
				return GetReceiverReferenceFromSingleTrigger(jsc, atomName);
			}else if(type == "AnimationPattern"){
				return GetReceiverReferenceFromMultiTriggers(jsc, atomName);				
			}else{
				return null;
			}
			
		}

		public static List<JSONNode> GetReceiverReferenceFromSingleTrigger(JSONClass jsc, string atomName){
			List<JSONNode> result = new List<JSONNode>();
			string[] actionTypes = {"startActions", "transitionActions", "endActions"};
			foreach(string actionType in actionTypes){
				JSONNode jsn = jsc["trigger"][actionType];
				if(jsn.Count != 0){
					for(int i = 0; i < jsn.Count; i++){
						string receiverName = jsn[i]["receiverAtom"]; //Needed!
						if(receiverName == atomName){
							result.Add(jsn[i]);
						}
					}
				}
			}			
			return result;
		}
		
		public static List<JSONNode> GetReceiverReferenceFromMultiTriggers(JSONClass jsc, string atomName){
			List<JSONNode> result = new List<JSONNode>();
			string[] actionTypes = {"startActions", "transitionActions", "endActions"};			
			JSONNode jsn = jsc["triggers"];
			for(int i = 0; i < jsn.Count; i++){
				foreach(string actionType in actionTypes){
					if(jsn[i][actionType].Count != 0){
						for(int j = 0; j < jsn[i][actionType].Count; j++){
							string receiverName = jsn[i][actionType][j]["receiverAtom"]; //Needed!
							if(receiverName == atomName){
								result.Add(jsn[i][actionType][j]);
							}
						}
					}
				}
			}
			return result;
		}
		
		//for test
		public static List<string> GetAllReceiverAtomAndTargetNamePair(){
			Dictionary<string, bool> hash = new Dictionary<string, bool>();
			List<string> result = new List<string>();
			
			List<Atom> atoms = SuperController.singleton.GetAtoms();
			foreach(Atom atom in atoms){
				JSONStorable js;
				JSONClass jsc;
				List<string> pairs;
				
				string[] types = {"CollisionTrigger", "Step", "Trigger", "AnimationPattern"};
				foreach(string type in types){
					js = atom.GetStorableByID(type);
					if(js != null){
						jsc = js.GetJSON();
						pairs = GetReceiverPairsFromTrigger(jsc, type);							
						foreach(string pair in pairs){
							if(hash.ContainsKey(pair)){
								continue;
							}
							hash.Add(pair, true);
							result.Add(pair);
						}
					}
				}
				
			}
			return result;
		}

		public static List<string> GetReceiverPairsFromTrigger(JSONClass jsc, string type){
			if(type == "CollisionTrigger" || type == "Trigger" || type == "Step"){
				return GetReceiverPairsFromSingleTrigger(jsc);
			}else if(type == "AnimationPattern"){
				return GetReceiverPairsFromMultiTriggers(jsc);				
			}else{
				return null;
			}
		}
		
		public static List<string> GetReceiverPairsFromSingleTrigger(JSONClass jsc){
			Dictionary<string, bool> hash = new Dictionary<string, bool>();
			List<string> result = new List<string>();
			string[] actionTypes = {"startActions", "transitionActions", "endActions"};
			foreach(string actionType in actionTypes){
				JSONNode jsn = jsc["trigger"][actionType];
				if(jsn.Count != 0){
					for(int i = 0; i < jsn.Count; i++){
						string pair = jsn[i]["receiverAtom"] + ":" + jsn[i]["receiverTargetName"];
						if(hash.ContainsKey(pair)){
							continue;
						}
						hash.Add(pair, true);
						result.Add(pair);
					}
				}
			}
			
			hash.Clear();
			return result;
		}
		
		public static List<string> GetReceiverPairsFromMultiTriggers(JSONClass jsc){
			Dictionary<string, bool> hash = new Dictionary<string, bool>();
			List<string> result = new List<string>();
			string[] actionTypes = {"startActions", "transitionActions", "endActions"};
			
			JSONNode jsn = jsc["triggers"];
			for(int i = 0; i < jsn.Count; i++){
				foreach(string actionType in actionTypes){
					if(jsn[i][actionType].Count != 0){
						for(int j = 0; j < jsn[i][actionType].Count; j++){
							string pair = jsn[i][actionType][j]["receiverAtom"] + ":" + jsn[i][actionType][j]["receiverTargetName"];
							if(hash.ContainsKey(pair)){
								continue;
							}
							hash.Add(pair, true);
							result.Add(pair);
						}
					}
				}
			}
			
			hash.Clear();
			return result;
		}
		
		public static IEnumerable<JSONStorable> GetAllPlugins(Atom atom){
			MVRPluginManager manager = GetMVRPluginManager(atom);
            if (manager != null)
            {
                var plugins = manager.GetJSON(true, true)["plugins"].AsObject;
                foreach(var pluginId in plugins.Keys)
                {
                    var receivers = atom
                        .GetStorableIDs()
                        .Where((sid) => sid.StartsWith(pluginId))
                        .Select((sid) => atom.GetStorableByID(sid));
                    foreach(var r in receivers)
                    {
                        yield return r;
                    }
                }
            }			
		} 
		
		public static JSONStorable GetReceiver(Atom atom, string pattern){
			JSONStorable targetJs = null;
			foreach(var js in GetAllPlugins(atom)){
				if(js.name.ToLower().Contains(pattern)){
					targetJs = js;
				}
			}
			return targetJs;
		}		
		 
		public static bool ContainsThisPlugin(Atom atom, string pattern){
			try{
				bool result = false;
				MVRPluginManager mvrpm = atom == null ? GetMVRPluginManager() : GetMVRPluginManager(atom);
				if(mvrpm != null){
					var pluginInfo = mvrpm.GetJSON(true, true).ToString();
					if(pluginInfo.Contains(pattern)){
						result = true;
					}
					// Utility.Debug($"plugininfo:{pluginInfo}, result:{result}, url:{pattern}");
				}
				return result;
			}catch(Exception e){
				LoggerE(e);
				return false;
			}
		}
				
		public static List<string> GetPluginIndexes(Atom atom){
			MVRPluginManager mvrpm = GetMVRPluginManager(atom);
			List<string> indexes = new List<string>();
			if(mvrpm != null){
				JSONClass jsc = mvrpm.GetJSON(true, true).AsObject;
				foreach(var key in jsc["plugins"].AsObject.Keys){
					indexes.Add(key);
				}
			}
			return indexes;
		}
		
		public static Dictionary<string, string> GetPluginsUrlDictionary(Atom atom, bool bInverse = false){
			MVRPluginManager mvrpm = GetMVRPluginManager(atom);
			Dictionary<string, string> dic = new Dictionary<string, string>();
			
			if(mvrpm != null){
				JSONClass jsc = mvrpm.GetJSON(true, true).AsObject;
				foreach(var key in jsc["plugins"].AsObject.Keys){
					string value = jsc["plugins"][key];
					if(!bInverse){
						dic.Add(key, value);
					}else{
						if(!dic.ContainsKey(value)){
							dic.Add(value, key);
						}else{
							dic[value] = key;
						}
					}
				}
			}
			return dic;
		}
		
		public static void ReplaceNewOneAndReloadPlugins(Atom atom, string pattern){
			Dictionary<string, string> dicUrl = GetPluginsUrlDictionary(atom, bInverse:false);
			if(atom == null){ return; }
			
			List<MVRScript> scripts = GetMVRScripts(atom);
			foreach(var script in scripts){
				string id = script.GetJSON()["id"];
				
				if(id.Contains(pattern)){
					Match m = Regex.Match(id, "^(.+?)_(.+)$");
					string key = m.Groups[1].Value;
					string value = m.Groups[2].Value;				
					string url = dicUrl[key];
					
					string searchPattern = "*" + pattern + "*.var";
					string latest = GetLongestFilename(searchPattern);
					Match m1 = Regex.Match(latest, "^AddonPackages\\\\(.+)\\.var$");
					string latestFilename = m1.Groups[1].ToString();
					
					Match m2 = Regex.Match(url, "^.+?(:.+)$");
					string urlOldSuffix = m2.Groups[1].ToString();
					string urlNew = latestFilename + urlOldSuffix;
					
					MVRPluginManager mvrpm = GetMVRPluginManager(atom);
					JSONClass jsc = mvrpm.GetJSON();
					mvrpm.LateRestoreFromJSON(jsc);
				}
			}
		}
		
		public static void PrintPluginInfo(Atom atom, string pattern){
			MVRPluginManager mvrpm = GetMVRPluginManager(atom);
			if(mvrpm != null){
				var pluginInfo = mvrpm.GetJSON(true, true).ToString();
				Info("pluginId:" + pluginInfo);
			}
		}
		
		public static void PrintStorableID(Atom atom){
			List<string> ids = atom.GetStorableIDs();
			foreach(var id in ids){
				Info("id:" + id);
				JSONStorable js = atom.GetStorableByID(id);
				
			}
		}
		
		public static bool HasThisPlugin(Atom atom, string pattern){
			MVRPluginManager mvrpm = atom != null ? GetMVRPluginManager(atom) : GetMVRPluginManager();
			return HasThisPlugin(mvrpm, pattern);
		}

		public static bool HasThisPlugin(MVRPluginManager mvrpm, string pattern){
			bool result = false;
			if(mvrpm != null){
				var pluginInfo = mvrpm.GetJSON(true, true).ToString();
				Logger("pluginId:" + pluginInfo);
				var N = JSON.Parse(pluginInfo);
				for(int i = 0; i < N["plugins"].Count; i++){
					string name = N["plugins"][i].ToString();
					if(name.ToLower().Contains(pattern.ToLower())){
						result = true;
					}
				}
			}
			return result;
		}



		
		public static bool IsFemale(Atom atom){
			JSONStorable geometry = atom.GetStorableByID("geometry");
			DAZCharacterSelector character = geometry as DAZCharacterSelector;
			
			return (character.gender.ToString() == "Female") ? true : false;
		}

		public static void ReloadPlugins(Atom atom){
			MVRPluginManager mvrpm = GetMVRPluginManager(atom);
			if(mvrpm == null){return;}
			JSONClass jsc = mvrpm.GetJSON();
			mvrpm.LateRestoreFromJSON(jsc);
		}
		
		public static void AppendAndReloadPlugins(MVRPluginManager mvrpm, Atom atom, string url, int number = 99){
			if(mvrpm == null){return;}
			
			if(ContainsThisPlugin(atom, url)){
				return;
			}
			
			string numstr = (number < 10) ? "plugin#0" + number : "plugin#" + number;
			if(ContainsThisPlugin(atom, numstr)){
				return;
			}
			JSONClass jsc = mvrpm.GetJSON();

			jsc["plugins"].Add(numstr, url);
			mvrpm.LateRestoreFromJSON(jsc, true, true, true);
		}

		public static void AppendAndReloadPlugins(Atom atom, string url, int number = 99){
			MVRPluginManager mvrpm = atom == null ? GetMVRPluginManager() : GetMVRPluginManager(atom);
			AppendAndReloadPlugins(mvrpm, atom, url, number);
		}

		public static void AppendAndReloadPlugins(string url, int number = 99){
			MVRPluginManager mvrpm = GetMVRPluginManager();
			AppendAndReloadPlugins(mvrpm, null, url, number);
		}

		public static string GetEBATFilePath(){
			string rawFilePath = "Custom/Scripts/cotyounoyume/ExpressionBlushingAndTears/ADD_ME.cslist";
			string basePathNormal = "Custom/Scripts/cotyounoyume/ExpressionBlushingAndTears/ADD_ME.cslist";
			string basePathPaused = "Custom/Scripts/cotyounoyume/ExpressionBlushingAndTears/PAUSED_ADD_ME.cslist";
			string fullFilename = "*cotyounoyume.ExpressionBlushingAndTearsFullVer.*";
			string freeFilename = "*cotyounoyume.ExpressionBlushingAndTears.*";

			string file = null;
			string fullPath = null;

			string resultPath = null;
			if(FileManagerSecure.FileExists(rawFilePath)){
				resultPath = rawFilePath;
			}else{
				if(GetMaxVerPackage(fullFilename) != null){
					file = GetMaxVerPackage(fullFilename);
					fullPath = file + ":/" + basePathPaused;
					if(FileManagerSecure.FileExists(file + ":/" + basePathPaused)){
						resultPath = file + ":/" + basePathPaused;
					}else if(FileManagerSecure.FileExists(file + ":/" + basePathNormal)){
						resultPath = file + ":/" + basePathNormal;
					}
				}else if(GetMaxVerPackage(freeFilename) != null){
					file = GetMaxVerPackage(freeFilename);
					fullPath = file + ":/" + basePathPaused;
					if(FileManagerSecure.FileExists(file + ":/" + basePathPaused)){
						resultPath = file + ":/" + basePathPaused;
					}else if(FileManagerSecure.FileExists(file + ":/" + basePathNormal)){
						resultPath = file + ":/" + basePathNormal;
					}					
				}
			}

			return resultPath;
		}

		public static Vector3 GetDirection(Atom atom, string fromControl, string toControl){
			FreeControllerV3 fromC = GetFreeControllerV3(atom, fromControl);
			FreeControllerV3 toC = GetFreeControllerV3(atom, toControl);
			Vector3 direction = GetPosition(toC) - GetPosition(fromC);
			return direction;
		}
		
		public static Vector3 GetDirection(Transform from, Transform to){
			Vector3 direction = GetPosition(to) - GetPosition(from);
			return direction;
		}
		
		public static string GetMaxVerPackage(string pattern){
			string maxFile = null;
			int maxVer = 0;
			string[] files = FileManagerSecure.GetFiles("AddonPackages", pattern);
			foreach(var file in files){
				string[] words = file.Split('.');
				int ver = 0;
				ver = Int32.Parse(words[2]);
				if(ver > maxVer){
					maxFile = file;
					maxVer = ver;
				}
			}
			return maxFile;
		}
		
		public static float Lerp3(float start, float end, float delta){
			//f(t) = t^2 * (3 - 2*t)
			float std_delta = delta / (end - start);
			float progress = (std_delta * std_delta) * (3 - 2 * std_delta);
			return progress;
		}

		
		public static string ResolveStageName(int stage, int count){
			string name = (stage == 5) ? "Stage=" : "Stage<";
			name += (stage < 3) ? "3" : "5";
			if(stage == 5){
				int threshold = 99;
				if(count < 1){ threshold = 1; }
				else if(count < 3){ threshold = 3; }
				else if(count < 6){ threshold = 6; }
				else if(count < 9){ threshold = 9; }
				name += ":OrgasmCount<" + ((threshold == 99) ? threshold.ToString() : ("0" + threshold.ToString()));
			}else{
				name += ":OrgasmCount<99";
			}
			return name;
		}
		
		public static Quaternion SmoothDampQuaternion(Quaternion current, Quaternion target, ref Vector3 currentVelocity, float smoothTime)
		{
		  Vector3 c = current.eulerAngles;
		  Vector3 t = target.eulerAngles;
		  return Quaternion.Euler(
			Mathf.SmoothDampAngle(c.x, t.x, ref currentVelocity.x, smoothTime),
			Mathf.SmoothDampAngle(c.y, t.y, ref currentVelocity.y, smoothTime),
			Mathf.SmoothDampAngle(c.z, t.z, ref currentVelocity.z, smoothTime)
		  );
		}		
		
		public static Vector3 SmoothDampEuler(Vector3 current, Vector3 target, ref Vector3 currentVelocity, float smoothTime)
		{
		  return new Vector3(
			Mathf.SmoothDampAngle(current.x, target.x, ref currentVelocity.x, smoothTime),
			Mathf.SmoothDampAngle(current.y, target.y, ref currentVelocity.y, smoothTime),
			Mathf.SmoothDampAngle(current.z, target.z, ref currentVelocity.z, smoothTime)
		  );
		}

		public static void OpenUIByName(Atom atom, string uiName, bool IsMainUI = false){
			ChangeGameModeIfNeeded();
			SuperController.singleton.SelectController(atom.mainController);
			SuperController.singleton.ShowMainHUD();
			
			UITabSelector tab = atom.GetComponentInChildren<UITabSelector>();
			tab.SetActiveTab(uiName);
		}

		public static bool OpenCustomUI(Atom atom, string pluginNamePattern){
			ChangeGameModeIfNeeded();
			SuperController.singleton.SelectController(atom.mainController);
			SuperController.singleton.ShowMainHUD();
			
			UITabSelector tab = atom.GetComponentInChildren<UITabSelector>();
			tab.SetActiveTab("Plugins");
			
			Transform initialTransform = atom.transform;
			string path = "UI/UIPlaceHolderModel/UIModel/Canvas/Panel/Content/Plugins/Scroll View/Viewport/Content/PluginPanel(Clone)";
			Transform target = initialTransform.Find(path);
			
			bool result = false;
			foreach(Transform pluginPanelT in target.parent){
				Transform uidT = pluginPanelT.Find("Content/ScriptControllerPanel(Clone)/UID");
				if(uidT == null){ continue; }
				
				string uid = uidT.GetComponent<Text>().text;
				if(uid.Contains(pluginNamePattern)){
					Transform openUIButtonT = pluginPanelT.Find("Content/ScriptControllerPanel(Clone)/OpenUIButton");
					openUIButtonT.GetComponent<Button>().onClick.Invoke();
					result = true;
				}
			}
			
			return result;
		}
		
		public static bool ReloadButtonClick(Atom atom, string pluginNamePattern){
			Transform initialTransform = atom.transform;
			string path = "UI/UIPlaceHolderModel/UIModel/Canvas/Panel/Content/Plugins/Scroll View/Viewport/Content/PluginPanel(Clone)";

			Transform target = initialTransform.Find(path);
			
			if(target == null){
				return false;
			}
			
			bool result = false;
			foreach(Transform pluginPanelT in target.parent){
				Transform uidT = pluginPanelT.Find("Content/ScriptControllerPanel(Clone)/UID");
				if(uidT == null){ continue; }
				
				string uid = uidT.GetComponent<Text>().text;
				if(uid.Contains(pluginNamePattern)){
					Transform ReloadButtonT = pluginPanelT.Find("ReloadButton");
					ReloadButtonT.GetComponent<Button>().onClick.Invoke();
					result = true;
				}
			}
			
			return result;
		}
		
		public static bool RemoveButtonClickOnSessionPlugin(string pluginNamePattern){
			Transform initialTransform = SuperController.singleton.mainMenuUI.transform;
			string path = "Panel/Content/TabSessionPlugins/Scroll View/Viewport/Content/PluginPanel(Clone)";

			Transform target = initialTransform.Find(path);
			if(target == null){
				return false;
			}
			
			bool result = false;
			foreach(Transform pluginPanelT in target.parent){
				Transform uidT = pluginPanelT.Find("Content/ScriptControllerPanel(Clone)/UID");
				if(uidT == null){ continue; }
				
				string uid = uidT.GetComponent<Text>().text;
				if(uid.ToLower().Contains(pluginNamePattern.ToLower())){
					Transform RemoveButtonT = pluginPanelT.Find("RemoveButton");
					RemoveButtonT.GetComponent<Button>().onClick.Invoke();
					result = true;
				}
			}
			// Info($"ERROR: can't find pattern:{pluginNamePattern} (Probably already deleted?).");
			result = true;
			return result;
		}
		
		public static Transform GetTimelineScriptUIRootTransform(Atom atom){
			Transform result = null;
			
			string initialPath = "UI/UIPlaceHolderModel/UIModel/Canvas/Panel/Content/Plugins";
			Transform startT = atom.transform.Find(initialPath);

			foreach(Transform t in startT){
				if(t.name == "ScriptUI(Clone)"){
					foreach(var text in t.GetComponentsInChildren<Text>()){
						if(text.text == "▶ All"){
							result = t;
							break;
						}
					}
				}
			}

			return result;
		}
		
		public static Transform GetTransformByPath(Atom atom, string path){
			return atom.transform.Find(path);
		}
		
		public static Transform GetTransformByPath(Transform transform, string path){
			return transform.Find(path);
		}
		
		public static void SurveyTransform(Transform t, string filter = ""){
			SurveyTransform(t, "", "", filter);
		}
		
		public static string GetTransformPath(Transform t){
			string path = "";
			if(t == null){ return path; }
			
			path = "/" + t.gameObject.name;
			while(t.parent != null){
				path = "/" + t.parent.gameObject.name + path; 
				t = t.parent;
			}
			return path;
		}
		
		public static void SurveyTransform(Transform t, string indent, string path, string filter){
			string indentBase = "    ";
			Text text = t.GetComponentInChildren<Text>();
			string label = (text != null) ? text.text : "";
			path = path + "/" + t.name;
			if(filter == ""){
				Info(indent + "name:" + t.name + ", class:" + t + ", label:" + label);
				Info(indent + "path:" + path);
			}else{
				if(t.name.Contains(filter) || label.Contains(filter)){
					Info(indent + "name:" + t.name + ", class:" + t + ", label:" + label);
					Info(indent + "path:" + path);
				}
			}
			foreach(Transform child in t){
				if(child.name != "Text"){
					SurveyTransform(child, indent + indentBase, path, filter);
				}
			}
		}
		
		public static Transform GetTimelineHeadersParent(Atom atom){
			Transform result = null;
			var items = atom.gameObject.GetComponentsInChildren<RectTransform>();
			foreach(var item in items){
				if(item.name == "Header"){
					result = item.parent;
				}
			}
			return result;
		}
		
		public static MVRScript GetSessionPlugin(string pluginNamePattern){
			MVRScript result = null;
			List<MVRScript> plugins = GetSessionPlugins();
			foreach(MVRScript plugin in plugins){
				string fullname = plugin.ToString();
				if(fullname.Contains(pluginNamePattern)){
					result = plugin;
				}
			}
			return result;
		}
		
		public static List<MVRScript> GetSessionPlugins(){
			List<MVRScript> results = new List<MVRScript>();
			foreach(MVRScript script in SuperController.singleton.transform.GetComponentsInChildren<MVRScript>(true)){
				results.Add(script);
			}
			
			return results;
		}
		
		public static void DumpRigidBody(Atom atom)
		{
			Debug($"var dPosW = new Dictionary<string, Vector3>();");
			Debug($"var dPosL = new Dictionary<string, Vector3>();");
			Debug($"var dRotW = new Dictionary<string, Vector3>();");
			Debug($"var dRotL = new Dictionary<string, Vector3>();");
			Debug($"var dQuaW = new Dictionary<string, Quaternion>();");
			Debug($"var dQuaL = new Dictionary<string, Quaternion>();");

			foreach(var rb in atom.linkableRigidbodies)

			{
				var t = rb.transform;
				var name = t.name;
				if(name.ToLower().EndsWith("control")) { continue; }
				if(name.ToLower().StartsWith("hair")) { continue; }
				if(name == "Genesis2Female") { continue; }
				var posW = t.position;
				var posL = t.localPosition;
				var rotW = t.eulerAngles;
				var rotL = t.localEulerAngles;
				var quaW = t.rotation;
				var quaL = t.localRotation;

				DumpVector3ForCSharp("dPosW", name, posW);
				DumpVector3ForCSharp("dPosL", name, posL);
				DumpVector3ForCSharp("dRotW", name, rotW);
				DumpVector3ForCSharp("dRotL", name, rotL);
				DumpQuaternionForCSharp("dQuaW", name, quaW);
				DumpQuaternionForCSharp("dQuaL", name, quaL);
			}			
		}

		public static void DumpVector3ForCSharp(string varName, string name, Vector3 v)
		{
			Debug($"{varName}.Add(\"{name}\", new Vector3({v.x:F6}f, {v.y:F6}f, {v.z:F6}f));");
		}

		public static void DumpQuaternionForCSharp(string varName, string name, Quaternion q)
		{
			Debug($"{varName}.Add(\"{name}\", new Quaternion({q.x:F6}f, {q.y:F6}f, {q.z:F6}f, {q.w:F6}f));");
		}

		public static void LogSpacer(int count = 8, string level = "Debug"){
			for(int i = 0; i < count; i++){
				if(level == "Debug")
				{
					Debug("****");
				}
				else
				{
					Info("****");
				}
			}
		}
		
		public static string GetLongestFilename(string pattern){
			string[] files = FileManagerSecure.GetFiles("AddonPackages", pattern);
			if(files.Length == 1){
				return files[0];
			}else if(files.Length == 0){
				return "";
			}else{
				string filename = "";
				for(int i = 0; i < files.Length - 1; i++){
					string longer = CompareString(files[i], files[i+1]);
					filename = CompareString(longer, filename);
				}
				return filename;
			}
		}

		// public static string GetPluginVarName(string pattern, string path){
		// 	var filename = GetLongestFilename(pattern);
		// 	if(filename == "") return "";
		// 	var varname = CorrectVarname(filename);
		// 	Utility.Debug($"filename:{filename}, varname:{varname}");
		// 	return varname + ":\\" + path;

		// }

		public static string GetPluginURL(string pattern, string path){
			var filename = GetLongestFilename(pattern);
			if(filename == "") return "";
			var varname = CorrectVarname(filename);
			var url = (varname + ":\\" + path).Replace('\\', '/');
			return url;
		}

		public static void LoadPluginIfNeeded(Atom atom, string url, string shortName, int num = 97){
			if(HasThisPlugin(atom, shortName)){ 
				return; 
			}
			AppendAndReloadPlugins(atom, url, number:num);
		}
		
		public static string GetVAMMoanURL(){
			return GetLongestFilename("*vammoan*") + ":\\" + "Custom\\Scripts\\VAMMoan\\VAMMoan.cslist";
		}
		
		public static void LoadVAMMoanIfNeeded(Atom atom){
			string url = "Custom\\Scripts\\VAMMoan\\VAMMoan.cslist";
			if(Utility.HasThisPlugin(atom, "VAMMoan")){ return; }
			string varname = GetLongestFilename("*vammoan*");
			if(varname == ""){ return; }
			
			string fullpath = varname + ":\\" + url;
			AppendAndReloadPlugins(atom, fullpath, number:98);
		}
		
		public static Atom GetAtomWithPluginNamePattern(string pattern, Atom targetAtom = null){
			pattern = pattern.ToLower();
			var atoms = targetAtom == null ? SuperController.singleton.GetAtoms() : new List<Atom>(){targetAtom};
			foreach(var atom in atoms){
				foreach(var pluginName in GetPluginList(atom)){
					// Utility.Debug($"pluginName:{pluginName}, pattern:{pattern}");
					if(!pluginName.ToLower().Contains(pattern)) continue;
					// Utility.Debug($"  Hit!: pluginName:{pluginName}, pattern:{pattern}");
					return atom;
				}
			}
			return null;
		}

		public static string GetPluginNameWithPluginNamePattern(string pattern, Atom targetAtom = null){
			pattern = pattern.ToLower();
			var atoms = targetAtom == null ? SuperController.singleton.GetAtoms() : new List<Atom>(){targetAtom};
			foreach(var atom in atoms){
				foreach(var pluginName in GetPluginList(atom)){
					if(!pluginName.ToLower().Contains(pattern)) continue;
					return pluginName;
				}
			}
			return null;
		}

		// public static Atom GetAtomAndPluginPairFromAtoms(string pattern, Atom targetAtom = null){
		// 	pattern = pattern.ToLower();
		// 	var atoms = targetAtom == null ? SuperController.singleton.GetAtoms() : new List<Atom>(){targetAtom};
		// 	foreach(var atom in atoms){
		// 		foreach(var pluginName in GetPluginList(atom)){
		// 			if(!pluginName.ToLower().Contains(pattern)) continue;
		// 			return new AtomAndPluginName(atom, pluginName);
		// 		}
		// 	}
		// 	return null;
		// }

		public static List<string> GetPluginsByPattern(Atom targetAtom, string pattern){
			var list = new List<string>();
			pattern = pattern.ToLower();
			foreach(var pluginName in GetPluginList(targetAtom)){
				if(!pluginName.ToLower().Contains(pattern)) continue;
				list.Add(pluginName);
			}

			return list;
		}

		public static List<string> GetPluginList(Atom atom){
			List<string> result = new List<string>();
			List<string> ids = atom.GetStorableIDs();
			Dictionary<string, bool> dic = new Dictionary<string, bool>();
			foreach(var id in ids){
				if(!id.StartsWith("plugin#")) continue;
				dic.Add(id, true);
			}

			return dic.Keys.ToList<string>();
		}

		public static bool GetOnOffPlugin(string pattern){
			var atom = GetAtomWithPluginNamePattern(pattern);
			if(atom == null) return false;
			var pluginName = GetPluginNameWithPluginNamePattern(pattern);
			JSONStorable jss = atom.GetStorableByID(pluginName);
			return jss.GetBoolParamValue("enabled");
		}

		public static bool GetOnOffPlugin(Atom atom, string pattern){
			var pluginName = GetPluginNameWithPluginNamePattern(pattern, atom);
			JSONStorable jss = atom.GetStorableByID(pluginName);
			return jss.GetBoolParamValue("enabled");
		}

		public static void ToggleOnOffPlugin(Atom atom, string pattern, bool isOn){
			var pluginName = GetPluginNameWithPluginNamePattern(pattern, atom);
			JSONStorable jss = atom.GetStorableByID(pluginName);
			bool currentStatus = jss.GetBoolParamValue("enabled");
			jss.SetBoolParamValue("enabled", isOn);
		}

		public static void ToggleOnOffPlugin(string pattern, bool isOn){
			var atom = GetAtomWithPluginNamePattern(pattern);
			if(atom == null) return;
			var pluginName = GetPluginNameWithPluginNamePattern(pattern);

			JSONStorable jss = atom.GetStorableByID(pluginName);
			bool currentStatus = jss.GetBoolParamValue("enabled");
			jss.SetBoolParamValue("enabled", isOn);
		}

		public static string NormalizePluginName(string orig, bool prefix = false){
			string[] pattern = {"_"};
			string[] words = orig.Split(pattern, StringSplitOptions.None);
			string pluginPrefix = words[0];
			string pluginName = orig.Remove(0, pluginPrefix.Length+1);
			return (prefix) ? pluginPrefix : pluginName;
		}

		public static string CompareString(string left, string right){
			if(left.Length == right.Length){
				return (left.CompareTo(right) > 0) ? left : right;
			}else{
				return (left.Length > right.Length) ? left : right;
			}
		}
		
		public static Transform GetBlankUID(Atom atom){
			Transform result = null;
			
			Transform initialTransform = atom.transform;
			string path = "UI/UIPlaceHolderModel/UIModel/Canvas/Panel/Content/Plugins/Scroll View/Viewport/Content/PluginPanel(Clone)";
			Transform target = initialTransform.Find(path);
			
			foreach(Transform pluginPanelT in target.parent){
				Transform uidT = pluginPanelT.Find("Content/ScriptControllerPanel(Clone)/UID");
				if(uidT == null){
					result = pluginPanelT;
				}
			}
			return result;
		}
				
		public static ClothSimControl GetClothSimControl(DAZClothingItem dci){
			return dci.GetComponentInChildren<ClothSimControl>();
		}

		public static ClothSettings GetClothSettings(DAZClothingItem dci){
			ClothSimControl csc = GetClothSimControl(dci);
			if(csc != null){
				return csc.clothSettings;
			}else{
				return null;
			}
		}
				
		public static ClothSettings GetClothSettings(ClothSimControl csc){
			if(csc != null){
				return csc.clothSettings;
			}else{
				return null;
			}
		}
		
		public static RuntimeData GetRuntime(DAZClothingItem dci){
			ClothSettings cs = GetClothSettings(dci);
			if(cs != null){
				return cs.Runtime;
			}else{
				return null;
			}
		}
		
		public static bool HasRuntime(DAZClothingItem dci){
			ClothSettings cs = GetClothSettings(dci);
			if(cs != null){
				return (cs.Runtime != null);
			}else{
				return false;
			}
		}
		
		public static Atom GetCustomUnityAssetByName(string name){
			Atom result = null;
			foreach(var atom in SuperController.singleton.GetAtoms()){
				if(atom.uid.Contains(name)){
					result = atom;
				}
			}
			return result;
		}
		
		public static string GetGender(Atom atom){
			JSONStorable geometry = atom.GetStorableByID("geometry");
			DAZCharacterSelector character = geometry as DAZCharacterSelector;
			return character.gender.ToString();
		}

		public static void AddBlankUIDIfNeeded(Atom atom){
			Transform target = GetBlankUID(atom);
			if(target == null){
				Transform AddPluginButtonT = atom.transform.Find("UI/UIPlaceHolderModel/UIModel/Canvas/Panel/Content/Plugins/AddPluginButton");
				AddPluginButtonT.GetComponent<Button>().onClick.Invoke();
			}
		}
		
		public static bool GetMemoryBool(Atom memoryAtom, string param, bool readFromLoadJson, bool debugOn = false){
			bool result = false;
			if(readFromLoadJson){
				JSONClass jsc = GetAtomControl(memoryAtom);
				result = jsc[param].AsBool;
				if(debugOn){
					Info("GetMemoryBool: param:" + param + ", jscIsNotNull:" + (jsc != null) + ", value:" + result);
				}
			}else{
				bool bExist = memoryAtom.IsBoolJSONParam(param);
				result  = memoryAtom.GetBoolParamValue(param);
				if(debugOn){
					Info("GetMemoryBool: param:" + param + ", bExist:" + bExist + ", value:" + result);
				}
			}
			
			return result;
		}
		
		public static void ChangeDefaultValueBool(Atom memoryAtom, string param, bool value, bool debugOn = false){
			bool bExist = memoryAtom.IsBoolJSONParam(param);
			if(bExist){
				JSONStorableBool jsb = memoryAtom.GetBoolJSONParam(param);
				if(debugOn){
					bool defaultValOld = jsb.defaultVal;
					Info("ChangeDefaultValueBool: param:" + param + ", bExist:" + bExist + ", defaultValOld:" + defaultValOld + ", defaultValNew:" + value);
				}				
				jsb.defaultVal = value;
			}else{
				SetMemoryBool(memoryAtom, param, false, value, debugOn);
				if(debugOn){
					Info("ChangeDefaultValueBool: param:" + param + ", bExist:" + bExist + ", defaultValOld:None, defaultValNew:" + value);
				}				
			}
		}
		
		public static void SetMemoryBool(Atom memoryAtom, string param, bool readFromLoadJson, bool value, bool debugOn = false){
			bool bExist = memoryAtom.IsBoolJSONParam(param);
			if(bExist){
				memoryAtom.SetBoolParamValue(param, value);
			}else{
				JSONStorableBool jsb = new JSONStorableBool(param, value);
				memoryAtom.RegisterBool(jsb);
			}
			if(debugOn){
				Info("SetMemoryBool: param:" + param + ", bExist:" + bExist + ", value:" + value);
			}
		}
		
		public static void SetMemoryBoolIfNull(Atom memoryAtom, string param, bool readFromLoadJson, bool value, bool debugOn = false){
			if(HasMemoryBool(memoryAtom, param, readFromLoadJson, debugOn)){
				bool inputValue = GetMemoryBool(memoryAtom, param, readFromLoadJson, debugOn);
				SetMemoryBool(memoryAtom, param, readFromLoadJson, inputValue, debugOn);
				if(debugOn){
					Info("SetMemoryBoolIfNull: param:" + param + ", notNull:true, inputValue:" + inputValue);
				}
			}else{
				SetMemoryBool(memoryAtom, param, readFromLoadJson, value, debugOn);
				if(debugOn){
					Info("SetMemoryBoolIfNull: param:" + param + ", notNull:false, value:" + value);
				}
			}
		}		
		
		public static bool HasMemoryBool(Atom memoryAtom, string param, bool readFromLoadJson, bool debugOn = false){
			bool result = false;
			if(readFromLoadJson){
				JSONClass jsc = GetAtomControl(memoryAtom);
				if(jsc[param] != null){
					result = true;
				}
				if(debugOn){
					Info("HasMemoryBool: param:" + param + ", readFromLoadJson:" + readFromLoadJson + ", result:" + result);
				}
			}else{
				bool bExist = memoryAtom.IsBoolJSONParam(param);
				result  = bExist;
				if(debugOn){
					Info("HasMemoryBool: param:" + param + ", readFromLoadJson:" + readFromLoadJson + ", result:" + result);
				}
			}
			return result;
		}
		
		public static float GetMemoryFloat(Atom memoryAtom, string param, bool readFromLoadJson, bool debugOn = false){
			float result = 0f;
			if(readFromLoadJson){
				JSONClass jsc = GetAtomControl(memoryAtom);
				result = jsc[param].AsFloat;
				if(debugOn){
					Info("GetMemoryFloat: param:" + param + ", jscIsNotNull:" + (jsc != null) + ", value:" + result);
				}
			}else{
				bool bExist = memoryAtom.IsBoolJSONParam(param);
				result  = memoryAtom.GetFloatParamValue(param);
				if(debugOn){
					Info("GetMemoryBool: param:" + param + ", bExist:" + bExist + ", value:" + result);
				}
			}
			
			return result;
		}
		
		public static void ChangeDefaultValueFloat(Atom memoryAtom, string param, float value, bool debugOn = false){
			bool bExist = memoryAtom.IsFloatJSONParam(param);
			if(bExist){
				JSONStorableFloat jsb = memoryAtom.GetFloatJSONParam(param);
				if(debugOn){
					float defaultValOld = jsb.defaultVal;
					Info("ChangeDefaultValueBool: param:" + param + ", bExist:" + bExist + ", defaultValOld:" + defaultValOld + ", defaultValNew:" + value);
				}				
				jsb.defaultVal = value;
			}else{
				SetMemoryFloat(memoryAtom, param, false, value, debugOn);
				if(debugOn){
					Info("ChangeDefaultValueBool: param:" + param + ", bExist:" + bExist + ", defaultValOld:None, defaultValNew:" + value);
				}				
			}
		}
		
		public static void SetMemoryFloat(Atom memoryAtom, string param, bool readFromLoadJson, float value, bool debugOn = false){
			bool bExist = memoryAtom.IsFloatJSONParam(param);
			if(bExist){
				memoryAtom.SetFloatParamValue(param, value);
			}else{
				JSONStorableFloat jsb = new JSONStorableFloat(param, value, -1000f, 1000f);
				memoryAtom.RegisterFloat(jsb);
			}
			if(debugOn){
				Info("SetMemoryBool: param:" + param + ", bExist:" + bExist + ", value:" + value);
			}
		}
		
		public static void DeleteMemoryFloat(Atom memoryAtom, string param, bool debugOn = false){
			JSONStorableFloat jsb = new JSONStorableFloat(param, 0f, -1000f, 1000f);
			memoryAtom.DeregisterFloat(jsb);
			if(debugOn){
				Info("DeleteMemoryFloat: param:" + param);
			}
		}
		
		public static void SetMemoryFloatIfNull(Atom memoryAtom, string param, bool readFromLoadJson, float value, bool debugOn = false){
			if(HasMemoryFloat(memoryAtom, param, readFromLoadJson, debugOn)){
				float inputValue = GetMemoryFloat(memoryAtom, param, readFromLoadJson, debugOn);
				SetMemoryFloat(memoryAtom, param, readFromLoadJson, inputValue, debugOn);
				if(debugOn){
					Info("SetMemoryFloatIfNull: param:" + param + ", notNull:true, inputValue:" + inputValue);
				}
			}else{
				SetMemoryFloat(memoryAtom, param, readFromLoadJson, value, debugOn);
				if(debugOn){
					Info("SetMemoryFloatIfNull: param:" + param + ", notNull:false, value:" + value);
				}
			}
		}		
		
		public static bool HasMemoryFloat(Atom memoryAtom, string param, bool readFromLoadJson, bool debugOn = false){
			bool result = false;
			if(readFromLoadJson){
				JSONClass jsc = GetAtomControl(memoryAtom);
				if(jsc[param] != null){
					result = true;
				}
				if(debugOn){
					Info("HasMemoryFloat: param:" + param + ", readFromLoadJson:" + readFromLoadJson + ", result:" + result);
				}
			}else{
				bool bExist = memoryAtom.IsFloatJSONParam(param);
				result  = bExist;
				if(debugOn){
					Info("HasMemoryFloat: param:" + param + ", readFromLoadJson:" + readFromLoadJson + ", result:" + result);
				}
			}
			return result;
		}

		public static string GetMemoryString(Atom memoryAtom, string param, bool readFromLoadJson, bool debugOn = false){
			string result = "";
			if(readFromLoadJson){
				JSONClass jsc = GetAtomControl(memoryAtom);
				result = jsc[param];
				if(debugOn){
					Info("GetMemoryString: param:" + param + ", jscIsNotNull:" + (jsc != null) + ", value:" + result);
				}
			}else{
				bool bExist = memoryAtom.IsStringJSONParam(param);
				result  = memoryAtom.GetStringParamValue(param);
				if(debugOn){
					Info("GetMemoryString: param:" + param + ", bExist:" + bExist + ", value:" + result);
				}
			}
			
			return result;
		}
		
		public static void ChangeDefaultValueString(Atom memoryAtom, string param, string value, bool debugOn = false){
			bool bExist = memoryAtom.IsStringJSONParam(param);
			if(bExist){
				JSONStorableString jss = memoryAtom.GetStringJSONParam(param);
				if(debugOn){
					string defaultValOld = jss.defaultVal;
					Info("ChangeDefaultValueString: param:" + param + ", bExist:" + bExist + ", defaultValOld:" + defaultValOld + ", defaultValNew:" + value);
				}				
				jss.defaultVal = value;
			}else{
				SetMemoryString(memoryAtom, param, false, value, debugOn);
				if(debugOn){
					Info("ChangeDefaultValueString: param:" + param + ", bExist:" + bExist + ", defaultValOld:None, defaultValNew:" + value);
				}				
			}
		}
		
		public static void SetMemoryString(Atom memoryAtom, string param, bool readFromLoadJson, string value, bool debugOn = false){
			bool bExist = memoryAtom.IsStringJSONParam(param);
			if(bExist){
				memoryAtom.SetStringParamValue(param, value);
			}else{
				JSONStorableString jss = new JSONStorableString(param, value);
				memoryAtom.RegisterString(jss);
			}
			if(debugOn){
				Info("SetMemoryBool: param:" + param + ", bExist:" + bExist + ", value:" + value);
			}
		}
		
		public static void SetMemoryStringIfNull(Atom memoryAtom, string param, bool readFromLoadJson, string value, bool debugOn = false){
			// Debug($"routeY1");
			// debugOn = true;
			if(HasMemoryString(memoryAtom, param, readFromLoadJson, debugOn)){
				string inputValue = GetMemoryString(memoryAtom, param, readFromLoadJson, debugOn);
				// Debug($"routeY2, inputValue:{inputValue}");
				SetMemoryString(memoryAtom, param, readFromLoadJson, inputValue, debugOn);
				if(debugOn){
					Info("SetMemoryStringIfNull: param:" + param + ", notNull:true, inputValue:" + inputValue);
				}
			}else{
				// Debug($"routeY3");
				SetMemoryString(memoryAtom, param, readFromLoadJson, value, debugOn);
				if(debugOn){
					Info("SetMemoryStringIfNull: param:" + param + ", notNull:false, value:" + value);
				}
			}
			// Debug($"routeY4");
		}		
		
		public static bool HasMemoryString(Atom memoryAtom, string param, bool readFromLoadJson, bool debugOn = false){
			bool result = false;
			if(readFromLoadJson){
				JSONClass jsc = GetAtomControl(memoryAtom);
				if(jsc[param] != null){
					result = true;
				}
				if(debugOn){
					Info("HasMemoryString: param:" + param + ", readFromLoadJson:" + readFromLoadJson + ", result:" + result);
				}
			}else{
				bool bExist = memoryAtom.IsStringJSONParam(param);
				result  = bExist;
				if(debugOn){
					Info("HasMemoryString: param:" + param + ", readFromLoadJson:" + readFromLoadJson + ", result:" + result);
				}
			}
			return result;
		}


		public static JSONClass GetAtomControl(Atom memoryAtom){
			JSONClass atomJsc = GetJSONClassFromArrayTypeJSONNodeByID(SuperController.singleton.loadJson["atoms"], memoryAtom.uid);
			if(atomJsc == null){ return null; }
				
			JSONClass atomControlJsc = GetJSONClassFromArrayTypeJSONNodeByID(atomJsc["storables"], "AtomControl");
			return atomControlJsc;
		}
		
		public static JSONClass GetJSONClassFromArrayTypeJSONNodeByID(JSONNode jsnRoot, string id){
			JSONClass targetJsc = null;
			foreach(JSONNode jsn in jsnRoot.AsArray){
				JSONClass jsc = jsn.AsObject;
				string jscid = jsc["id"];
				if(jscid == id){
					targetJsc = jsc;
				}
			}
			return targetJsc;
		}
		
		public static int GetPluginVer(Atom atom, string pluginName){
			if(atom == null){ return 0; }
			
			int ver = 0;
			Dictionary<string, string> dicUrl = GetPluginsUrlDictionary(atom);
			
			List<MVRScript> scripts = GetMVRScripts(atom);
			foreach(var script in scripts){
				string id = script.GetJSON()["id"];
				
				if(id.Contains(pluginName)){
					Match m1 = Regex.Match(id, "^(.+?)_(.+)$");
					string key = m1.Groups[1].Value;
					string value = m1.Groups[2].Value;				
					string url = dicUrl[key];
			
					Match m2 = Regex.Match(url, "^.+?\\.(\\d+):.+$");
					if(m2 != null){
						ver = Convert.ToInt32(m2.Groups[1].Value);
					}
				}
			}
			return ver;
		}		

		public static Color ConvertHSVtoRGB(HSVColor hsvColor)
		{
			return Color.HSVToRGB(hsvColor.H, hsvColor.S, hsvColor.V);
		}

		public static HSVColor ConvertRGBtoHSV(Color color)
		{
			HSVColor hsvColor = new HSVColor{ H = 0f, S = 0f, V = 0f };
			Color.RGBToHSV(color, out hsvColor.H, out hsvColor.S, out hsvColor.V);
			return hsvColor;
		}

		public static HSVColor ConvertJSNtoHSV(JSONNode jsn)
		{
			HSVColor hsvColor = new HSVColor { H = 0f, S = 0f, V = 0f };
			hsvColor.H = jsn["h"].AsFloat;
			hsvColor.S = jsn["s"].AsFloat;
			hsvColor.V = jsn["v"].AsFloat;
			return hsvColor;
		}

		public static HSVColor ConvertJSNtoHSV(string jsonString)
		{
			return ConvertJSNtoHSV(JSONNode.Parse(jsonString));
		}

		public static void DumpTags(Atom atom)
		{
			foreach(var dci in GetClothes(atom))
			{
				if(!dci.active) continue;
				string dcitags = dci.tags.ToLower();
				var tags = dci.tagsArray;
				bool hasUserPrefs = dci.hasUserPrefs;
				Debug($"cloth:{dci.displayName}, tags:{dcitags}, hasUserPrefs:{hasUserPrefs}");
				foreach(var tag in tags)
				{
					Debug($"\tcloth:{dci.displayName}, tag:{tag}");
				}
				var jss = dci.GetComponent<string>();
				Debug($"jss is null?:" + (jss == null));
				
			}
		}

		public static bool CheckTags(DAZClothingItem dci, string tags)
		{
			string[] words = tags.Replace(" ", "").ToLower().Split(',');
			foreach(var word in words)
			{
				if(dci.CheckMatchTag(word)) return true;
			}
			return false;
		}

		public static bool CheckTagsOrigin(DAZClothingItem dci, string tags)
		{
			string[] words = tags.Replace(" ", "").ToLower().Split(',');
			string dcitags = dci.tags.ToLower();
			foreach(var word in words)
			{
				if(dcitags.Contains(word)) return true;
			}
			return false;
		}

		public static void SetTagOrigin(DAZClothingItem dci, string tag)
		{
			tag = tag.ToLower();
			if(!CheckTagsOrigin(dci, tag)){
				dci.tags = dci.tags + $",{tag}";
			}
		}

		public static void UnsetTagOrigin(DAZClothingItem dci, string tag)
		{
			tag = tag.ToLower();
			if(CheckTagsOrigin(dci, tag))
			{
				string tags = dci.tags;
				if(tags == tag){
					tags = "";
				}else{
					tags = tags.Replace($",{tag}", "");
				}
				dci.tags = tags;
			}else{
				// Debug($"Not Match: tag={tag}, dci.tags={dci.tags}");
			}
		}

		public static string CorrectVarname(string filename)
		{
			if(filename == "") return "";
			return CutoffPrefix(CutoffExtension(filename));
		}

		public static string CutoffVersion(string pattern)
		{
			return Regex.Replace(pattern, "\\.\\d+\\.var$", "");
		}

		public static string CutoffExtension(string pattern)
		{
			return Regex.Replace(pattern, "\\.var$", "");
		}

		public static string CutoffPrefix(string filename)
		{
			return Regex.Replace(filename, "^.+\\\\([^\\\\]+)$", "$1");
		}

	}
}
