using System;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;

namespace Cotyounoyume
{
    public class WorldUIBase
    {
        public EBATCore core;
        public WorldUI worldUI;
        MVRScript plugin;
		public string UINameBase = "EBAT_UI_";
		public string UINameBaseTemporary = "EBAT_UITemporary_";
		public Atom UIAtomTemporary;
        public Canvas canvas;
        public Canvas canvasTemporary;
        private float UIScale;

        public float defaultCanvasOffsetUp = 0.5f;
		public float defaultCanvasOffsetRight = 0.8f;
		public float defaultCanvasOffsetForward = 0.4f;

		public int fontSize = 26;
		public int fontSizeAlt = 20;
		
		private float xSpacing = 0.3f;
		private float xIndent = 0.15f;
		private float ySpacing = -0.05f;
		private float xOffsetMinimize = 0.08f;
		private float xOffset = 0.04f;
		private float xOffset2= -0.12f;
		private float yOffset = -0.08f;
		private float smallRatio = 0.75f;
		private float smallRatioForSize = 0.80f;
		public string extOpen  = "[▶]";
		public string extClose = "[◀]";


        public WorldUIBase(EBATCore core, WorldUI worldUI, float UIScale)
        {
            this.core = core;
            this.plugin = core as MVRScript;
            this.worldUI = worldUI;
            this.UIScale = UIScale;
        }

		public void LoadSettings(){
			try{
				if(!core.settingsRoot["Common"].AsObject.HasKey("Default Position")){
					core.settingsRoot["Common"]["Default Position"]["Up"]      = defaultCanvasOffsetUp.ToString();
					core.settingsRoot["Common"]["Default Position"]["Right"]   = defaultCanvasOffsetRight.ToString();
					core.settingsRoot["Common"]["Default Position"]["Forward"] = defaultCanvasOffsetForward.ToString();
				}else{
					defaultCanvasOffsetUp      = core.settingsRoot["Common"]["Default Position"]["Up"].AsFloat;
					defaultCanvasOffsetRight   = core.settingsRoot["Common"]["Default Position"]["Right"].AsFloat;
					defaultCanvasOffsetForward = core.settingsRoot["Common"]["Default Position"]["Forward"].AsFloat;		
				}
				
			}catch(Exception e){
				core.LoggerE(e);
			}
		}
		
		public void ResetCanvasPosition(){
			defaultCanvasOffsetUp      = core.settingsRoot["Common"]["Default Position"]["Up"].AsFloat;
			defaultCanvasOffsetRight   = core.settingsRoot["Common"]["Default Position"]["Right"].AsFloat;
			defaultCanvasOffsetForward = core.settingsRoot["Common"]["Default Position"]["Forward"].AsFloat;
			
			if(core.UIAtom == null){ return; }
			core.UIAtom.mainController.transform.position = Utility.GetPositionFromTargetBodyname(core.atom, "headControl");			
			core.UIAtom.mainController.transform.Translate(Vector3.up * defaultCanvasOffsetUp);
			core.UIAtom.mainController.transform.Translate(Vector3.right * defaultCanvasOffsetRight);
			core.UIAtom.mainController.transform.Translate(Vector3.forward * defaultCanvasOffsetForward);
		}

 		public void DefaultCanvasPosition(Transform transform){
			Vector3 headPosition = Utility.GetPositionFromTargetBodyname(core.atom, "headControl");
			float upOffset      = defaultCanvasOffsetUp;
			float rightOffset   = defaultCanvasOffsetRight;
			float forwardOffset = defaultCanvasOffsetForward;
			Utility.LookFront(transform);
			
			transform.position = Utility.GetPositionFromTargetBodyname(core.atom, "headControl");
			transform.Translate(Vector3.up * upOffset);
			transform.Translate(Vector3.right * rightOffset);
			transform.Translate(Vector3.forward * forwardOffset);
		}
		
		public void SetCanvasPosition(string type, Transform transform, float upOffset, float rightOffset, float forwardOffset){
			bool bDefault = false;
			if(type == "temporary" || (type != "temporary" && !core.HasMemoryFloat("UI Position.x"))){
				bDefault = true;
			}else if(core.GetMemoryFloat("UI Position.x") == 0f && core.GetMemoryFloat("UI Position.y") == 0f && core.GetMemoryFloat("UI Position.z") == 0f){
				bDefault = true;
			}
			
			if(bDefault){
				transform.position = Utility.GetPositionFromTargetBodyname(core.atom, "headControl");
				transform.Translate(Vector3.up * upOffset);
				transform.Translate(Vector3.right * rightOffset);
				transform.Translate(Vector3.forward * forwardOffset);
			}else{
				transform.position = new Vector3(
					core.GetMemoryFloat("UI Position.x"),
					core.GetMemoryFloat("UI Position.y"),
					core.GetMemoryFloat("UI Position.z")
				);
				
				transform.rotation = new Quaternion(
					core.GetMemoryFloat("UI Rotation.x"),
					core.GetMemoryFloat("UI Rotation.y"),
					core.GetMemoryFloat("UI Rotation.z"),
					core.GetMemoryFloat("UI Rotation.w")
				);
			}
		}
		public void MakeMenuBase(string type = ""){
			
			worldUI.MakeStructureExpandButton();
			worldUI.MakeStructureCategoryButton();
			// worldUI.MakeStructureNormalButton();

			Vector3 headPosition = Utility.GetPositionFromTargetBodyname(core.atom, "headControl");
			float upOffset      = (type == "temporary") ?  0.2f : defaultCanvasOffsetUp;
			float rightOffset   = (type == "temporary") ? -0.2f : defaultCanvasOffsetRight;
			float forwardOffset = (type == "temporary") ?  0.0f : defaultCanvasOffsetForward;
			string UIName = (type == "temporary") ? UINameBaseTemporary : UINameBase;
			if(core.UIAtom == null){
				MakeCanvas(type, upOffset, rightOffset, forwardOffset);
				
				core.StartCoroutine(core.CreateAtom("Empty", UIName, (tmpatom)=>
				{
					Utility.LookFront(tmpatom.mainController.transform);
					SetCanvasPosition(type, tmpatom.mainController.transform, upOffset, rightOffset, forwardOffset);
					
					/*
					tmpatom.mainController.transform.position = Utility.getPositionFromTargetBodyname(core.atom, "headControl");
					tmpatom.mainController.transform.Translate(Vector3.up * upOffset);
					tmpatom.mainController.transform.Translate(Vector3.right * rightOffset);
					tmpatom.mainController.transform.Translate(Vector3.forward * forwardOffset);
					*/
					Canvas tmpCanvas = (type == "temporary") ? canvasTemporary : canvas;
					tmpCanvas.transform.SetParent(tmpatom.mainController.transform, true);
					if(type == "temporary"){
						UIAtomTemporary = tmpatom;
					}else{
						core.UIAtom = tmpatom;
//						SetCanvasPosition();
					}
					if(type != "temporary"){
						core.customUnityAssetItem.MakeSphereForUI(core.UIAtom.mainController.transform);
					}
				}));
			}else{
				if(type != "temporary"){
					core.customUnityAssetItem.MakeSphereForUI(core.UIAtom.mainController.transform);
				}
			}
		}
		
		public void MakeCanvas(string type = "", float upOffset = 0f, float rightOffset = 0f, float forwardOffset = 0f){
            GameObject canvasObject = new GameObject();
			Canvas tmpCanvas = null;
            tmpCanvas = canvasObject.AddComponent<Canvas>();
            tmpCanvas.renderMode = RenderMode.WorldSpace;
            tmpCanvas.pixelPerfect = false;
            SuperController.singleton.AddCanvas(tmpCanvas);

            CanvasScaler cs = canvasObject.AddComponent<CanvasScaler>();
            cs.scaleFactor = 100.0f;
            cs.dynamicPixelsPerUnit = 1f;

            canvasObject.AddComponent<GraphicRaycaster>();

            tmpCanvas.transform.localScale = new Vector3(this.UIScale, this.UIScale, this.UIScale);
			Utility.LookFront(tmpCanvas.transform);
			SetCanvasPosition(type, tmpCanvas.transform, upOffset, rightOffset, forwardOffset);
			/*
			Vector3 headPosition = Utility.getPositionFromTargetBodyname(core.atom, "headControl");
			tmpCanvas.transform.position = Utility.getPositionFromTargetBodyname(core.atom, "headControl");
			tmpCanvas.transform.Translate(Vector3.up * upOffset);
			tmpCanvas.transform.Translate(Vector3.right * rightOffset);
			tmpCanvas.transform.Translate(Vector3.forward * forwardOffset);
			*/
			if(type == "temporary"){
				canvasTemporary = tmpCanvas;
			}else{
				canvas = tmpCanvas;
			}
		}

		public void RemoveMenuBase(){
			if (core.UIAtom != null){
				SuperController.singleton.RemoveAtom(core.UIAtom);
			}
			if (UIAtomTemporary != null){
				SuperController.singleton.RemoveAtom(UIAtomTemporary);
			}
			if (SuperController.singleton != null){
				SuperController.singleton.RemoveCanvas(canvas);
			}			
			if (SuperController.singleton != null){
				SuperController.singleton.RemoveCanvas(canvasTemporary);
			}			
		}

        public void RemoveButton(UIDynamicButton button)
        {
            plugin.RemoveButton(button);
        }

		public void RemoveMenuBodyImpl(UIDynamicButton[] buttons, UIDynamicSlider[] sliders, UIDynamicTextField[] texts = null){
			foreach(UIDynamicButton button in buttons){
				// Utility.Debug($"button:" + button.label);
				if(button != null){ plugin.RemoveButton(button);} 
			}
			
			if(sliders != null)
			{
				foreach(UIDynamicSlider slider in sliders){
					if(slider != null){ Utility.RemoveSliderForCanvas(slider);}
				}			
			}
			
			if(texts != null)
			{
				foreach(UIDynamicTextField text in texts){
					if(text != null){ Utility.RemoveTextFieldForCanvas(text);}
				}
			}
		}
		
		public void RemoveMenuBodyImpl(UIDynamicButton[] buttons, UIDynamicTextField[] texts){
			foreach(UIDynamicButton button in buttons){
				if(button != null){ plugin.RemoveButton(button);} 
			}
			
			foreach(UIDynamicTextField text in texts){
				if(text != null){ Utility.RemoveTextFieldForCanvas(text);}
			}
		}

		#region Define Alias
		public void MakeMenuBodyPartExpandCollapseButtonForMain(bool expandOn, Dictionary<string, UIDynamicButton> uiDic){
			if(uiDic == null) return;
			var type = "Main";
			string dispName = type + (expandOn ? extOpen : extClose);
			var x = expandOn ? -2 : -1;
			var y = 0;
			var h = worldUI.buttonSizeY;
			if(expandOn){
				y = worldUI.GetLineNumber() + 1;
				// Utility.Debug($"bar:{core.GetMemoryBool("Bar")}, slider1:{core.GetMemoryBool("Slider1")}");
				var barOnOff = IsBarOrSliderOn() ? 30 : 40;
				h = barOnOff + y * 50;
			}

			var ui = MakeButton(dispName, x, worldUI.line + y, height:h);
			ui.button.onClick.AddListener(() => OnExpandCollapse(type));
			if(uiDic.ContainsKey(type)) uiDic[type] = ui;
			else uiDic.Add(type, ui);
			worldUI.line = !expandOn ? worldUI.line+1 : worldUI.line;
		}

		public bool IsBarOrSliderOn(){
			if((worldUI.IsCategoryOn() && core.GetMemoryBool("Bar")) || (!worldUI.IsCategoryOn() && core.GetMemoryBool("Slider1"))) return true;
			return false;
		}

		public void MakeMenuBodyPartExpandCollapseButton(string type, bool expandOn, Dictionary<string, UIDynamicButton> uiDic){
			if(type == "Main"){
				MakeMenuBodyPartExpandCollapseButtonForMain(expandOn, uiDic);
				return;
			}
			Dictionary<string, int> dicY = new Dictionary<string, int>(){ 
				{"Addon1", 9}, {"Addon2", 12}, {"Test", 3}, 
			};
			Dictionary<string, float> dicH = new Dictionary<string, float>(){ 
				{"Addon1", worldUI.buttonSizeY10}, {"Addon2", worldUI.buttonSizeY13}, {"Test", worldUI.buttonSizeY4}, 
			};

			string dispName = type + (expandOn ? extOpen : extClose);
			var x = expandOn ? -1 : -2;
			var y = expandOn ?  0 : dicY[type];
			var h = expandOn ? worldUI.buttonSizeY : dicH[type];
			if(type == "Addon2" && !expandOn){
				if(core.GetMemoryBool("isCompact")){
					y -= 1;
					h -= 50f;
				}
				if(core.oneArmEachOnGlobal){
					y += 1;
					h += 50f;
				}
			}
			var ui = MakeButton(dispName, x, worldUI.line + y, height:h);
			ui.button.onClick.AddListener(() => OnExpandCollapse(type));
			if(uiDic.ContainsKey(type)) uiDic[type] = ui;
			else uiDic.Add(type, ui);
			worldUI.line = expandOn ? worldUI.line+1 : worldUI.line;
		}

		public void OnExpandCollapse(string type){
			var state = core.GetMemoryBool(type);
			core.SetMemoryBool(type, !state);
			worldUI.RemoveMenuBody();
			worldUI.MakeMenuBody();
		}

        public string GetCreatorName(string name)
        {
            string[] words = name.Split('.');
            return (words.Length > 0) ? words[0] : "";
        }

        public void OpenHubDetail(string name, string manualId = "")
        {
            try{
                var hubBrowser = SuperController.singleton.hubBrowser;
                // if(!dlHelperCore.ResourceIDMap.ContainsKey(name)) return;
                
                string hub_name = Utility.CutoffExtension(name);
                // string hub_id = hubBrowser.GetPackageHubResourceId(hub_name);
                string hub_id = manualId != null ? manualId : hubBrowser.GetPackageHubResourceId(hub_name);
                var idResolved = (hub_id == null) ? false : true;
                // Utility.Debug($"hub_id:{hub_id}");
                // Utility.Debug("hub_id is null?: " + (hub_id == null));

                var targetCreator = GetCreatorName(name);
                if(idResolved)
                {
                    hubBrowser.OpenDetail(hub_id);
                }
                else
                {
                    // Utility.Debug($"creator:{targetCreator}");
                    hubBrowser.CreatorFilter = targetCreator;
                    hubBrowser.Show();
                }

                // Utility.Debug($"name:{name}, hub_id:{hub_id}");
            }catch(Exception e){
                Utility.LoggerE(e);
            }
        }


		public float GetWidth(string type, float width, bool isHeader = false, bool isHeader2 = false){
			float size = worldUI.buttonSizeX;
			if(type == "button"){
				size    = worldUI.buttonSizeX;
			}else if(type == "buttonAlt"){
				size    = worldUI.buttonSizeMinimizeX;
			}else if(type == "spacer"){
				size    = worldUI.spacerSizeX;
				return size * (core.GetMemoryBool("isSmall") ? smallRatio * smallRatioForSize : 1.0f);				
			}else if(type == "toggle"){
				size    = worldUI.toggleSizeX;
			}else if(type == "slider"){
				size    = worldUI.sliderSizeX;
			}

			if(isHeader2 && type != "buttonAlt"){
				size = worldUI.headerX;
			}else if(isHeader && type != "buttonAlt"){
				size = worldUI.headerX;
			}
			
			if(width != 0f){
				if(width == 55f) width *= 0.80f;
				return width * (core.GetMemoryBool("isSmall") ? smallRatio * smallRatioForSize : 1.0f);
			}else{
				return size * (core.GetMemoryBool("isSmall") ? smallRatio * smallRatioForSize : 1.0f);
			}
		}
		
		public float GetHeight(string type, float height){
			float size = worldUI.buttonSizeY;
			if(type == "button"){
				size    = worldUI.buttonSizeY;
			}else if(type == "buttonAlt"){
				size    = worldUI.buttonSizeY;
			}else if(type == "spacer"){
				size    = worldUI.spacerSizeY;
			}else if(type == "toggle"){
				size    = worldUI.toggleSizeY;
			}else if(type == "slider"){
				size    = worldUI.sliderSizeY;
			}
			if(height != 0f){
				return height * (core.GetMemoryBool("isSmall") ? smallRatio : 1.0f);
			}else{
				return size * (core.GetMemoryBool("isSmall") ? smallRatio : 1.0f);
			}
		}
		
		public float GetXPos(int x, bool isMinimizeButton = false, bool isMinisizeButton = false){
			return GetXPos((float)x, isMinimizeButton, isMinisizeButton);
		}
		
		public float GetXPos(float x, bool isMinimizeButton = false, bool isMinisizeButton = false){
			if(isMinimizeButton){
				return xOffsetMinimize * (core.GetMemoryBool("isSmall") ? smallRatio : 1.0f);
			}else if(x == -2){
				var xpos = xOffset2 * (core.GetMemoryBool("isSmall") ? smallRatio : 1.0f);
				return xpos;
				// return xpos - (core.GetMemoryBool("isSmall") ? 0.02f : 0f);
			}else if(x == -1){
				var xpos = xOffset * (core.GetMemoryBool("isSmall") ? smallRatio : 1.0f);
				return xpos + (core.GetMemoryBool("isSmall") ? 0.02f : 0f);
			}else if(isMinisizeButton){
				return xIndent + (xOffset + xSpacing * 0.4f * x) * (core.GetMemoryBool("isSmall") ? smallRatio * 1.2f : 1.0f);
			}else{
				return xIndent + (xOffset + xSpacing * x) * (core.GetMemoryBool("isSmall") ? smallRatio : 1.0f);
			}			
		}
		
		public float GetYPos(int y){
			return (yOffset + ySpacing * y) * (core.GetMemoryBool("isSmall") ? smallRatio : 1.0f);
		}
		
		public int GetFontSize(){
			return core.GetMemoryBool("isSmall") ? fontSizeAlt : fontSize;
		}
		
		public UIDynamic MakeSpacer(int x, int y, float width = 0f, float height = 0f){
			width = GetWidth("spacer", width, x == -1, x == -2);
			height = GetHeight("spacer", height);
			float xpos = GetXPos(x);
			float ypos = GetYPos(y);
			
			UIDynamic uiSpacer = Utility.CreateSpacer(plugin, canvas, width, height);
			Utility.TranslateUI(uiSpacer, xpos, ypos);
			return uiSpacer;
		}
		
		public float xposForDL = 0.7f;
		public float yposForDL = -1.0f;

		public UIDynamicButton MakeButtonTemporary(float width = 100f, float height = 60f, float xpos = 0.25f, float ypos = -0.05f, int fontSize = 50, string label = "OK", bool forDL = false){
			UIDynamicButton uiButton = Utility.CreateButton(plugin, canvasTemporary, label, width, height, fontSize);
			if(forDL) {
				Utility.TranslateUI(uiButton, xposForDL, yposForDL -0.05f);
			}else{
				Utility.TranslateUI(uiButton, xpos, ypos);
			}
			return uiButton;
		}
		
		public UIDynamicButton MakeButton(string name, float x, int y, float width = 0f, float height = 0f, bool isMinimizeButton = false, bool isMinisizeButton = false){
			int fontSizeVar = GetFontSize();
			string type = isMinimizeButton ? "buttonAlt" : "button";
			width = GetWidth(type, width, x == -1, x == -2);
			height = GetHeight(type, height);				
			float xpos = GetXPos(x, isMinimizeButton, isMinisizeButton);
			float ypos = GetYPos(y);
			
			UIDynamicButton uiButton = Utility.CreateButton(plugin, canvas, name, width, height, fontSizeVar);
			Utility.TranslateUI(uiButton, xpos, ypos);
			return uiButton;
		}
		
		public UIDynamicTextField MakeTextField(string text, float x, int y, float width = 0f, float height = 0f, bool isMinimizeButton = false){
			int fontSizeVar = GetFontSize();
			string type = isMinimizeButton ? "buttonAlt" : "button";
			width = GetWidth(type, width, x == -1, x == -2);
			height = GetHeight(type, height);				
			float xpos = GetXPos(x, isMinimizeButton);
			float ypos = GetYPos(y);
			
			UIDynamicTextField uiText = Utility.CreateTextField(plugin, canvas, text, width, height, fontSizeVar);
			Utility.TranslateUI(uiText, xpos, ypos);
			return uiText;
		}
		
		public UIDynamicTextField MakeTextFieldTemporary(string text, float width = 0f, float height = 0f, int fontSize = 50, bool forDL = false){
			UIDynamicTextField uiText = Utility.CreateTextField(plugin, canvasTemporary, text, width, height, fontSize);
			if(forDL){
				Utility.TranslateUI(uiText, xposForDL, yposForDL);
			}
			return uiText;
		}
		
		public UIDynamicButton MakeButton(string name, int x, int y, float width = 0f, float height = 0f, bool isMinimizeButton = false, bool isMinisizeButton = false){
			return MakeButton(name, (float)x, y, width, height, isMinimizeButton, isMinisizeButton);
		}
		
		public UIDynamicButton MakeButtonForAddon(string name, int x, int y, float width = 0f, float height = 0f){
			return MakeButton(name, x, y, width, height);
		}		
		
		public UIDynamicTextField MakeTextFieldForAddon(string name, int x, int y, float width = 0f, float height = 0f){
			return MakeTextField(name, x, y, width, height);
		}		
		
		public UIDynamicToggle MakeToggle(string name, int x, int y, float width = 0f, float height = 0f){
			try{
			int fontSizeVar = GetFontSize();
			width = GetWidth("toggle", width, x == -1, x == -2);
			height = GetHeight("toggle", height);
			float xpos = GetXPos(x);
			float ypos = GetYPos(y);
			
			UIDynamicToggle uiToggle = Utility.CreateToggle(plugin, canvas, name, width, height, fontSizeVar);
			Utility.TranslateUI(uiToggle, xpos, ypos);
			return uiToggle;
				
			}catch(Exception e){
				core.LoggerE(e);
				return null;
			}
		}
		
		public UIDynamicSlider MakeSlider(string name, int x, int y, float min, float max, float current){
			int fontSizeVar = GetFontSize();
			float width = GetWidth("slider", 0, x == -1, x == -2);
			float height = GetHeight("slider", 0);
			float xpos = GetXPos(x);
			float ypos = GetYPos(y);
			
			UIDynamicSlider uiSlider = Utility.CreateSliderAlt(plugin, canvas, name, width, height, min, max, current, true, fontSizeVar, core.customUi.jsUISmall.val ? smallRatio : 1.0f);
			Utility.TranslateUI(uiSlider, xpos, ypos);
			return uiSlider;
		}
		
		#endregion

   }
}