using System;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using Object = UnityEngine.Object;
using System.Text.RegularExpressions;

namespace Cotyounoyume
{
    public class AddonCUAUndresser
    {
        Dictionary<Atom, string> AssetNameMap = new Dictionary<Atom, string>();
        public EBATCore core;
        public bool excludeHeadOnGlobal = true;

        public AddonCUAUndresser(EBATCore core){
            this.core = core;
        }

        public void ReadyForUndressing(){
            var cuas = Utility.GetLinkedCUAs();
            foreach(var cua in cuas){
                var assetName = GetAssetName(cua);
                if(assetName == "") continue;

                if(AssetNameMap.ContainsKey(cua)) AssetNameMap[cua] = assetName;
                else AssetNameMap.Add(cua, assetName);
            }
        }

        public void LoadSettings(){
	        excludeHeadOnGlobal = core.settings["General"]["Undress CUAs Exclude Head"].AsBool;
        }

        public void Undress(){
            ReadyForUndressing();

            var cuas = Utility.GetLinkedCUAs();
            foreach(var cua in cuas){
                var rbName = Utility.GetLinkToRB(cua);
                if(excludeHeadOnGlobal && rbName == "head") continue;
                SetAssetName(cua, "");
            }
        }

        public void RecoverUndressing(){
            var cuas = Utility.GetLinkedCUAs();
            foreach(var cua in cuas){
                if(!AssetNameMap.ContainsKey(cua)) continue;
                SetAssetName(cua, AssetNameMap[cua]);
            }
        }

        public void SetAssetName(Atom cua, string value){
           JSONStorable js = cua.GetStorableByID("asset");
			if(js == null) return;

			JSONStorableStringChooser jsAssetName = js.GetStringChooserJSONParam("assetName");
			if(jsAssetName == null) return;

            js.SetStringChooserParamValue("assetName", value);
        }

        public string GetAssetName(Atom cua){
            JSONStorable js = cua.GetStorableByID("asset");
			if(js == null) return "";

			JSONStorableStringChooser jsAssetName = js.GetStringChooserJSONParam("assetName");
			if(jsAssetName == null) return "";

            // Utility.Debug($"atom:{cua}, js:{js}, jsAssetName:{jsAssetName}, val:{jsAssetName.val}");
			return jsAssetName.val;
        }

        public void OnDestroy(){
            RecoverUndressing();
            AssetNameMap.Clear();
        }
    }
}