using System;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using Object = UnityEngine.Object;
using System.Text.RegularExpressions;

namespace Cotyounoyume
{
    public class AddonClothSimItem : MonoBehaviour 
    {
		public Atom atom;
		public EBATCore core;
		public AddonManager1 addon1;
		
		public ClothesStateChanger csc;
		public bool isUndressActive = false;
		
		public string[] ngwords = {
			"eye", "nose", "face", "mouth", "glasses",
			"hair", "brow",
			"drip", "makeup",
			"mtk2", 
		};

		public AddonClothSimItem(EBATCore core){
			atom = core.atom;
			this.core = core;
		}
		
		public void LoadSettings(){
			try{
				this.addon1 = core.addonManager1;
			}catch (Exception e){
				core.LoggerE(e);
			}
		}
				
		public void ApplySettings(){
			BackupClothesState();
		}

		public void Manage(float time){
		}	
		
		public void OnDestroy(){
			csc.OnDestroy();
		}
		
		public void OnOrgasmStart()
		{
		}
		
		public void OnOrgasmStop()
		{

		}

		public void CheckClothAttribute(){
			csc.CheckClothAttribute();
		}
		
		public void BackupClothesState(){
			try{
				if(csc != null){
					csc.ChangeClothesIfNeeded();
					SetRemoveOptions();
					csc.ReloadStatus();
				}else{
					csc = new ClothesStateChanger(core.atom, core, ngwords);
					SetRemoveOptions();
					csc.ReloadStatus();
				}
			}catch(Exception e){
				core.LoggerE(e);
			}			
		}

		public void OnLoadDone(){
			SetRemoveOptions();
			SetUndressOptions();
		}

		public void SetRemoveOptions(){
			if(!core.IsLoaded()) return; 
			
			csc.SetRemoveOptions(
				core.addonManager1.holdNGWordsOnGlobal,
				core.addonManager1.holdClothSimOnGlobal, 
				core.addonManager1.holdSkinWrapOnGlobal, 
				core.addonManager1.wrapAlterToSimOnGlobal);
		}

		public void SetUndressOptions(){
			if(!core.IsLoaded()) return; 

			csc.SetUndressOptions(
				core.addonManager1.forcedUndressOnGlobal,
				core.addonManager1.undressReverseOnGlobal
				);
		}
		
		public void ChangeClothes(){
			csc.ChangeClothesIfNeeded();
		}
		
		public void AlterSkinWrap(string tag = ""){
			try{
				csc.WrapAlter(false, tag);			
			}catch(Exception e){
				core.LoggerE(e);
			}						
		}
		
		public void WrapNormal(string tag = ""){
			try{
				csc.WrapNormal(tag);			
			}catch(Exception e){
				core.LoggerE(e);
			}						
		}
		
		public void RemoveClothes(string tag = ""){
			try{
				csc.RemoveClothes(tag);
			}catch(Exception e){
				core.LoggerE(e);
			}						
		}
		
		public void RemoveResetClothes(){
			try{
				csc.RemoveResetClothes();
			}catch(Exception e){
				core.LoggerE(e);
			}						
		}
		
		public void RecoverClothesState(){
			try{
				isUndressActive = false;
				csc.Recover();
			}catch(Exception e){
				core.LoggerE(e);
			}			
		}		
		
		public void ResetClothesState(){
			try{
				csc.Reset();
			}catch(Exception e){
				core.LoggerE(e);
			}			
		}		


		public void DoInitializeIfNeeded(bool BitsAndPiecesOn = false, string tag = "", bool NotForcedUndressOverride = false, bool EnableDetach = true)
		{
			csc.DoInitializeIfNeeded(BitsAndPiecesOn, tag, NotForcedUndressOverride, EnableDetach);
		}
		
		public void Undress(
			float Power = 0f,
			float Duration = 0f,
			bool BitsAndPiecesOn = false,
			bool NotForcedUndressOverride = false,
			bool EnableDetach = true,
			string Tags = ""
		)
		{
			csc.Undress(Power, Duration, BitsAndPiecesOn, NotForcedUndressOverride, EnableDetach, Tags);
		}

	}
}