#region using
using System;
using UnityEngine;
using UnityEngine.UI;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
using MeshVR;
using GPUTools.Cloth.Scripts;
using GPUTools.Cloth.Scripts.Runtime.Data;
using GPUTools.Cloth.Scripts.Geometry.Data;
#endregion

namespace Cotyounoyume{
	public class ClothesStateChanger
	{
		public List<ClothManager> clothManagers = new List<ClothManager>();
		public List<ClothManager> clothManagersUndressed = new List<ClothManager>();
		public Atom atom;
		public EBATCore core;
		public string[] ngwords;
		public bool holdNGWords   = true;
		public bool holdClothSim  = true;
		public bool holdSkinWraps = true;
		public bool wrapAlterToSimOn = false;
		public bool doneOnce = false;

		public float Power = 2f;
		public float Duration = 2f;
		public float Step = 0.1f;
		public bool UndressForced = true;
		public bool IsReverse = false;

		
		public ClothesStateChanger(Atom atom, EBATCore core, string[] ngwords){
			this.core = core;
			
			clothManagers = new List<ClothManager>();
			clothManagersUndressed = new List<ClothManager>();
			this.ngwords = ngwords;
			this.atom = atom;
			ChangeClothes();
			doneOnce = true;
		}

		public void ReloadStatus(){
			if(IsClothManagersWrong()) return;
			
			foreach(var cm in clothManagers){
				cm.ReloadStatus();
			}
			foreach(var cm in clothManagersUndressed){
				cm.ReloadStatus();
			}
		}

		public bool IsClothManagersWrong(){
			if(clothManagers != null) return false;
			Utility.Debug("UnknownError: clothManagers is null. remake clothManagers.");
			return true;
		}

		public void ConfigurateParameter(bool UndressForced = true, bool IsReverse = false, float Power = 3f, float Duration = 2f, float Step = 0.1f)
		{
			this.UndressForced = UndressForced;
			this.IsReverse = IsReverse;
			this.Power = Power;
			this.Duration = Duration;
			this.Step = Step;
		}
		
		public Dictionary<DAZClothingItem, int> MakeHitMap(){
			try{
				if(IsClothManagersWrong()) return new Dictionary<DAZClothingItem, int>();
			
				JSONStorable geometryStorable = atom.GetStorableByID("geometry");
				DAZCharacterSelector geometryForClothes = geometryStorable as DAZCharacterSelector;			
				List<DAZClothingItem> newClothes = geometryForClothes.clothingItems.ToList();
				
				Dictionary<DAZClothingItem, int> hitMap = new Dictionary<DAZClothingItem, int>();
				foreach(var cm in clothManagers){
					if(!hitMap.ContainsKey(cm.cloth)){
						hitMap.Add(cm.cloth, 1);
					}else{
						hitMap[cm.cloth] += 1;
					}
				}
				foreach(var dci in newClothes){
					if(!hitMap.ContainsKey(dci)){
						hitMap.Add(dci, -1);
					}else{
						hitMap[dci] -= 1;
					}
				}
				return hitMap;
			}catch(Exception e){
				Utility.LoggerE(e);
				return null;
			}
		}
		
		public void ChangeClothesIfNeeded(){
			if(IsClothManagersWrong()) return;

			ReloadStatus();
			bool isNeeded = false;
			Dictionary<DAZClothingItem, int> hitMap = MakeHitMap();
			foreach(KeyValuePair<DAZClothingItem, int> kvp in hitMap){
				if(kvp.Value != 0){
					isNeeded = true;
				}
			}
			if(isNeeded){
				ChangeClothes();
			}
		}
		
		public void ChangeClothes(){
			try{
				JSONStorable geometryStorable = atom.GetStorableByID("geometry");
				DAZCharacterSelector geometryForClothes = geometryStorable as DAZCharacterSelector;
				List<DAZClothingItem> clothes = geometryForClothes.clothingItems.ToList();
				
				Dictionary<DAZClothingItem, int> hitMap = MakeHitMap();				
				foreach(KeyValuePair<DAZClothingItem, int> kvp in hitMap){
					if(kvp.Value == 0){			//exists both array
						//nop
					}else if(kvp.Value == 1){	//exists only clothManagers
						RecoverSpecificCloth(kvp.Key);
						RemoveSpecificClothFromManager(kvp.Key);
					}else if(kvp.Value == -1){	//exists only new clothes
						if(kvp.Key.active && kvp.Key.creatorName != "cotyounoyume"){
							clothManagers.Add(new ClothManager(core, this, kvp.Key));
						}
					}
				}
			}catch(Exception e){
				SuperController.LogMessage("*** Exception:" + e);
			}			
		}
		
		public void SetRemoveOptions(bool bHoldNGWords, bool bHoldClothSim, bool bHoldSkinWraps, bool wrapAlterToSimOn){
			this.holdNGWords      = bHoldNGWords;
			this.holdClothSim     = bHoldClothSim;
			this.holdSkinWraps    = bHoldSkinWraps;
			this.wrapAlterToSimOn = wrapAlterToSimOn;
		}
				
		public void SetUndressOptions(bool forcedUndress, bool undressReverse){
			this.UndressForced = forcedUndress;
			this.IsReverse = undressReverse;
		}

		public bool IsNG(DAZClothingItem dci){
			bool result = false;
			foreach(var name in ngwords){
				if(dci.displayName.ToLower().Contains(name)){
					result = true;
				}
			}
			return result;
		}
		
		public void RemoveClothes(string tag = ""){
			if(IsClothManagersWrong()) return;
			
			try{
				
			int clothNum = clothManagers.Count;
			for(int i = clothNum - 1; i >= 0; i--){
				bool hold = false;
				bool hasRuntime = clothManagers[i].hasRuntime;
				bool hasCs      = clothManagers[i].hasCs;
				bool hasAltWrap = clothManagers[i].hasAltWrap;
				DAZClothingItem cloth = clothManagers[i].cloth;

				if(tag != "" && !clothManagers[i].cloth.CheckMatchTag(tag)) continue;

				if(holdNGWords){
					if(IsNG(cloth)){
						hold = true;
					}
				}
				if(holdClothSim){
					if(hasRuntime){
						hold = true;
					}
				}
				if(holdSkinWraps){
					if(hasAltWrap){
						hold = true;
					}
				}
				if(!hold){
					RemoveClothAndBackup(i);
				}
			}
			
			}catch(Exception e){
				Utility.LoggerE(e);
			}
		}

		public void RemoveClothAndBackup(int i){
			if(IsClothManagersWrong()) return;
			
			ClothManager cm = clothManagers[i];
			clothManagersUndressed.Add(cm);
			JSONStorable geometryStorable = atom.GetStorableByID("geometry");
			DAZCharacterSelector geometryForClothes = geometryStorable as DAZCharacterSelector;
			geometryForClothes.SetActiveClothingItem(cm.cloth, active: false);
			clothManagers.RemoveAt(i);
		}
		
		public void RestoreClothFromBackup(int i){
			ClothManager cm = clothManagersUndressed[i];
			clothManagers.Add(cm);
			JSONStorable geometryStorable = atom.GetStorableByID("geometry");
			DAZCharacterSelector geometryForClothes = geometryStorable as DAZCharacterSelector;
			geometryForClothes.SetActiveClothingItem(cm.cloth, active: true);
			clothManagersUndressed.RemoveAt(i);
		}
		
		public void RemoveResetClothes(){
			int clothNum = clothManagersUndressed.Count;
			for(int i = clothNum - 1; i >= 0; i--){
				RestoreClothFromBackup(i);
			}
		}
		
		public void CheckClothAttribute(){
			ChangeClothesIfNeeded();

			if(IsClothManagersWrong()) return;
			
			for(int i = 0; i < clothManagers.Count; i++){
				clothManagers[i].CheckClothAttribute();
			}
		}
		
		public void RecoverSpecificCloth(DAZClothingItem dci){
			if(IsClothManagersWrong()) return;
			
			foreach(var cm in clothManagers){
				if(dci == cm.cloth){
					cm.Recover();
					cm.WrapNormal();
				}
			}
		}
		
		public void RemoveSpecificClothFromManager(DAZClothingItem dci){
			if(IsClothManagersWrong()) return;
			
			int clothNum = clothManagers.Count;
			for(int i = clothNum - 1; i >= 0; i--){
				if(dci == clothManagers[i].cloth){
					clothManagers.RemoveAt(i);
				}
			}
		}
		
		public void Recover(){
			try{
				ClearOptionalTags();

				if(IsClothManagersWrong()) return;
			
				for(int i = 0; i < clothManagers.Count; i++){
					clothManagers[i].Recover();
				}
				for(int i = 0; i < clothManagersUndressed.Count; i++){
					clothManagersUndressed[i].Recover();
				}
			}catch(Exception e){
				SuperController.LogMessage("*** Exception:" + e);
			}			
		}
		
		public void Reset(){
			try{
				if(IsClothManagersWrong()) return;
			
				for(int i = 0; i < clothManagers.Count; i++){
					clothManagers[i].Reset();
				}
				for(int i = 0; i < clothManagersUndressed.Count; i++){
					clothManagersUndressed[i].Reset();
				}
			}catch(Exception e){
				SuperController.LogMessage("*** Exception:" + e);
			}			
		}
		
		public void WrapAlter(bool isForce = false, string tag = ""){
			if(IsClothManagersWrong()) return;
			
			for(int i = 0; i < clothManagers.Count; i++){
				if(tag != "" && !clothManagers[i].cloth.CheckMatchTag(tag)) continue;
				clothManagers[i].wm.WrapAlter();
			}
		}
		
		public void WrapNormal(string tag = ""){
			if(IsClothManagersWrong()) return;
			
			for(int i = 0; i < clothManagers.Count; i++){
				if(tag != "" && !clothManagers[i].cloth.CheckMatchTag(tag)) continue;

				clothManagers[i].wm.WrapNormal();
			}
		}
				
		public void DoInitializeIfNeeded(bool BitsAndPiecesOn = false, string tag = "", bool NotForcedUndressOverride = false, bool EnableDetach = true)
		{
			if(IsClothManagersWrong()) return;
			
			for(int i = 0; i < clothManagers.Count; i++){
				if(tag != "" && !clothManagers[i].cloth.CheckMatchTag(tag)) continue;

				if(!holdNGWords || !IsNG(clothManagers[i].cloth)){
					clothManagers[i].DoInitializeIfNeeded(BitsAndPiecesOn, NotForcedUndressOverride, EnableDetach);
				}
			}
		}

		public void Undress(
			float Power = 0f,
			float Duration = 0f,
			bool BitsAndPiecesOn = false,
			bool NotForcedUndressOverride = false,
			bool EnableDetach = true,
			string Tags = ""
		)
		{
			if(IsClothManagersWrong()) return;
			
			foreach(var cm in clothManagers)
			{
				if(Tags != "" && !Utility.CheckTags(cm.cloth, Tags)) continue;
				if(!holdNGWords || !IsNG(cm.cloth)){
					cm.UndressIndividual(Power, Duration, BitsAndPiecesOn, NotForcedUndressOverride, EnableDetach);
				}
			}
		}

		public void ClearOptionalTags()
		{
			if(IsClothManagersWrong()) return;
			
			foreach(var cm in clothManagers)
			{
				cm.UnsetTagOrigin("inUndressing");
				cm.UnsetTagOrigin("Initialized");
			}
		}

		public void DisplayTags()
		{
			if(IsClothManagersWrong()) return;
			
			foreach(var cm in clothManagers)
			{
				Utility.Debug($"cloth:{cm.cloth.name}, tags:{cm.cloth.tags}");
			}
		}


		public void OnDestroy(){
			if(doneOnce){
				for(int i = 0; i < clothManagers.Count; i++){
					clothManagers[i].OnDestroy();
					clothManagers[i] = null;
				}
				clothManagers = null;
				for(int i = 0; i < clothManagersUndressed.Count; i++){
					clothManagersUndressed[i].OnDestroy();
					clothManagersUndressed[i] = null;
				}
				clothManagers = null;
			}
		}
	}
	
	public class ClothManager
	{
		public EBATCore core;
		public ClothesStateChanger csch;
		public DAZClothingItem cloth;
		public WrapManager wm;
		public ClothSettingsManager csm;
		public ClothSimControl csc;
		public RuntimeData rtd;
		
		public bool simEnabledOrig = false;
		public bool isBuiltin = false;
		public bool hasRuntime = false;
		public bool hasCsc     = false;
		public bool hasCs      = false;
		public bool hasAltWrap = false;
		public bool doneOnce = false;
		public bool wrapAlterToSimOn = false;

		public float elapsedTime = 0f;
		public float additionalPower = 0f;
		
		public ClothManager(EBATCore core, ClothesStateChanger csch , DAZClothingItem cloth){
			this.core = core;
			this.csch = csch;
			this.cloth = cloth;
			
			wm  = new WrapManager(core, this);
			csm = new ClothSettingsManager(core, this);
			ReloadStatus();
			
			// isBuiltin = !cloth.uid.Contains("/");
			// hasCsc     = csm.hasCsc;
			// hasCs      = csm.hasCs;
			// hasRuntime = wm.hasRuntime;
			// hasAltWrap = wm.hasAltWrap;
			// doneOnce = true;

			// csc = Utility.GetClothSimControl(cloth);
			// rtd = Utility.GetRuntime(cloth);	
		}
		
		public void ReloadStatus(){
			csm.ReloadStatus();
			wm.ReloadStatus();
			
			isBuiltin = !cloth.uid.Contains("/");
			hasCsc     = csm.hasCsc;
			hasCs      = csm.hasCs;
			hasRuntime = wm.hasRuntime;
			hasAltWrap = wm.hasAltWrap;
			csc = Utility.GetClothSimControl(cloth);
			rtd = Utility.GetRuntime(cloth);	
		}
		
		public bool IsIrregular(){
			return csm.isIrregular;
		}

		public void SetUndressing(bool IsOn = true)
		{
			if(IsOn) SetTagOrigin("inUndressing");
			else UnsetTagOrigin("inUndressing");
		}

		public bool IsInUndressing()
		{
			return CheckTagsOrigin("inUndressing");
		}

		public void CheckClothAttribute(){
			try{
				Utility.Info("clothname:" + cloth.displayName);
				bool hasCsc = (csm.csc != null);
				bool hasCs  = (csm.cs  != null);
				bool hasGeometryData = false;
				ClothGeometryData geometryData = null;
				if(hasCs){
					geometryData = csm.cs.GeometryData;
					hasGeometryData = (geometryData != null);
				}
				if(hasCs != hasRuntime){
					Utility.Info(" ******** IRREGULAR DATA *********:" +  cloth.displayName + "(" + cloth.uid + ")");
					if(hasCs){
						csm.cs.ProcessGeometry();
					}
				}
				Utility.Info("  hasGeometryData:" +  hasGeometryData);
				Utility.Info("  uid:" +  cloth.uid);
				Utility.Info("  hasAltWrap:" +  hasAltWrap);
				Utility.Info("  hasRuntime:" +  hasRuntime);
				Utility.Info("  wm:" +  wm);
				Utility.Info("  csm:" +  csm);
				Utility.Info("  hasCsc:" +  hasCsc);
				Utility.Info("  hasCs:" +  hasCs);
			}catch(Exception e){
				Utility.LoggerE(e);
			}
			
		}
		
		public void WrapNormal(){
			wm.Recover();
		}
		public void Recover(){
			UnsetTagOrigin("Initialized");
			UnsetTagOrigin("inUndressing");
			csm.Recover();
		}

		public void Reset(){
			csm.Reset();
		}

		public void UndressIndividual(
			float Power = 0f,
			float Duration = 0f,
			bool BitsAndPiecesOn = false,
			bool NotForcedUndressOverride = false,
			bool EnableDetach = true
		)
		{
			if(IsInUndressing()) {
				RechargePowerAndTime(Power);
			}
			else
			{
				core.StartCoroutine(UndressIndividualCoroutine(Power, Duration, BitsAndPiecesOn, NotForcedUndressOverride, EnableDetach)); 
			}
		}

		public void RechargePowerAndTime(float Power = 0f)
		{
			elapsedTime = 0f;
			additionalPower += 0.5f * ((Power != 0f) ? Power : csch.Power);
		}

		public IEnumerator UndressIndividualCoroutine(float Power = 0f, float Duration = 0f, bool BitsAndPiecesOn = false, bool NotForcedUndressOverride = false, bool EnableDetach = true)
		{
			DoInitializeIfNeeded(BitsAndPiecesOn, NotForcedUndressOverride, EnableDetach);
			yield return new WaitForSeconds(0.001f);
			SetUndressing(true);

			float tmpDuration = (Duration != 0f) ? Duration : csch.Duration;
			for(elapsedTime = 0f; elapsedTime <= tmpDuration; elapsedTime+= csch.Step)
			{
				additionalPower *= 0.9f;
				float ratio = 1f - (elapsedTime / csch.Duration);
				float tmpPower = ((Power != 0f) ? Power : csch.Power ) + additionalPower;

				csm.PowerControl(ratio, tmpPower);
				yield return new WaitForSeconds(csch.Step);
			}
			SetUndressing(false);
		}

		public void DoInitializeIfNeeded(bool BitsAndPiecesOn = false, bool NotForcedUndressOverride = false, bool EnableDetach = true)
		{
			try{
				csm.DoInitializeIfNeeded(BitsAndPiecesOn, NotForcedUndressOverride, EnableDetach);
			}catch(Exception e){
				Utility.LoggerE(e);
			}

		}

		public void OnDestroy(){
			if(doneOnce){
				wm.OnDestroy();
				csm.OnDestroy();
			}
		}

		public bool IsInitialized()
		{
			return CheckTagsOrigin("Initialized");
		}

		public bool CheckTagsOrigin(string tags)
		{
			return Utility.CheckTagsOrigin(cloth, tags);
		}
		
		public void SetTagOrigin(string tag)
		{
			Utility.SetTagOrigin(cloth, tag);
		}
		
		public void UnsetTagOrigin(string tag)
		{
			Utility.UnsetTagOrigin(cloth, tag);
		}
	}
	
	public class ClothSettingsManager
	{
		public EBATCore core;
		public ClothManager cm;
		public ClothSimControl csc;
		public ClothSettings   cs;
		public RuntimeData     rtd;
		
		public bool hasRuntime = false;
		public bool hasCsc     = false;
		public bool hasCs      = false;
		public bool isIrregular = false;
		
		public bool    backupBreakEnabled;
		public float   backupBreakThreshold;
		public float   backupFriction;
		public float   backupWeight;
		public float   backupGravityMultiplier;
		public float   backupStiffness;
		public float   backupDistanceScale;
		public bool    backupSimEnabled;
		public float   backupCollisionPower;
		public float   backupParticleRadius;
		public float   backupStretchability;
		public float   backupJointStrength;
		
		public float   durationTime = 0f;
		public float   elapsedTime = 0f;
		public float   power = 0f;
		public float   progress = 0f;

		public float[] backupParticlesStrength;
		public float[] backupParticlesBlend;
		public bool    doneOnce = false;
		public bool    arrayCopyDone = false;
		public bool    csCopyDone = false;

		public ClothSettingsManager(EBATCore core, ClothManager cm){
			this.core = core;
			this.cm = cm;
			ReloadStatus();
			doneOnce = true;
			// ConfiguratePowerAndTime();
		}	

		public void ReloadStatus(){
			if(!csCopyDone || !arrayCopyDone){
				this.csc = cm.csc;
				this.rtd = cm.rtd;
				cs  = (csc != null) ? Utility.GetClothSettings(csc) : null ;
				hasRuntime = (rtd != null);
				hasCsc     = (csc != null);
				hasCs      = (cs  != null);
				
				if(hasCs && csCopyDone == false){
					backupBreakEnabled      = cs.BreakEnabled;
					backupBreakThreshold    = cs.BreakThreshold;
					backupFriction          = cs.Friction;
					backupWeight            = cs.Weight;
					backupGravityMultiplier = cs.GravityMultiplier;
					backupStiffness         = cs.Stiffness;
					backupDistanceScale     = cs.DistanceScale;
					backupCollisionPower    = cs.CollisionPower;
					backupParticleRadius    = cs.ParticleRadius;
					backupStretchability    = cs.Stretchability;
					backupJointStrength     = cs.JointStrength;
					
					backupSimEnabled        = csc.simEnabledJSON.val;
					csCopyDone = true;
				}
				
				if(hasRuntime && arrayCopyDone == false){
					backupParticlesStrength = new float[cs.GeometryData.ParticlesStrength.Length];
					backupParticlesBlend    = new float[cs.GeometryData.ParticlesBlend.Length];
					Array.Copy(cs.GeometryData.ParticlesStrength, backupParticlesStrength, cs.GeometryData.ParticlesStrength.Length);
					Array.Copy(cs.GeometryData.ParticlesBlend   , backupParticlesBlend   , cs.GeometryData.ParticlesBlend.Length   );
					arrayCopyDone = true;
				}
			}
		}
		
		public IEnumerator SimEnableCoroutine(){
			if(!csc.simEnabledJSON.val){
				csc.SetSimEnabled(true);
				yield return new WaitForSeconds(0.1f);
				csc.SetSimEnabled(false);
				csc.Reset();
			}
			yield return null;
		}
		
		public void Recover(){
			if(!hasRuntime){ return; }
			csc.clothSettings.BreakEnabled      = backupBreakEnabled;
			csc.clothSettings.BreakThreshold    = backupBreakThreshold;
			csc.clothSettings.Friction          = backupFriction;
			csc.clothSettings.Weight            = backupWeight;		
			csc.clothSettings.GravityMultiplier = backupGravityMultiplier;
			csc.clothSettings.Stiffness         = backupStiffness;
			csc.clothSettings.DistanceScale     = backupDistanceScale;
			csc.clothSettings.CollisionPower    = backupCollisionPower;
			csc.clothSettings.ParticleRadius    = backupParticleRadius;
			csc.clothSettings.Stretchability    = backupStretchability;
			csc.clothSettings.JointStrength     = backupJointStrength;

			csc.SetSimEnabled(backupSimEnabled);
			if(hasRuntime){
				Array.Copy(backupParticlesStrength, cs.GeometryData.ParticlesStrength, backupParticlesStrength.Length);
				Array.Copy(backupParticlesBlend   , cs.GeometryData.ParticlesBlend   , backupParticlesBlend.Length);
			}
		}
		
		public void Reset(){
			if(!hasRuntime){ return; }
			
			csc.Reset();
		}

		public void DoInitializeIfNeeded(bool BitsAndPiecesOn = false, bool NotForcedUndressOverride = false, bool EnableDetach = true)
		{
			if(cm.IsInitialized()) { return; }
			BreakControl(true, EnableDetach, BitsAndPiecesOn, false, 0.05f, NotForcedUndressOverride);
			cm.SetTagOrigin("Initialized");
		}

		public void BreakMolecularLevel(float radius = 0.05f)
		{
			if(cs.GeometryData != null){
				for(int i = 0; i < cs.GeometryData.ParticlesStrength.Length; i++){
					cs.GeometryData.ParticlesBlend[i] = radius;
				}
			}			
		}

		public void GetCSCIfNeede()
		{
			if(csc == null)
			{
				if(cm.csc == null) {
					cm.ReloadStatus();
				}
					
				csc = cm.csc;
				cs  = (csc != null) ? Utility.GetClothSettings(csc) : null ;
			}
		}


		public void BreakControl(bool ResetOn = false, bool DetachableOn = false, bool BitsAndPiecesOn = false, bool StiffnessOn = false, float Radius = 0.05f, bool NotForcedUndressOverride = false)
		{
			try{
				GetCSCIfNeede();
				if(!hasRuntime) return;
				csc.clothSettings.BreakEnabled   = true;
				csc.clothSettings.JointStrength = 0.05f;			

				if(ResetOn) csc.SetSimEnabled(true);
				if(StiffnessOn) csc.clothSettings.Stiffness = 1.0f;
				if(DetachableOn) csc.SyncAllowDetatch(true);
				csc.clothSettings.BreakThreshold = (BitsAndPiecesOn) ? 0.1f : 0f;
				if(cm.csch.UndressForced && !NotForcedUndressOverride)
				{
					cs.DistanceScale  = 1.01f;
					cs.JointStrength = 0.05f;
					cs.CollisionPower = 0.5f;
						BreakMolecularLevel(Radius);
				}
				if(ResetOn) Reset();
			}catch(Exception e){
				Utility.LoggerE(e);
			}
		}

		public void PowerControl(float ratio, float power)
		{
			try{
				if(!hasCs) return;
				float sign = (cm.csch.IsReverse) ? -1f : 1f;
				csc.clothSettings.GravityMultiplier = Mathf.Lerp(sign * 0.5f, sign * power, ratio);
				csc.clothSettings.Weight            = Mathf.Lerp(       0.5f,        power, ratio);
				csc.clothSettings.Friction          = Mathf.Lerp(       1.0f,        0.1f, ratio);
			}catch(Exception e){
				Utility.LoggerE(e);
			}
		}

		public void OnDestroy(){
			if(!hasCs){ return; }
			
			if(doneOnce){
				Recover();
				
				csc = null;
				cs  = null;
				rtd = null;
				backupParticlesStrength = null;
				backupParticlesBlend = null;
			}else{
				//nop
			}
		}
	}
}