using System;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;

namespace Cotyounoyume
{
	public struct ButtonStructW1{
		public string name;
		public int x;
		public bool compactOn;
		public float width;
		public Action action;
		public Func<bool> func;

		public ButtonStructW1(string name, int x, bool compactOn, Action action, float width = 0f, Func<bool> func = null){
			this.name = name;
			this.x = x;
			this.compactOn = compactOn;
			this.action = action;
			this.width = width;
			this.func = func;
		}
	}	

    public class WorldUIAddon1{
		public EBATCore core;
		public WorldUI worldUi;
		public AddonManager1 addonManager1;
		
		public AddonClothSimItem addonClothSim;
		public AddonTakeControlItem addonTakeControl;		
		public AddonCUAUndresser addonCUAUndresser;
		
		public Dictionary<string, UIDynamicButton> uiExpandButtons = new Dictionary<string, UIDynamicButton>();
		public Dictionary<string, UIDynamicButton> uiButtons = new Dictionary<string, UIDynamicButton>();
		public Dictionary<string, ButtonStructW1> ButtonStructW1Dic;
		public string extOpen  = "[▶]";
		public string extClose = "[◀]";
		
		public WorldUIAddon1(WorldUI worldUi){
			this.worldUi = worldUi;
			this.core = worldUi.core;
		}

		public void LoadSettings(){
			this.addonManager1 = core.addonManager1;
			this.addonClothSim         = addonManager1.addonClothSim;
			this.addonTakeControl      = core.addonTakeControl;
			this.addonCUAUndresser     = addonManager1.addonCUAUndresser;
		}
		
		public void ApplySettings(){
			
		}

		public void ReadMemory(){
			try{
				if(ButtonStructW1Dic == null) Utility.Debug("ButtonStructW1Dic is null");
				foreach(var name in ButtonStructW1Dic.Keys.ToList()){
					if(!name.StartsWith("C:")) continue;
					var bs = ButtonStructW1Dic[name];
					if(bs.func == null) continue;
					if(!core.HasMemoryBool(name)){
						core.SetMemoryBool(name, bs.func());
					}else{
						var current = bs.func();
						var mem = core.GetMemoryBool(name);
						if(current != mem){
							bs.action();
						}
					}
				}
			}catch(Exception e){
				core.LoggerE(e);
			}
		}

		public void RemoveMenuBodyAddon(){
			// var buttons = uiButtons.Values.ToList().Concat(uiActionButtons.Values).ToArray();
			worldUi.worldUIBase.RemoveMenuBodyImpl(uiButtons.Values.ToArray(), null, null);
			worldUi.worldUIBase.RemoveMenuBodyImpl(uiActionButtons.Values.ToArray(), null, null);
			worldUi.worldUIBase.RemoveMenuBodyImpl(uiExpandButtons.Values.ToArray(), null, null);
		}
		
		public void MakeButtonStructures()
		{
			ButtonStructW1Dic = new Dictionary<string, ButtonStructW1>(){
				{"Undress Reset",     new ButtonStructW1("Reset"            , -1, false, () => OnUndressRecoverClick(), 0, null)},

				{"Undress CUAs",      new ButtonStructW1("Undress CUAs"   ,    0, false, () => OnUndressCUAsClick(), 0, null)},
				{"C:Exclude Head",    new ButtonStructW1("Exclude Head"   ,    1, false, () => OnConfigXClickCommon("C:Exclude Head"), 0, () => core.addonManager1.addonCUAUndresser.excludeHeadOnGlobal)},
				{"H:Hint17",          new ButtonStructW1("Help for CUAs..." ,  2, false, () => OnHint17Click(), 0, null)},

				{"Undress Overall",   new ButtonStructW1("Undress Overall"  ,  0, false, () => OnUndressOverallClick(), 0, null)},
				{"C:Undress Detach",  new ButtonStructW1("Undress Detach"   ,  1, false, () => OnConfigXClickCommon("C:Undress Detach"), 0, () => core.addonManager1.undressDetachOnGlobal)},
				{"C:Undress Forced",  new ButtonStructW1("Undress Forced"   ,  2, false, () => OnConfigXClickCommon("C:Undress Forced"), 0, () => core.addonManager1.forcedUndressOnGlobal)},
				{"C:Reverse Gravity", new ButtonStructW1("Reverse Gravity"  ,  3, false, () => OnConfigXClickCommon("C:Reverse Gravity"), 0, () => core.addonManager1.undressReverseOnGlobal)},

				{"Outfit",            new ButtonStructW1("Outfit"           ,  0, false, () => OnUndressOutfitClick(), 0, null)},
				{"Underwear",         new ButtonStructW1("Underwear"        ,  1, false, () => OnUndressUnderwearClick(), 0, null)},
				{"Misc",              new ButtonStructW1("Misc"             ,  2, false, () => OnUndressMiscClick(), 0, null)},
				{"H:Hint13",          new ButtonStructW1("Help for Partial...",3, true , () => OnHint13Click(), 0, null)},

				{"Remove Clothes",    new ButtonStructW1("Remove Clothes"   ,  0, false, () => OnUndressRemoveClick(), 0, null)},
				{"C:Opt1(NG Words)",  new ButtonStructW1("Opt1(NG Words)"   ,  1, false, () => OnConfigXClickCommon("C:Opt1(NG Words)"), 0, () => core.addonManager1.holdNGWordsOnGlobal)},
				{"C:Opt2(ClothSims)", new ButtonStructW1("Opt2(ClothSims)"  ,  2, false, () => OnConfigXClickCommon("C:Opt2(ClothSims)"), 0, () => core.addonManager1.holdClothSimOnGlobal)},
				{"C:Opt3(SkinWraps)", new ButtonStructW1("Opt3(SkinWraps)"  ,  3, false, () => OnConfigXClickCommon("C:Opt3(SkinWraps)"), 0, () => core.addonManager1.holdSkinWrapOnGlobal)},

				{"Alter Skin Wrap",   new ButtonStructW1("Alter Skin Wrap"      ,  0, false, () => OnAltSkinWrapClick(), 0, null)},
				{"C:SkinWrapToClothSim", new ButtonStructW1("SkinWrapToClothSim",  1, false, () => OnConfigXClickCommon("C:SkinWrapToClothSim"), 0, () => core.addonManager1.wrapAlterToSimOnGlobal)},
				{"H:Hint12",          new ButtonStructW1("Help for Undress..."  ,  2, true , () => OnHint12Click(), 0, null)},

				{"Stop",              new ButtonStructW1("Stop"                , -1, false, () => OnRemoveGameObjectsClick(), 0, null)},
				{"Expose Breasts",    new ButtonStructW1("Expose Breasts"      ,  0, false, () => OnExposeBreastsClick(), 0, null)},
				{"Tear Off Shirt",    new ButtonStructW1("Tear Off Shirt"      ,  1, false, () => OnExposeBreastsStrongClick(), 0, null)},
				{"Pull up a Skirt",   new ButtonStructW1("Pull up a Skirt"     ,  2, false, () => OnPullUpSkirtClick(), 0, null)},

				{"Restart",           new ButtonStructW1("Restart"                , -1, false, () => OnReturnControlOfPluginsClick(), 0, null)},
				{"Stop Other Plugins",new ButtonStructW1("Stop Other Plugins"     ,  0, false, () => OnTakeControlOfPluginsClick(), 0, null)},
				{"H:Hint1",           new ButtonStructW1("Help for stop other..." ,  1, true , () => OnHint1Click(), 0, null)},
			};
		}

		public void MakeMenuBodyPartAddon1Util(List<string> names, int y){		
			foreach(var name in names){	
				if(!ButtonStructW1Dic.ContainsKey(name)) continue;
				if(name.StartsWith("H:") && core.GetMemoryBool("isCompact")) continue;

				var bw1 = ButtonStructW1Dic[name];
				var displayName = bw1.name;
				var ui = MakeButton(displayName, bw1.x, worldUi.line + GetY(name), bw1.width, GetHeight(name));
				if(uiButtons.ContainsKey(name)) uiButtons[name] = ui;
				else uiButtons.Add(name, ui);
				uiButtons[name].button.onClick.AddListener(() => bw1.action());
				if(name.StartsWith("H:")) core.worldUi.ColoringButton(uiButtons[name], 0f, 0f, 0f);
				if(name.StartsWith("C:")){
					core.worldUi.ColoringButton(uiButtons[name], 198f/255f, 250f/255f, 192f/255f);
					var ext = LableChecker(name) ? "[ON]" : "[OFF]";
					uiButtons[name].buttonText.text = $"{displayName}{ext}";
				} 
			}
		}

		public void OnExpandCollapse(string type){
			var state = core.GetMemoryBool(type);
			core.SetMemoryBool(type, !state);
			worldUi.RemoveMenuBody();
			worldUi.MakeMenuBody();
		}

		public bool LableChecker(string name){
			if(!ButtonStructW1Dic.ContainsKey(name)){
				Utility.Debug($"ERROR: unknown name:{name}");
				return false;
			}
			 
			return ButtonStructW1Dic[name].func();
		}

		public float GetHeight(string name){
			int y = GetY(name);
			if(y == 7) return worldUi.buttonSizeY8;
			else return worldUi.buttonSizeY;
		}

		public int GetY(string name){
			if(name == "Undress Reset") return 7;
			else return 0;
		}

		public void MakeMenuBodyAddon(){
			var name = "Addon1";
			if(core.GetMemoryBool(name)){
				worldUi.worldUIBase.MakeMenuBodyPartExpandCollapseButton(name, false, uiExpandButtons);
				MakeMenuBodyPartAddon1();
			}else{
				if(!core.GetMemoryBool("isCompact")){
					worldUi.worldUIBase.MakeMenuBodyPartExpandCollapseButton(name, true, uiExpandButtons);
				}
			}
			worldUi.ColoringButton(uiExpandButtons.Values.ToArray(), 235f/255f, 161f/255f, 52f/255f);			
		}

		public Dictionary<string, Action> actions = new Dictionary<string, Action>();
		public Dictionary<string, UIDynamicButton> uiActionButtons = new Dictionary<string, UIDynamicButton>();

		public List<string> parts = new List<string>(){
			"hat",	"mask",	"glasses",	"accessory",	"jewely",	"gloves",	"stockings",
			"bra",	"shirt",	"sweater",	"top",	"dress",	"bodysuit",	"underwear",
			"panties",	"skirt",	"pants",	"bottom",	"shorts",	"socks",	"shoes",
		};

		public void MakeMenuBodyPartAddon1UndressPartial(){
			var xindents = new List<float>(){
				worldUi.quick1, worldUi.quick2, worldUi.quick3, worldUi.quick4, worldUi.quick5, worldUi.quick6, worldUi.quick7 
			};

			actions = new Dictionary<string, Action>();
			int num = 0;
			foreach(var part in parts){
				actions.Add(part, () => { UndressPartialCommon(part);});
				UIDynamicButton uiButton = MakeButton(part, xindents[num % 7], worldUi.line + (int)(num/7), worldUi.buttonSizeXQuick);
				if(uiActionButtons.ContainsKey(part)) uiActionButtons[part] = uiButton;
				else uiActionButtons.Add(part, uiButton);
				uiActionButtons[part].button.onClick.AddListener(() => actions[part]());
				num++;
			}
			worldUi.line+= 3;
		}

		public void MakeMenuBodyPartAddon1(){
			string[] list0= {"Undress Reset", "Undress CUAs", "C:Exclude Head", "H:Hint17",};
			string[] list1= {"Undress Overall", "C:Undress Detach", "C:Undress Forced", "C:Reverse Gravity",};
			string[] list2= {"Outfit", "Underwear", "Misc", "H:Hint13",};
			string[] list3= {"Remove Clothes", "C:Opt1(NG Words)", "C:Opt2(ClothSims)", "C:Opt3(SkinWraps)", };
			string[] list4 = {"Alter Skin Wrap", "C:SkinWrapToClothSim", "H:Hint12", };
			string[] list5 = {"Stop", "Expose Breasts", "Tear Off Shirt", "Pull up a Skirt", };
			string[] list6 = {"Restart", "Stop Other Plugins", "H:Hint1", };
			
			MakeMenuBodyPartAddon1Util(list0.ToList<string>(), worldUi.line); worldUi.line++;
			MakeMenuBodyPartAddon1Util(list1.ToList<string>(), worldUi.line); worldUi.line++;
			MakeMenuBodyPartAddon1Util(list2.ToList<string>(), worldUi.line); worldUi.line++;
			MakeMenuBodyPartAddon1UndressPartial();
			MakeMenuBodyPartAddon1Util(list3.ToList<string>(), worldUi.line); worldUi.line++;
			MakeMenuBodyPartAddon1Util(list4.ToList<string>(), worldUi.line); worldUi.line++;
			MakeMenuBodyPartAddon1Util(list5.ToList<string>(), worldUi.line); worldUi.line++;
			MakeMenuBodyPartAddon1Util(list6.ToList<string>(), worldUi.line); worldUi.line++;
		}
		
		public void OnExpandCollapseValueOnly(string type, bool expandOn){
			core.SetMemoryBool(type, expandOn);
		}
		
		public void LabelSwitch(){
			labels.ForEach(name => {ChangeLabel(name);});
		}

		public List<string> labels = new List<string>(){
			"C:Undress Detach", "C:Undress Forced", "C:Reverse Gravity", "C:Opt1(NG Words)", "C:Opt2(ClothSims)", "C:Opt3(SkinWraps)", "C:SkinWrapToClothSim",
		};

		public void ChangeLabel(string name){
			if(ButtonStructW1Dic == null) return;
			if(!ButtonStructW1Dic.ContainsKey(name)) return;
			if(uiButtons == null) return;
			if(!uiButtons.ContainsKey(name)) return;

			var bw2 = ButtonStructW1Dic[name];
			var ext = LableChecker(name) ? "[ON]" : "[OFF]";
			uiButtons[name].buttonText.text = ButtonStructW1Dic[name].name + ext;
		}
		
		public void OnUndressCUAsClick(){
			addonCUAUndresser.Undress();
		}

		public void UndressPreparingCommon(){
			addonClothSim.ChangeClothes();
			addonClothSim.BackupClothesState();
			addonClothSim.SetRemoveOptions();
		}

		public void UndressPartialCommon(string tags){
			core.EmbarrassALittle();
			UndressPreparingCommon();
			core.DoDelayedAction(() => core.DoAddonManager2Function("CoveringTargetsForUndress"), 1.5f);
			addonClothSim.Undress(core.addonManager1.undressPower, core.addonManager1.undressSeconds, false, false, core.addonManager1.undressDetachOnGlobal, tags);
		}

		public void OnUndressOverallClick(){
			UndressPartialCommon("");
		}

		public void OnUndressOutfitClick(){
			UndressPartialCommon("shirt,sweater,dress,bodysuit,shorts,skirt,pants");
		}

		public void OnUndressUnderwearClick(){
			UndressPartialCommon("bra,panties");
		}

		public void OnUndressMiscClick(){
			UndressPartialCommon("hat,jewely,accessory,glasses,jewely,gloves,stockings,socks,shoes,underwear");
		}
		
		public void OnAltSkinWrapClick(){
			addonClothSim.ChangeClothes();
			addonClothSim.BackupClothesState();
			addonClothSim.AlterSkinWrap();
		}
		
		public void OnUndressRecoverClick(){
			addonCUAUndresser.RecoverUndressing();
			addonClothSim.RemoveResetClothes();
			addonClothSim.RecoverClothesState();
			addonClothSim.WrapNormal();
			addonClothSim.ResetClothesState();
			addonClothSim.ChangeClothes();
			addonClothSim.BackupClothesState();
		}

		public void OnRemoveGameObjectsClick(){
			core.RemoveGameObjectByName("clothGrabber");
			core.customUnityAssetItem.suffixCount = 0;
		}

		public void OnExposeBreastsClick(){
			addonClothSim.ChangeClothes();
			core.customUnityAssetItem.MakeShirtSpreaders();;
		}

		public void OnExposeBreastsStrongClick(){
			addonClothSim.ChangeClothes();
			core.customUnityAssetItem.MakeShirtSpreaders(strStrengthType:"high");;
		}

		public void OnPullUpSkirtClick(){
			core.customUnityAssetItem.MakePullingSkirtUppers();
		}
		
		public void OnUndressRemoveClick(){
			addonClothSim.RecoverClothesState();
			addonClothSim.BackupClothesState();
			addonClothSim.SetRemoveOptions();
			addonClothSim.RemoveClothes();
		}

		public Dictionary<string, string> ExternalNameMap = new Dictionary<string, string>(){
			{"C:Exclude Head", "Undress CUAs Exclude Head"},
			{"C:Undress Detach", "Undress Detach"},
			{"C:Undress Forced", "Forced Undress"},
			{"C:Reverse Gravity", "Reverse Gravity"},
			{"C:Opt1(NG Words)", "Hold Clothes containing NGWords"},
			{"C:Opt2(ClothSims)", "Hold Clothes with ClothSim"},
			{"C:Opt3(SkinWraps)", "Hold Clothes with AlterSkinWrap"},
			{"C:SkinWrapToClothSim", "Alter SkinWrap to ClothSim"},
		};

		public void OnConfigXClickCommon(string keyName){
			if(!ExternalNameMap.ContainsKey(keyName)) return;
			OnConfigXClickCommon(keyName, ExternalNameMap[keyName]);
		}

		public void OnConfigXClickCommon(string keyName, string displayName){
			core.customUi.ToggleExternal(displayName);
			LabelSwitch();
			if(!ButtonStructW1Dic.ContainsKey(keyName)) return;
			ButtonStructW1 bs = ButtonStructW1Dic[keyName];
			var mem = bs.func();
			core.SetMemoryBool(keyName, mem);
		}

		// public void OnUndressOpt1Click(){
		// 	core.customUi.ToggleExternal("Hold Clothes containing NGWords");
		// 	core.worldUi.LabelSwitch();
		// }

		// public void OnUndressOpt2Click(){
		// 	core.customUi.ToggleExternal("Hold Clothes with ClothSim");
		// 	core.worldUi.LabelSwitch();
		// }

		// public void OnUndressOpt3Click(){
		// 	core.customUi.ToggleExternal("Hold Clothes with AlterSkinWrap");
		// 	core.worldUi.LabelSwitch();
		// }

		// public void OnUndressOpt4Click(){
		// 	core.customUi.ToggleExternal("Alter SkinWrap to ClothSim");
		// 	core.worldUi.LabelSwitch();
		// }

		// public void OnUndressOpt5Click(){
		// 	core.customUi.ToggleExternal("Forced Undress");
		// 	core.worldUi.LabelSwitch();
		// }

		// public void OnUndressOpt6Click(){
		// 	core.customUi.ToggleExternal("Reverse Gravity");
		// 	core.worldUi.LabelSwitch();
		// }

		// public void OnUndressOpt7Click(){
		// 	core.customUi.ToggleExternal("Undress Detach");
		// 	core.worldUi.LabelSwitch();
		// }

		// public void OnUndressOpt8Click(){
		// 	core.customUi.ToggleExternal("Undress CUAs Exclude Head");
		// 	core.worldUi.LabelSwitch();
		// }

		public void OnTakeControlOfPluginsClick(){
			addonTakeControl.TakeControlOfPlugins();
			addonTakeControl.TakeControlOfEverythingButPlugins();
		}

		public void OnHint1Click(){
			core.hint.StopOtherPluginsHint();
		}

		public void OnHint12Click(){
			core.hint.RemoveOptionsHint();
		}

		public void OnHint13Click(){
			core.hint.UndressPartialHint();
		}

		public void OnHint17Click(){
			core.hint.UndressCUAsHint();
		}

		public void OnReturnControlOfPluginsClick(){
			addonTakeControl.ReturnControlOfPlugins();
			addonTakeControl.ReturnControlOfEverythingButPlugins();
		}
		

		public UIDynamicButton MakeButton(string name, int x, int y, float width = 0f, float height = 0f){
			return worldUi.MakeButtonForAddon(name, x, y, width, height);
		}

		public UIDynamicButton MakeButton(string name, float x, int y, float width = 0f, float height = 0f, bool isMinimizeButton = false, bool isMinisizeButton = false){
			return worldUi.MakeButton(name, x, y, width, height, isMinimizeButton, isMinisizeButton);
		}
		public UIDynamicTextField MakeTextField(string text, int x, int y, float width = 0f, float height = 0f){
			return worldUi.MakeTextFieldForAddon(text, x, y, width, height);
		}

		public UIDynamicSlider MakeSlider(string name, int x, int y, float min, float max, float current){
			return worldUi.MakeSlider(name, x, y, min, max, current);
		}
	}
}