using System;
using UnityEngine;
using UnityEngine.UI;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
using MeshVR;
using GPUTools.Cloth.Scripts;
using GPUTools.Cloth.Scripts.Runtime.Data;
using GPUTools.Cloth.Scripts.Geometry.Data;

namespace Cotyounoyume{
	public class WrapManager
	{
		public EBATCore core;
        public ClothManager cm;
		public bool hasAltWrap = false;
		public bool hasRuntime = false;
		public List<string> AltWrapNames;
		public DAZSkinWrapSwitcher wrapSwitcher = null;
		public bool doneOnce = false;
		
		public WrapManager(EBATCore core, ClothManager cm){
			this.core = core;
            this.cm = cm;
			ReloadStatus();
			doneOnce = true;
		}
		
		public void ReloadStatus(){
			if(!hasAltWrap){
				AltWrapNames = new List<string>();
				hasRuntime = Utility.HasRuntime(cm.cloth);
				
				wrapSwitcher = cm.cloth.GetComponentInChildren<DAZSkinWrapSwitcher>();
				if(wrapSwitcher != null && wrapSwitcher.currentWrapName != ""){
					if(wrapSwitcher.IsStringChooserJSONParam("wrapName")){
						foreach(var name in wrapSwitcher.GetStringChooserJSONParamChoices("wrapName")){
							if(name != "Normal"){
								AltWrapNames.Add(name);
							}
						}
					}
					if(AltWrapNames.Count > 0) hasAltWrap = true;
				}else{
					wrapSwitcher = null;
				}
			}
		}
		
		public void Recover(){
			if(!hasAltWrap){ return; }
			
			WrapNormal();
		}
		
		public void WrapNormal(){
			if(!hasAltWrap){ return; }
			
			if(hasAltWrap && wrapSwitcher.currentWrapName != "Normal"){
				wrapSwitcher.SetCurrentWrapName("Normal");
			}
		}
		
		public void WrapAlter(){
			if(!hasAltWrap){ return; }
			
			if(hasAltWrap){
				if(cm.csch.wrapAlterToSimOn || !hasRuntime){
					int min = 0;
					int max = AltWrapNames.Count;
					if(max == 0) return;
					int rand = UnityEngine.Random.Range(min, max);
					string altWrapName = AltWrapNames[rand];
					wrapSwitcher.SetCurrentWrapName(altWrapName);
				}
			}
		}
		
		public void OnDestroy(){
			if(!hasAltWrap){ return; }
			
			if(doneOnce){
				WrapNormal();
				AltWrapNames.Clear();
				AltWrapNames = null;
				wrapSwitcher = null;
			}else{
				//nop
			}
		}
	}
}